module crc32                                    ! Cyclic redundancy check (CRC-32)

! Routines:
! 
! buildCRCTbl            Builds CRC-32 table 
! updtCRCInt1            Updates a running CRC-32
! writeFilCRCInt1        Writes data to file, updating the running CRC-32
! writeFilCRCInt2        Writes data to file, updating the running CRC-32
! writeFilCRCInt4        Writes data to file, updating the running CRC-32
! writeFilCRCInt8        Writes data to file, updating the running CRC-32
! writeFilCRCReal4       Writes data to file, updating the running CRC-32
! writeFilCRCReal8       Writes data to file, updating the running CRC-32
! writeFilCRCLogical     Writes data to file, updating the running CRC-32
! writeFilCRCStr         Writes data to file, updating the running CRC-32
! writeFilCRCStrArr      Writes data to file, updating the running CRC-32
! readFilCRCInt1         Reads data from file, updating the running CRC-32
! readFilCRCInt2         Reads data from file, updating the running CRC-32
! readFilCRCInt4         Reads data from file, updating the running CRC-32
! readFilCRCInt8         Reads data from file, updating the running CRC-32
! readFilCRCReal4        Reads data from file, updating the running CRC-32
! readFilCRCReal8        Reads data from file, updating the running CRC-32
! readFilCRCLogical      Reads data from file, updating the running CRC-32
! readFilCRCStr          Reads data from file, updating the running CRC-32
! readFilCRCStrArr       Reads data from file, updating the running CRC-32

  use basicTypes
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––



  !========================================================[ VARIABLES ]============================================================

  integer(i4b) :: &
   CRCTbl(0:255)


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  interface writeFilCRC
    module procedure writeFilCRCInt1
    module procedure writeFilCRCInt2
    module procedure writeFilCRCInt4
    module procedure writeFilCRCInt8
    module procedure writeFilCRCReal4
    module procedure writeFilCRCReal8
    module procedure writeFilCRCLogical
    module procedure writeFilCRCStr
    module procedure writeFilCRCStrArr
  end interface

  interface readFilCRC
    module procedure readFilCRCInt1
    module procedure readFilCRCInt2
    module procedure readFilCRCInt4
    module procedure readFilCRCInt8
    module procedure readFilCRCReal4
    module procedure readFilCRCReal8
    module procedure readFilCRCLogical
    module procedure readFilCRCStr
    module procedure readFilCRCStrArr
  end interface


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine buildCRCTbl
!
!   builds CRC-32 table (cf. [536])
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     k
    integer(i4b) :: &
     c, &
     n

    do n=0,255
      c=n
      do k=0,7
        if (iand(c,1) == 0) then
          c=ishft(c,-1)
        else
          c=ieor(z'edb88320',ishft(c,-1_i4b)) ! int(z'edb88320',i4b) would be correct but lead to (uncritical) arithmetic overflow
        end if
      end do
      CRCTbl(n)=c
    end do

  end subroutine buildCRCTbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine updtCRCInt1(CRC,arr)
!
!   updates a running CRC-32 (cf. [536])
!
!   Arguments:
!     CRC                            current CRC
!     arr                            data to update CRC with
!   
!   Requirements:
!     - buildCRCTbl() must be called before
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr(:)
    intent(inout) :: &
     CRC
    intent(in) :: &
     arr

    do i=1,ubound(arr,1)
      CRC=ieor(CRCTbl(iand(ieor(CRC,int(arr(i),i4b)),int(z'ff',i4b))),z'ffffffff') ! int(z'ffffffff',i4b) would be correct but lead to (uncritical) arithmetic overflow
    end do

  end subroutine updtCRCInt1




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeFilCRCInt1(un,arr,CRC)
!
!   writes data to file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr(:)
    intent(in) :: &
     un, &
     arr
    intent(inout) :: &
     CRC

    call updtCRCInt1(CRC,arr)
    write(un) arr

  end subroutine writeFilCRCInt1




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeFilCRCInt2(un,arr,CRC)
!
!   writes data to file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i2b) :: &
     arr(:)
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    intent(in) :: &
     un, &
     arr
    intent(inout) :: &
     CRC

    call updtCRCInt1(CRC,transfer(arr,arr1))
    write(un) arr

  end subroutine writeFilCRCInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeFilCRCInt4(un,arr,CRC)
!
!   writes data to file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC, &
     arr(:)
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    intent(in) :: &
     un, &
     arr
    intent(inout) :: &
     CRC

    call updtCRCInt1(CRC,transfer(arr,arr1))
    write(un) arr

  end subroutine writeFilCRCInt4




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeFilCRCInt8(un,arr,CRC)
!
!   writes data to file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i8b) :: &
     arr(:)
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    intent(in) :: &
     un, &
     arr
    intent(inout) :: &
     CRC

    call updtCRCInt1(CRC,transfer(arr,arr1))
    write(un) arr

  end subroutine writeFilCRCInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeFilCRCReal4(un,arr,CRC)
!
!   writes data to file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    real(sp) :: &
     arr(:)
    intent(in) :: &
     un, &
     arr
    intent(inout) :: &
     CRC

    call updtCRCInt1(CRC,transfer(arr,arr1))
    write(un) arr

  end subroutine writeFilCRCReal4




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeFilCRCReal8(un,arr,CRC)
!
!   writes data to file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un, &
     i
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    real(dp) :: &
     arr(:)
    intent(in) :: &
     un, &
     arr
    intent(inout) :: &
     CRC

    do i=1,size(arr,1)
      call updtCRCInt1(CRC,transfer(arr(i),arr1)) ! transfer() needs to be done by component
      write(un) arr(i)
    end do

  end subroutine writeFilCRCReal8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeFilCRCLogical(un,arr,CRC)
!
!   writes data to file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    logical :: &
     arr(:)
    intent(in) :: &
     un, &
     arr
    intent(inout) :: &
     CRC

    call updtCRCInt1(CRC,transfer(arr,arr1))
    write(un) arr

  end subroutine writeFilCRCLogical




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeFilCRCStr(un,str,CRC)
!
!   writes data to file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     str                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    character(*) :: &
     str
    intent(in) :: &
     un, &
     str
    intent(inout) :: &
     CRC

    call updtCRCInt1(CRC,transfer(str,arr1))
    write(un) str

  end subroutine writeFilCRCStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeFilCRCStrArr(un,arr,CRC)
!
!   writes data to file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    character(*) :: &
     arr(:)
    intent(in) :: &
     un, &
     arr
    intent(inout) :: &
     CRC

    call updtCRCInt1(CRC,transfer(arr,arr1))
    write(un) arr

  end subroutine writeFilCRCStrArr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readFilCRCInt1(un,arr,CRC)
!
!   reads data from file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr(:)
    intent(in) :: &
     un
    intent(inout) :: &
     CRC, &
     arr

    read(un) arr
    call updtCRCInt1(CRC,arr)

  end subroutine readFilCRCInt1




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readFilCRCInt2(un,arr,CRC)
!
!   reads data from file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i2b) :: &
     arr(:)
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    intent(in) :: &
     un
    intent(inout) :: &
     CRC, &
     arr

    read(un) arr
    call updtCRCInt1(CRC,transfer(arr,arr1))

  end subroutine readFilCRCInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readFilCRCInt4(un,arr,CRC)
!
!   reads data from file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC, &
     arr(:)
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    intent(in) :: &
     un
    intent(inout) :: &
     CRC, &
     arr

    read(un) arr
    call updtCRCInt1(CRC,transfer(arr,arr1))

  end subroutine readFilCRCInt4




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readFilCRCInt8(un,arr,CRC)
!
!   reads data from file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i8b) :: &
     arr(:)
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    intent(in) :: &
     un
    intent(inout) :: &
     CRC, &
     arr

    read(un) arr
    call updtCRCInt1(CRC,transfer(arr,arr1))

  end subroutine readFilCRCInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readFilCRCReal4(un,arr,CRC)
!
!   reads data from file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    real(sp) :: &
     arr(:)
    intent(in) :: &
     un
    intent(inout) :: &
     CRC, &
     arr

    read(un) arr
    call updtCRCInt1(CRC,transfer(arr,arr1))

  end subroutine readFilCRCReal4




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readFilCRCReal8(un,arr,CRC)
!
!   reads data from file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un, &
     i
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    real(dp) :: &
     arr(:)
    intent(in) :: &
     un
    intent(inout) :: &
     CRC, &
     arr

    do i=1,size(arr,1)
      read(un) arr(i)
      call updtCRCInt1(CRC,transfer(arr(i),arr1)) ! transfer() needs to be done by component
    end do

  end subroutine readFilCRCReal8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readFilCRCLogical(un,arr,CRC)
!
!   reads data from file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    logical :: &
     arr(:)
    intent(in) :: &
     un
    intent(inout) :: &
     CRC, &
     arr

    read(un) arr
    call updtCRCInt1(CRC,transfer(arr,arr1))

  end subroutine readFilCRCLogical




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readFilCRCStr(un,str,CRC)
!
!   reads data from file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     str                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    character(*) :: &
     str
    intent(in) :: &
     un
    intent(inout) :: &
     CRC, &
     str

    read(un) str
    call updtCRCInt1(CRC,transfer(str,arr1))

  end subroutine readFilCRCStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readFilCRCStrArr(un,arr,CRC)
!
!   reads data from file, updating the running CRC-32
!
!   Arguments:
!     un                             file unit
!     arr                            data
!     CRC                            current CRC
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     un
    integer(i4b) :: &
     CRC
    integer(i1b) :: &
     arr1(1)                         ! size irrelevant
    character(*) :: &
     arr(:)
    intent(in) :: &
     un
    intent(inout) :: &
     CRC, &
     arr

    read(un) arr
    call updtCRCInt1(CRC,transfer(arr,arr1))

  end subroutine readFilCRCStrArr




end module
