module kdTree                                               ! kd-trees

! Routines:
! 
! newTree                Returns: pointer to new kd-tree built from given data
! newLeaf                Returns: pointer to new kd-tree leaf
! getValArrInrNode       Returns: array of values of all leaves below an inner node
! getValArrLeaf          Returns: value of a leaf
! getVectArrInrNode      Returns: array of vectors of all leaves below a given inner node
! getVectArrLeaf         Returns: vector of a leaf
! estimIntgInrNode       Returns: integral estimated from all leaves below an inner node
! estimIntgLeaf          Returns: zero (in place of integral estimated over one leaf)

  use basicTypes
  use maths
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !========================================================[ CONSTANTS ]============================================================

  integer,parameter :: &
   N_LEAVES_OMP_MIN=8, &
   N_LEAVES_OMP_MAX=64
  
  private :: &
   N_LEAVES_OMP_MAX
   


  !=====================================================[ TYPE DEFINITIONS ]========================================================

  type,abstract :: tyKDTreeNode
    integer :: &
     nLeaves, &
     nDim
    contains
      procedure(getValArr),deferred :: getValArr
      procedure(getVectArr),deferred :: getVectArr
      procedure(estimIntg),deferred :: estimIntg
  end type
  abstract interface
    function getValArr(tree)
      import
      real(qp),allocatable :: &
       getValArr(:)
      class(tyKDTreeNode) :: &
       tree
      intent(in) :: &
       tree
    end function
    function getVectArr(tree)
      import
      real(qp),allocatable :: &
       getVectArr(:,:)
      class(tyKDTreeNode) :: &
       tree
!       intent(in) :: &
!        tree
    end function
    function estimIntg(tree,m)
      import
      integer :: &
       m
      real(qp) :: &
       estimIntg
      class(tyKDTreeNode) :: &
       tree
      intent(in) :: &
!        tree, &                     ! strangely leads to error "INTENT (IN) conflicts with INTENT(IN)" at compile time
       m
    end function
  end interface

  type,extends(tyKDTreeNode) :: tyKDTreeLeaf
    real(dp),allocatable :: &
     vect(:)
    real(qp) :: &
     val
    contains
      procedure :: getValArr => getValArrLeaf
      procedure :: getVectArr => getVectArrLeaf
      procedure :: estimIntg => estimIntgLeaf
!       final :: finaliseLeaf        ! finalisation not yet implemented in gfortran
  end type

  type,extends(tyKDTreeNode) :: tyKDTreeInnerNode
!     real(qp) :: &
    class(tyKDTreeNode),pointer :: &
     childL, &
     childR
    contains
      procedure :: getValArr => getValArrInrNode
      procedure :: getVectArr => getVectArrInrNode
      procedure :: estimIntg => estimIntgInrNode
!       final :: finaliseInrNode     ! finalisation not yet implemented in gfortran
  end type


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––


  interface newKDTree
    module procedure newTree
    module procedure newLeaf
  end interface
    


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function newTree(arrVect,arrVal)
!
!   Returns:                         pointer to new kd-tree built from given data
!
!   Arguments:
!     arrVect                        array of vectors
!     arrVal                         array of corresponding values
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     arrVect(:,:)                    ! indexes: samples, components
    real(qp) :: &
     arrVal(:)
    integer :: &
     nDim, &
     kMaxVce, &
     j
    integer(i8b) :: &
     nLeaves, &
     s(size(arrVect,1))
    class(tyKDTreeNode) :: &
     newTree
    pointer :: &
     newTree
    intent(in) :: &
     arrVect, &
     arrVal
    
    nLeaves=size(arrVect,1)
    nDim=size(arrVect,2)
    
    if (nLeaves == 1) then
      newTree=>newKDTree(arrVect(1,:),arrVal(1))
    else
      allocate(tyKDTreeInnerNode::newTree)
      
      select type (newTree)
      type is (tyKDTreeInnerNode)

        newTree%nLeaves=nLeaves
        newTree%nDim=nDim
        
        ! vector component with max. sample variance
        kMaxVce=maxloc([(arrVar(arrVect(:,j)), j=1,nDim)],1)
        
        ! median value in comp. kMaxVce
        call indxArr(arrVect(:,kMaxVce),s)
        if (nLeaves == 2) then
          newTree%childL=>newKDTree(arrVect(s(1),:),arrVal(s(1)))
          newTree%childR=>newKDTree(arrVect(s(2),:),arrVal(s(2)))
        else
          block
            integer(i8b) :: &
             iMed
            real(qp) :: &
             medVal
             
            iMed=ceiling(nLeaves/2._dp)
            medVal=arrVect(s(iMed),kMaxVce)
            do
              call dec(iMed)
              if (iMed == 1 .OR. arrVect(s(iMed),kMaxVce) /= medVal) exit
            end do

            !$OMP parallel sections if (N_LEAVES_OMP_MIN <= nLeaves .AND. nLeaves <= N_LEAVES_OMP_MAX)
            !$OMP section
            newTree%childL=>newKDTree(arrVect(s(:iMed),:),arrVal(s(:iMed)))
            !$OMP section
            newTree%childR=>newKDTree(arrVect(s(iMed+1:),:),arrVal(s(iMed+1:)))
            !$OMP end parallel sections
          end block
        end if

      end select

    end if

  end function newTree
    
    
    

!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function newLeaf(vect,val)
!
!   Returns:                         pointer to new kd-tree leaf
!
!   Arguments:
!     vect                           vector 
!     val                            corresponding value
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     allocStat
    real(dp) :: &
     vect(:)
    real(qp) :: &
     val
    type(tyKDTreeLeaf) :: &
     newLeaf
    pointer :: &
     newLeaf
    intent(in) :: &
     vect, &
     val
    
    allocate(newLeaf,stat=allocStat)
    call testAllocErr(allocStat,62)
    allocate(newLeaf%vect(size(vect)),source=vect,stat=allocStat) ! also copies value of vect into newLeaf%vect (checked)
    call testAllocErr(allocStat,63)
    newLeaf%val=val
    newLeaf%nLeaves=1
    newLeaf%nDim=size(vect)
    
  end function newLeaf
    
    
    

!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function getValArrInrNode(tree)
!
!   Returns:                         array of values of all leaves below an inner node
!
!   Arguments:
!     tree                           inner node
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     allocStat
    real(qp),allocatable :: &
     getValArrInrNode(:)
    class(tyKDTreeInnerNode) :: &
     tree
    intent(in) :: &
     tree
    
    select type (tree)
    type is (tyKDTreeInnerNode)
      allocate(getValArrInrNode(tree%nLeaves),stat=allocStat)
      call testAllocErr(allocStat,65)
      getValArrInrNode(:tree%childL%nLeaves)=tree%childL%getValArr()
      getValArrInrNode(tree%childL%nLeaves+1:)=tree%childR%getValArr()
    end select
    
  end function getValArrInrNode
    
    

!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function getValArrLeaf(tree)
!
!   Returns:                         value of a leaf
!
!   Arguments:
!     tree                           leaf
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     allocStat
    integer(i8b) :: &
     nLeavesL
    real(qp),allocatable :: &
     getValArrLeaf(:)
    class(tyKDTreeLeaf) :: &
     tree
    intent(in) :: &
     tree
    
    select type (tree)
    type is (tyKDTreeLeaf)
      allocate(getValArrLeaf(1),stat=allocStat)
      call testAllocErr(allocStat,64)
      getValArrLeaf(1)=tree%val
    end select
    
  end function getValArrLeaf
    
    
    

!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function getVectArrInrNode(tree)
!
!   Returns:                         array of vectors of all leaves below a given inner node
!
!   Arguments:
!     tree                           inner node
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     allocStat
    real(qp),allocatable :: &
     getVectArrInrNode(:,:)
    class(tyKDTreeInnerNode) :: &
     tree
!     intent(in) :: &
!      tree
    
    select type (tree)
    type is (tyKDTreeInnerNode)
      allocate(getVectArrInrNode(tree%nLeaves,tree%nDim),stat=allocStat)
      call testAllocErr(allocStat,73)
      getVectArrInrNode(:tree%childL%nLeaves,:)=tree%childL%getVectArr()
      getVectArrInrNode(tree%childL%nLeaves+1:,:)=tree%childR%getVectArr()
    end select
    
  end function getVectArrInrNode
    
    

!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function getVectArrLeaf(tree)
!
!   Returns:                         vector of a leaf
!
!   Arguments:
!     tree                           leaf
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     allocStat
    integer(i8b) :: &
     nLeavesL
    real(qp),allocatable :: &
     getVectArrLeaf(:,:)
    class(tyKDTreeLeaf) :: &
     tree
!     intent(in) :: &
!      tree
    
    select type (tree)
    type is (tyKDTreeLeaf)
      allocate(getVectArrLeaf(1,tree%nDim),stat=allocStat)
      call testAllocErr(allocStat,74)
      getVectArrLeaf(1,:)=tree%vect
    end select
    
  end function getVectArrLeaf
    
    

!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function estimIntgInrNode(tree,m)
!
!   Returns:                         integral estimated from all leaves below an inner node
!
!   Arguments:
!     tree                           inner node
!     m                              max. # leaves in each subvolume considered for estimation
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    class(tyKDTreeInnerNode) :: &
     tree
    integer :: &
     m, &
     allocStat
    real(qp) :: &
     estimIntgInrNode, &
     intg, &
     fArr(:)
    allocatable :: &
     fArr
    intent(in) :: &
!      tree, &                       ! strangely leads to error "INTENT (IN) conflicts with INTENT(IN)" at compile time
     m
  
    select type (tree)
    type is (tyKDTreeInnerNode)
      if (tree%nLeaves <= m) then
        allocate(fArr(tree%nLeaves),stat=allocStat)
        call testAllocErr(allocStat,66)
        estimIntgInrNode=median(tree%getValArr(),NO_FLAGS)*arrVol(tree%getVectArr())
      else
        intg=0._qp
        !$OMP parallel sections if (N_LEAVES_OMP_MIN <= tree%nLeaves .AND. tree%nLeaves <= N_LEAVES_OMP_MAX) reduction(+: intg)
        !$OMP section
        intg=tree%childL%estimIntg(m)
        !$OMP section
        intg=tree%childR%estimIntg(m)
        !$OMP end parallel sections
        estimIntgInrNode=intg
      end if
    end select

  end function estimIntgInrNode

    
    

!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function estimIntgLeaf(tree,m)
!
!   Returns:                         zero (in place of integral estimated over one leaf)
!
!   Arguments:
!     tree                           inner node
!     m                              (ignored)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    class(tyKDTreeLeaf) :: &
     tree
    integer :: &
     m, &
     allocStat, &
     s(:)
    real(qp) :: &
     estimIntgLeaf, &
     fArr(:)
    allocatable :: &
     fArr, &
     s
    intent(in) :: &
!      tree, &                       ! strangely leads to error "INTENT (IN) conflicts with INTENT(IN)" at compile time
     m
  
    estimIntgLeaf=0._dp

  end function estimIntgLeaf
    
    
    

! !:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! !
! !
!     subroutine finaliseLeaf(tree)
! !
! !
! !:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! 
!     implicit none
!     type(tyKDTreeLeaf) :: &
!      tree
!     intent(inout) :: &
!      tree
! 
!     deallocate(tree%vect)
!     
!   end subroutine finaliseLeaf
!     
!     
! 
! !:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! !
! !
!     subroutine finaliseInrNode(node)
! !
! !
! !:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
! 
!     implicit none
!     type(tyKDTreeInnerNode) :: &
!      node
!     intent(inout) :: &
!      node
! 
!     deallocate(node%childL,node%childR)
!     
!   end subroutine finaliseInrNode

end module
