module maths                                               ! Mathematics

! Routines:
! 
! isEven4                  Returns: whether a number is even
! isEven8                  Returns: whether a number is even
! incInt                   Increases value of integer variable by one
! incInt2                  Increases value of integer variable by one
! incInt8                  Increases value of integer variable by one
! decInt                   Decreases value of integer variable by one
! decInt2                  Decreases value of integer variable by one
! decInt8                  Decreases value of integer variable by one
! genSeq4                  Returns: sequence of natural numbers between and including two given numbers
! genSeq8                  Returns: sequence of natural numbers between and including two given numbers
! identInt8                Returns: the given natural number
! nxtInt8                  Returns: the next natural number
! allEqInt                 Returns: whether all elements of an array are equal
! allEqReal                Returns: whether all elements of an array are equal
! allEqDbl                 Returns: whether all elements of an array are equal
! allEqBool                Returns: whether all elements of an array are equal
! isInfDbl                 Returns: whether argument is infinity
! isInfQP                  Returns: whether argument is infinity
! eqlReal                  Returns: whether two variables are equal to within uncertainty of representation
! eqlDbl                   Returns: whether two variables are equal to within uncertainty of representation
! realRound                Returns: real number a rounded to an integer multiple of real number b in one of 3 modes
! realRoundInt             Returns: real number a rounded to an integer multiple of real number b in one of 3 modes
! dblRound                 Returns: real number a rounded to an integer multiple of real number b in one of 3 modes
! dblRoundInt              Returns: real number a rounded to an integer multiple of real number b in one of 3 modes
! i1Sub4                   Returns: lower index of sub-loop for given original loop
! i1Sub8                   Returns: lower index of sub-loop for given original loop
! i2Sub4                   Returns: upper index of sub-loop for given original loop
! i2Sub8                   Returns: upper index of sub-loop for given original loop
! trReal                   Returns: trace of a matrix
! trDbl                    Returns: trace of a matrix
! det2x2Real               Returns: determinant of a 2×2 matrix
! det2x2Dbl                Returns: determinant of a 2×2 matrix
! invMatrix2x2Real         Returns: inverse of a 2×2 matrix
! invMatrix2x2Dbl          Returns: inverse of a 2×2 matrix
! arrRgeDbl                Returns: minimum range containing all elements of an array
! minSinZeroPi             Returns: min_{x ∈ I} sin(x), I ⊂ [0,pi]
! maxSinZeroPi             Returns: max_{x ∈ I} sin(x), I ⊂ [0,pi]
! brktMin                  Returns: triplet (ax,bx,cx) bracketing a local minimum in sgn*fct
! posMin                   Returns: position of local minimum in sgn*fct between ax and cx
! dens1DUni                Returns: number proportional to uniform density U(-1,1)
! dens1DTriang             Returns: number proportional to triangular density (support (-1,1))
! dens1DEpa                Returns: number proportional to Epanechnikov density (support (-1,1))
! dens1DNorm               Returns: number proportional to normal density N(0,1)
! normalCDF                Returns: CDF of normal distribution
! kerDensEst1D             Performs kernel density estimation (KDE) based on 1D samples
! KSStat                   Returns: one-sample Kolmogorov-Smirnov statistic
! KSSignifLvl              Returns: Kolmogorov-Smirnov p-value of K
! dens2DEpa                Returns: number proportional to Epanechnikov density of |x|² (support: unit sphere in R²)
! dens2DBiWeight           Returns: number proportional to biweight density of |x|² (support: unit sphere in R²)
! dens2DTriWeight          Returns: number proportional to triweight density of |x|² (support: unit sphere in R²)
! dens2DNorm               Returns: number proportional to normal density of |x|² (support: unit sphere in R²)
! dens2DContourHeight      Returns: density level above which a given density over two dimensions covers a certain probability
! dens2DContourHeightPhi   Returns: function whose root w.r.t. h is a density level used by dens2DContourHeight
! medianDbl                Returns: median element of an array
! medianQP                 Returns: median element of an array
! arrMeanReal              Returns: mean of a set of samples
! arrMeanDbl               Returns: mean of a set of samples
! arrMeanQP                Returns: mean of a set of samples
! arrVarReal               Returns: variance of a set of samples
! arrVarDbl                Returns: variance of a set of samples
! arrVarQP                 Returns: variance of a set of samples
! arrCovReal               Returns: covariance of two sets of samples
! arrCovDbl                Returns: covariance of two sets of samples
! arrVolDbl                Returns: volume of smallest area confined by coordinate planes containing all given points
! arrVolQP                 Returns: volume of smallest area confined by coordinate planes containing all given points

  use basicTypes
  use coords
  use intervals
  use sorting
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !=====================================================[ TYPE DEFINITIONS ]========================================================
  
  
  ! ROUNDING
  
  type tyRoundMd
    integer :: up,down,nearest
  end type
  
  
  
  ! INTEGRAL ESTIMATION
  
  type tyIntgEstDbl
    real(dp) :: low,up,trapez
  end type
  type tyIntgEstQP
    real(qp) :: low,up,trapez
  end type
  
  
  ! KERNEL DENSITY ESTIMATION
  
  type tyKer1DInt
    integer :: norm,uni,triang,epa
  end type
  type tyKer1DDbl
    real(dp) :: norm,uni,triang,epa
  end type
  type tyKer1DBool
    logical :: norm,uni,triang,epa
  end type
  type tyKer2DInt
    integer :: norm,epa,biWeight,triWeight
  end type
  type tyKer2DDbl
    real(dp) :: norm,epa,biWeight,triWeight
  end type
  type tyKer2DBool
    logical :: norm,epa,biWeight,triWeight
  end type


  !========================================================[ CONSTANTS ]============================================================

  real(dp),parameter :: &
   PI=3.14159265358979323846264338327950288419716939937511_dp, &
   TWO_PI=2._dp*PI, &
   SQRT_PI=sqrt(PI), &
   SQRT_2_PI=sqrt(TWO_PI), &
   PI_HALF=PI/2._dp, &
   DEG_IN_RAD=TWO_PI/360._dp, &
   DEG_IN_ARCSEC=3600._dp, &
   ARCSEC_IN_RAD=DEG_IN_RAD/DEG_IN_ARCSEC, &
   MAS_IN_RAD=.001_dp*ARCSEC_IN_RAD, &
   MAS_IN_DEG=.001_dp*ARCSEC_IN_RAD/DEG_IN_RAD, &
   ARCSEC_IN_DEG=ARCSEC_IN_RAD/DEG_IN_RAD, &
   RJD_OFS=2.4e6_dp, &
   D_IN_S=24.*3600._dp, &            ! 1d/1s
   YR_IN_D=365.25_dp, &              ! 1yr/1d
   PC_IN_KM=3.08567758128e13_dp, &   ! 1pc/1km; value corrected 10.04.2010
   AU_IN_M=1.49597870691e11_dp, &    ! 1AU/1m  [http://ssd.jpl.nasa.gov/?constants]
   LY_IN_KM=9.4607304725808e12_dp, & ! [http://www.iau.org/public/measuring/]
   LY_IN_PC=LY_IN_KM/PC_IN_KM, &
   M_SUN_IN_MJUP=1047.66670178025913831244_dp, & ! M_Sun/M_Jup
   M_JUP_IN_KG=1.8986e27_dp, &       ! Mass of Jupiter in kg; NASA Jupiter Fact Sheet 2007
   M_SUN_IN_KG=M_SUN_IN_MJUP*M_JUP_IN_KG, & ! M_Sun in kg
   GRAVC_MKS=6.67428e-11_dp, &       ! Newtonian constant of gravitation; [m³/(kg s²)]; CODATA 2006; correct usage ensured 12.12.2009
   GRAVC=GRAVC_MKS*M_JUP_IN_KG, &    ! [m³/(M_Jup s²)]
   SUBSTELLAR_MASS_MAX_PHYS=80._dp, & ! physical upper limit for the mass of what's called a planet (or BD) [M_J]
   PLANET_MASS_MAX_PHYS=13._dp, &    ! physical upper limit for the mass of what's called a planet [M_J]
   CUBE_ROOT_2_PI_GRAVC=(TWO_PI*GRAVC)**(1._dp/3), &
   LOG_SCL_PSD_MIN=1e-12_dp

  character(*),parameter :: &
   DATE_DESC_AND_UNIT_GENERIC="Time [d]", &
   DATE_DESC_AND_UNIT_GENERIC_SHORT="Time [d]", &
   DESC_STDEV="standard deviation"

  
  ! ROUNDING
  
  type(tyRoundMd),parameter :: &
   ROUND_MD=tyRoundMd(1,2,3)
  
  
  ! KERNEL DENSITY ESTIMATION
  
  integer,parameter :: &
   N_KER_1D_TY=4, &
   N_KER_2D_TY=4
  real(dp),parameter :: &
   KER_1D_NORM_FAC(N_KER_1D_TY)=[real(dp) :: 1/sqrt(TWO_PI),.5,1,.75], &
   KER_1D_REL_BANDW(N_KER_1D_TY)=[real(dp) :: 1,1.740,2.432,sqrt(5.)], & ! cf. [402.table6.3], [404.table3.1], [08.11.2010/1]
   KER_2D_REL_BANDW(N_KER_2D_TY)=[real(dp) :: 1,192**(1._dp/6),2.78,3.12], & ! cf. [11.11.2010/1], [404.p87]
   KER_1D_EXTEND_X(N_KER_1D_TY)=[real(dp) :: sqrt(-2*log(sqrt(TWO_PI)*DIFF_EPS_DP)),VAL_UNDEF_DBL,VAL_UNDEF_DBL,VAL_UNDEF_DBL], & ! cf [08.11.2010/4]
   KER_2D_EXTEND_X(N_KER_2D_TY)=[real(dp) :: sqrt(-2*log(TWO_PI*DIFF_EPS_DP)),VAL_UNDEF_DBL,VAL_UNDEF_DBL,VAL_UNDEF_DBL] ! cf [18.11.2010/1]
  logical,parameter :: &
   KER_1D_FIN_SUP(N_KER_1D_TY)=[.FALSE.,.TRUE.,.TRUE.,.TRUE.], & ! supports of kernels finite?
   KER_2D_FIN_SUP(N_KER_2D_TY)=[.FALSE.,.TRUE.,.TRUE.,.TRUE.] ! supports of kernels finite?
  type(tyKer1DInt),parameter :: &
   I_KER_1D=tyKer1DInt(1,2,3,4)
  type(tyKer2DInt),parameter :: &
   I_KER_2D=tyKer2DInt(1,2,3,4)


  !========================================================[ VARIABLES ]============================================================

  character(STR_LEN_STD) :: &
   dateDescAndUnit, &
   dateDescAndUnitPlt


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  
  ! BASIC PROCEDURES
  
  interface isEven
    module procedure isEven4
    module procedure isEven8
  end interface

  interface inc
    module procedure incInt
    module procedure incInt2
    module procedure incInt8
  end interface

  interface dec
    module procedure decInt
    module procedure decInt2
    module procedure decInt8
  end interface

  interface genSeq
    module procedure genSeq4
    module procedure genSeq8
  end interface

  interface allEq
    module procedure allEqInt
    module procedure allEqReal
    module procedure allEqDbl
    module procedure allEqBool
  end interface
  
  interface isInf
    module procedure isInfDbl
    module procedure isInfQP
  end interface

  
  ! ROUNDING

  interface eql
    module procedure eqlReal
    module procedure eqlDbl
  end interface

  interface round
    module procedure realRound
    module procedure realRoundInt
    module procedure dblRound
    module procedure dblRoundInt
  end interface

  
  ! LOOPS

  interface i1Sub
    module procedure i1Sub4
    module procedure i1Sub8
  end interface

  interface i2Sub
    module procedure i2Sub4
    module procedure i2Sub8
  end interface


  ! LINEAR ALGEBRA

  interface tr
    module procedure trReal
    module procedure trDbl
  end interface

  interface det
    module procedure det2x2Real
    module procedure det2x2Dbl
  end interface

  interface invMatrix
    module procedure invMatrix2x2Real
    module procedure invMatrix2x2Dbl
  end interface


  ! EXTREMA

  interface arrRge
    module procedure arrRgeDbl
  end interface

  
  ! DENSITIES AND STATISTICS

  interface median
    module procedure medianDbl
    module procedure medianQP
  end interface

  interface arrMean
    module procedure arrMeanReal
    module procedure arrMeanDbl
    module procedure arrMeanQP
  end interface

  interface arrVar
    module procedure arrVarReal
    module procedure arrVarDbl
    module procedure arrVarQP
  end interface

  interface arrCov
    module procedure arrCovReal
    module procedure arrCovDbl
  end interface

  interface arrVol
    module procedure arrVolDbl
    module procedure arrVolQP
  end interface


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


  !====================================================[ BASIC PROCEDURES ]=========================================================


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function isEven4(val)
!
!   Returns:                         whether a number is even
!
!   Arguments:
!     val                            number in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     val
    logical :: &
     isEven4
    intent(in) :: &
     val

    isEven4=modulo(val,2) == 0
    
  end function isEven4




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function isEven8(val)
!
!   Returns:                         whether a number is even
!
!   Arguments:
!     val                            number in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     val
    logical :: &
     isEven8
    intent(in) :: &
     val

    isEven8=modulo(val,2_i8b) == 0
    
  end function isEven8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine incInt(var)
!
!   increases value of integer variable by one
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     var
    intent(inout) :: &
     var

    var=var+1

  end subroutine incInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine incInt2(var)
!
!   increases value of integer variable by one
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i2b) :: &
     var
    intent(inout) :: &
     var

    var=var+1_i2b

  end subroutine incInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine incInt8(var)
!
!   increases value of integer variable by one
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     var
    intent(inout) :: &
     var

    var=var+1_i8b

  end subroutine incInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine decInt(var)
!
!   decreases value of integer variable by one
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     var
    intent(inout) :: &
     var

    var=var-1

  end subroutine decInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine decInt2(var)
!
!   decreases value of integer variable by one
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i2b) :: &
     var
    intent(inout) :: &
     var

    var=var-1_i2b

  end subroutine decInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine decInt8(var)
!
!   decreases value of integer variable by one
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     var
    intent(inout) :: &
     var

    var=var-1_i8b

  end subroutine decInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function genSeq4(min,max)
!
!   Returns:                         sequence of natural numbers between and including two given numbers
!
!   Arguments:
!     min                            minimum number
!     max                            maximum number
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i4b) :: &
     min, &
     max, &
     genSeq4(max-min+1), &
     i
    intent(in) :: &
     min, &
     max

    do i=min,max
      genSeq4(i-min+1)=i
    end do

  end function genSeq4




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function genSeq8(min,max)
!
!   Returns:                         sequence of natural numbers between and including two given numbers
!
!   Arguments:
!     min                            minimum number
!     max                            maximum number
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     min, &
     max, &
     genSeq8(max-min+1), &
     i
    intent(in) :: &
     min, &
     max

    do i=min,max
      genSeq8(i-min+1)=i
    end do

  end function genSeq8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function identInt8(x)
!
!   Returns:                         the given natural number
!
!   Arguments:
!     x                              integer number
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     x, &
     identInt8
    intent(in) :: &
     x

    identInt8=x

  end function identInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function nxtInt8(x)
!
!   Returns:                         the next natural number
!
!   Arguments:
!     x                              integer number
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     x, &
     nxtInt8
    intent(in) :: &
     x

    nxtInt8=x+1

  end function nxtInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function allEqInt(arr)
!
!   Returns:                         whether all elements of an array are equal
!
!   Arguments:
!     arr                            array in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     arr(:)
    logical :: &
     allEqInt
    intent(in) :: &
     arr

    if (size(arr) <= 1) then
      allEqInt=.TRUE.
    else
      allEqInt=all(arr == arr(lbound(arr,1)))
    end if

  end function allEqInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function allEqReal(arr)
!
!   Returns:                         whether all elements of an array are equal
!
!   Arguments:
!     arr                            array in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     arr(:)
    logical :: &
     allEqReal
    intent(in) :: &
     arr

    if (size(arr) <= 1) then
      allEqReal=.TRUE.
    else
      allEqReal=all(arr == arr(lbound(arr,1)))
    end if

  end function allEqReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function allEqDbl(arr)
!
!   Returns:                         whether all elements of an array are equal
!
!   Arguments:
!     arr                            array in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     arr(:)
    logical :: &
     allEqDbl
    intent(in) :: &
     arr

    if (size(arr) <= 1) then
      allEqDbl=.TRUE.
    else
      allEqDbl=all(arr == arr(lbound(arr,1)))
    end if

  end function allEqDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function allEqBool(arr)
!
!   Returns:                         whether all elements of an array are equal
!
!   Arguments:
!     arr                            array in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     arr(:), &
     allEqBool
    intent(in) :: &
     arr

    if (size(arr) <= 1) then
      allEqBool=.TRUE.
    else
      allEqBool=all(arr .eqv. arr(lbound(arr,1)))
    end if

  end function allEqBool




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function isInfDbl(val)
!
!   Returns:                         whether argument is infinity
!
!   Arguments:
!     val                            value in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     val
    logical :: &
     isInfDbl
    intent(in) :: &
     val
    
    isInfDbl=.NOT.isnan(val) .AND. abs(val) > HUGE_DP
    
  end function isInfDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function isInfQP(val)
!
!   Returns:                         whether argument is infinity
!
!   Arguments:
!     val                            value in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(qp) :: &
     val
    logical :: &
     isInfQP
    intent(in) :: &
     val
    
    isInfQP=.NOT.isnan(val) .AND. abs(val) > HUGE_QP
    
  end function isInfQP




  !========================================================[ ROUNDING ]=============================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function eqlReal(a,b)
!
!   Returns:                         whether two variables are equal to within uncertainty of representation
!
!   Arguments:
!     a                              variable in question
!     b                              variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     a, &
     b
    logical :: &
     eqlReal
    intent(in) :: &
     a, &
     b

    eqlReal=abs(a-b) <= 2.*DIFF_EPS_SP

  end function eqlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function eqlDbl(a,b)
!
!   Returns:                         whether two variables are equal to within uncertainty of representation
!
!   Arguments:
!     a                              variable in question
!     b                              variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     a, &
     b
    logical :: &
     eqlDbl
    intent(in) :: &
     a, &
     b

    eqlDbl=abs(a-b) <= 2._dp*DIFF_EPS_DP

  end function eqlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function realRound(a,b,roundMode)
!
!   Returns:                         real number a rounded to an integer multiple of real number b in one of 3 modes;
!                                    if b not present, a rounded to the nearest power of ten
!
!   Arguments:
!     a                              number to be rounded
!     b                              number to be rounded to
!     roundMode                      rounding mode:
!                                      ROUND_MD%up:              return lowest number divisible by b which is higher than or equal to a
!                                      ROUND_MD%down or omitted: return highest number divisible by b which is lower than or equal to a
!                                      ROUND_MD%nearest:         round up or down, whichever result is nearer to a
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     roundMode, &
     scl
    real :: &
     a, &
     realRound, &
     b
    optional :: &
     b, &
     roundMode
    intent(in) :: &
     a, &
     b, &
     roundMode

    if (present(b)) then
      if (present(roundMode)) then
        if (roundMode == ROUND_MD%up) then ! round up
          realRound=ceiling(a/abs(b))*abs(b) ! return next-highest number divisible by b
        else if (roundMode == ROUND_MD%down) then ! round down
          realRound=floor(a/abs(b))*abs(b) ! return next-lowest number divisible by b
        else if (roundMode == ROUND_MD%nearest) then ! round to nearest
          realRound=sign(floor(abs(a/b)+.5)*b,a)
        else
          realRound=sign(floor(abs(a/b)+.5)*b,a) ! if mode is unrecognised,round to nearest
        end if
      else
        realRound=sign(floor(abs(a/b)+.5)*b,a) ! if mode is missing,round to nearest
      end if
    else
      scl=floor(log10(abs(a)))
      if (abs(a)*10.**(-scl) < 5.) then
        realRound=sign(10.**scl,a)
      else
        realRound=sign(10.**(scl+1),a)
      end if
    end if

  end function realRound




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function realRoundInt(a, b, roundMode)
!
!   Returns:                         real number a rounded to an integer multiple of real number b in one of 3 modes;
!                                    if b not present, a rounded to the nearest power of ten; return type integer
!
!   Arguments:
!     a                              number to be rounded
!     b                              number to be rounded to
!     roundMode                      rounding mode:
!                                      ROUND_MD%up:              return lowest number divisible by b which is higher than or equal to a
!                                      ROUND_MD%down or omitted: return highest number divisible by b which is lower than or equal to a
!                                      ROUND_MD%nearest:         round up or down, whichever result is nearer to a
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     realRoundInt, &
     b, &
     roundMode
    real :: &
     a
    optional :: &
     roundMode
    intent(in) :: &
     a, &
     b, &
     roundMode

    if (present(roundMode)) then
      if (roundMode == ROUND_MD%up) then ! round up
        realRoundInt=ceiling(a/abs(b))*abs(b) ! return next-highest number divisible by b
      else if (roundMode == ROUND_MD%down) then ! round down
        realRoundInt=floor(a/abs(b))*abs(b) ! return next-lowest number divisible by b
      else if (roundMode == ROUND_MD%nearest) then ! round to nearest
        realRoundInt=int(sign(1.*floor(abs(a/b)+.5)*b,a))
      else
        realRoundInt=int(sign(1.*floor(abs(a/b)+.5)*b,a)) ! if mode is unrecognised,round to nearest
      end if
    else
      realRoundInt=int(sign(1.*floor(abs(a/b)+.5)*b,a)) ! if mode is missing,round to nearest
    end if

  end function realRoundInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function dblRound(a,b,roundMode)
!
!   Returns:                         real number a rounded to an integer multiple of real number b in one of 3 modes;
!                                    if b not present, a rounded to the nearest power of ten
!
!   Arguments:
!     a                              number to be rounded
!     b                              number to be rounded to
!     roundMode                      rounding mode:
!                                      ROUND_MD%up:              return lowest number divisible by b which is higher than or equal to a
!                                      ROUND_MD%down or omitted: return highest number divisible by b which is lower than or equal to a
!                                      ROUND_MD%nearest:         round up or down, whichever result is nearer to a
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     roundMode, &
     scl
    real(dp) :: &
     a, &
     dblRound, &
     b
    optional :: &
     b, &
     roundMode
    intent(in) :: &
     a, &
     b, &
     roundMode

    if (present(b)) then
      if (present(roundMode)) then
        if (roundMode == ROUND_MD%up) then ! round up
          dblRound=ceiling(a/abs(b))*abs(b) ! return next-highest number divisible by b
        else if (roundMode == ROUND_MD%down) then ! round down
          dblRound=floor(a/abs(b))*abs(b) ! return next-lowest number divisible by b
        else if (roundMode == ROUND_MD%nearest) then ! round to nearest
          dblRound=sign(floor(abs(a/b)+.5_dp)*b,a)
        else
          dblRound=sign(floor(abs(a/b)+.5_dp)*b,a) ! if mode is unrecognised,round to nearest
        end if
      else
        dblRound=sign(floor(abs(a/b)+.5_dp)*b,a) ! if mode is missing,round to nearest
      end if
    else
      scl=floor(log10(abs(a)))
      if (abs(a)*10._dp**(-scl) < 5._dp) then
        dblRound=sign(10._dp**scl,a)
      else
        dblRound=sign(10._dp**(scl+1),a)
      end if
    end if

  end function dblRound




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function dblRoundInt(a, b, roundMode)
!
!   Returns:                         real number a rounded to an integer multiple of real number b in one of 3 modes;
!                                    if b not present, a rounded to the nearest power of ten; return type integer
!
!   Arguments:
!     a                              number to be rounded
!     b                              number to be rounded to
!     roundMode                      rounding mode:
!                                      ROUND_MD%up:              return lowest number divisible by b which is higher than or equal to a
!                                      ROUND_MD%down or omitted: return highest number divisible by b which is lower than or equal to a
!                                      ROUND_MD%nearest:         round up or down, whichever result is nearer to a
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     dblRoundInt, &
     b, &
     roundMode
    real(dp) :: &
     a
    optional :: &
     roundMode
    intent(in) :: &
     a, &
     b, &
     roundMode

    if (present(roundMode)) then
      if (roundMode == ROUND_MD%up) then ! round up
        dblRoundInt=ceiling(a/abs(b))*abs(b) ! return next-highest number divisible by b
      else if (roundMode == ROUND_MD%down) then ! round down
        dblRoundInt=floor(a/abs(b))*abs(b) ! return next-lowest number divisible by b
      else if (roundMode == ROUND_MD%nearest) then ! round to nearest
        dblRoundInt=int(sign(1._dp*floor(abs(a/b)+.5_dp)*b,a))
      else
        dblRoundInt=int(sign(1._dp*floor(abs(a/b)+.5_dp)*b,a)) ! if mode is unrecognised,round to nearest
      end if
    else
      dblRoundInt=int(sign(1._dp*floor(abs(a/b)+.5_dp)*b,a)) ! if mode is missing,round to nearest
    end if

  end function dblRoundInt




  !==========================================================[ LOOPS ]==============================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function i1Sub4(i1,i2,inc,nSub,iSub)
!
!   Returns:                         lower index of sub-loop for given original loop
!
!   Arguments:
!     i1                             lower index of original loop
!     i2                             upper index of original loop
!     inc                            increment of original loop
!     nSub                           # sub-loops
!     iSub                           index of current sub-loop
!
!   enables the partitioning of a given loop
!   useful for fast loops with many iterations and for OMP parallel loops, where it enables progress output
!   all original loop indexes will be covered in the correct order by a set of nSub sub-loops indexed by iSub=1…nSub
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i4b) :: &
     i1, &
     i2, &
     inc, &
     nSub, &
     iSub, &
     i1Sub4
    intent(in) :: &
     i1, &
     i2, &
     inc, &
     nSub, &
     iSub

    i1Sub4=i1+(iSub-1)*sign(ceiling(real(abs(i2-i1)+1,dp)/nSub),inc)

  end function i1Sub4




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function i1Sub8(i1,i2,inc,nSub,iSub)
!
!   Returns:                         lower index of sub-loop for given original loop
!
!   Arguments:
!     i1                             lower index of original loop
!     i2                             upper index of original loop
!     inc                            increment of original loop
!     nSub                           # sub-loops
!     iSub                           index of current sub-loop
!
!   enables the partitioning of a given loop
!   useful for fast loops with many iterations and for OMP parallel loops, where it enables progress output
!   all original loop indexes will be covered in the correct order by a set of nSub sub-loops indexed by iSub=1…nSub
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     i1, &
     i2, &
     inc, &
     nSub, &
     iSub, &
     i1Sub8
    intent(in) :: &
     i1, &
     i2, &
     inc, &
     nSub, &
     iSub

    i1Sub8=i1+(iSub-1)*sign(int(ceiling(real(abs(i2-i1)+1_i8b,dp)/nSub),i8b),inc)

  end function i1Sub8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function i2Sub4(i1,i2,inc,nSub,iSub)
!
!   Returns:                         upper index of sub-loop for given original loop
!
!   Arguments:
!     i1                             lower index of original loop
!     i2                             upper index of original loop
!     inc                            increment of original loop
!     nSub                           # sub-loops
!     iSub                           index of current sub-loop
!
!   enables the partitioning of a given loop
!   useful for fast loops with many iterations and for OMP parallel loops, where it enables progress output
!   all original loop indexes will be covered in the correct order by a set of nSub sub-loops indexed by iSub=1…nSub
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i4b) :: &
     i1, &
     i2, &
     inc, &
     nSub, &
     iSub, &
     i2Sub4
    intent(in) :: &
     i1, &
     i2, &
     inc, &
     nSub, &
     iSub

    i2Sub4=i1+iSub*sign(ceiling(real(abs(i2-i1)+1,dp)/nSub),inc)-sign(1,inc)
    if (i1 < i2) then
      i2Sub4=min(i2,i2Sub4)
    else
      i2Sub4=max(i2,i2Sub4)
    end if
    if (sign(1,i2-i1) /= sign(1,inc)) i2Sub4=i1Sub(i1,i2,inc,nSub,iSub)-sign(1,inc)

  end function i2Sub4




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function i2Sub8(i1,i2,inc,nSub,iSub)
!
!   Returns:                         upper index of sub-loop for given original loop
!
!   Arguments:
!     i1                             lower index of original loop
!     i2                             upper index of original loop
!     inc                            increment of original loop
!     nSub                           # sub-loops
!     iSub                           index of current sub-loop
!
!   enables the partitioning of a given loop
!   useful for fast loops with many iterations and for OMP parallel loops, where it enables progress output
!   all original loop indexes will be covered in the correct order by a set of nSub sub-loops indexed by iSub=1…nSub
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     i1, &
     i2, &
     inc, &
     nSub, &
     iSub, &
     i2Sub8
    intent(in) :: &
     i1, &
     i2, &
     inc, &
     nSub, &
     iSub

    i2Sub8=i1+iSub*sign(int(ceiling(real(abs(i2-i1)+1_i8b,dp)/nSub),i8b),inc)-sign(1_i8b,inc)
    if (i1 < i2) then
      i2Sub8=min(i2,i2Sub8)
    else
      i2Sub8=max(i2,i2Sub8)
    end if
    if (sign(1_i8b,i2-i1) /= sign(1_i8b,inc)) i2Sub8=i1Sub(i1,i2,inc,nSub,iSub)-sign(1_i8b,inc)

  end function i2Sub8




  !=====================================================[ LINEAR ALGEBRA ]==========================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function trReal(M)
!
!   Returns:                         trace of a matrix
!
!   Arguments:
!     M                              matrix in question
!
!   M IS CONTIGUOUS (as of rev. 115)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    real :: &
     trReal, &
     M(:,:)
    intent(in) :: &
     M

    trReal=0.
    do i=1,size(M,1)
      trReal=trReal+M(i,i)
    end do


  end function trReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function trDbl(M)
!
!   Returns:                         trace of a matrix
!
!   Arguments:
!     M                              matrix in question
!
!   M IS CONTIGUOUS (as of rev. 115)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    real(dp) :: &
     trDbl, &
     M(:,:)
    intent(in) :: &
     M

    trDbl=0._dp
    do i=1,size(M,1)
      trDbl=trDbl+M(i,i)
    end do


  end function trDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function det2x2Real(M)
!
!   Returns:                         determinant of a 2×2 matrix
!
!   Arguments:
!     M                              matrix in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     det2x2Real, &
     M(2,2)
    intent(in) :: &
     M

    det2x2Real=M(1,1)*M(2,2)-M(1,2)*M(2,1)

  end function det2x2Real




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function det2x2Dbl(M)
!
!   Returns:                         determinant of a 2×2 matrix
!
!   Arguments:
!     M                              matrix in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     det2x2Dbl, &
     M(2,2)
    intent(in) :: &
     M

    det2x2Dbl=M(1,1)*M(2,2)-M(1,2)*M(2,1)

  end function det2x2Dbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function invMatrix2x2Real(M)
!
!   Returns:                         inverse of a 2×2 matrix
!
!   Arguments:
!     M                              matrix in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     invMatrix2x2Real(2,2), &
     M(2,2), &
     d
    intent(in) :: &
     M

    d=det(M)
    if (d == 0) then
      call undef(invMatrix2x2Real)
    else
      invMatrix2x2Real(1,1)=M(2,2)
      invMatrix2x2Real(1,2)=-M(1,2)
      invMatrix2x2Real(2,1)=-M(2,1)
      invMatrix2x2Real(2,2)=M(1,1)
      invMatrix2x2Real=invMatrix2x2Real/d
    end if

  end function invMatrix2x2Real




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function invMatrix2x2Dbl(M)
!
!   Returns:                         inverse of a 2×2 matrix
!
!   Arguments:
!     M                              matrix in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     invMatrix2x2Dbl(2,2), &
     M(2,2), &
     d
    intent(in) :: &
     M

    d=det(M)
    if (d == 0) then
      call undef(invMatrix2x2Dbl)
    else
      invMatrix2x2Dbl(1,1)=M(2,2)
      invMatrix2x2Dbl(1,2)=-M(1,2)
      invMatrix2x2Dbl(2,1)=-M(2,1)
      invMatrix2x2Dbl(2,2)=M(1,1)
      invMatrix2x2Dbl=invMatrix2x2Dbl/d
    end if

  end function invMatrix2x2Dbl




  !=========================================================[ EXTREMA ]=============================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrRgeDbl(arr)
!
!   Returns:                         minimum range containing all elements of an array
!
!   Arguments:
!     arr                            array in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     i
    double precision :: &
     arr(:), &
     mi, &
     ma, &
     arrRgeDbl
    intent(in) :: &
     arr
    
    mi=HUGE_DP
    ma=-HUGE_DP
    do i=1,size(arr,1)
      mi=min(mi,arr(i))
      ma=max(ma,arr(i))
    end do
      
    arrRgeDbl=ma-mi
    
  end function




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function minSinZeroPi(xMin,xMax)
!
!   Returns:                         min_{x ∈ I} sin(x),   I ⊂ [0,pi]
!
!   Arguments:
!     xMin                           lower bound of I
!     xMax                           upper bound of I
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     xMin, &
     xMax, &
     minSinZeroPi
    intent(in) :: &
     xMin, &
     xMax

    if (xMin < PI_HALF) then
      if (xMax < PI_HALF) then
        minSinZeroPi=sin(xMin)
      else
        if (xMin < PI-xMax) then
          minSinZeroPi=sin(xMin)
        else
          minSinZeroPi=sin(xMax)
        end if
      end if
    else
      minSinZeroPi=sin(xMax)
    end if

  end function minSinZeroPi




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function maxSinZeroPi(xMin,xMax)
!
!   Returns:                         max_{x ∈ I} sin(x),   I ⊂ [0,pi]
!
!   Arguments:
!     xMin                           lower bound of I
!     xMax                           upper bound of I
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     xMin, &
     xMax, &
     maxSinZeroPi
    intent(in) :: &
     xMin, &
     xMax

    if (xMin < PI_HALF) then
      if (xMax < PI_HALF) then
        maxSinZeroPi=sin(xMax)
      else
        maxSinZeroPi=1
      end if
    else
      maxSinZeroPi=sin(xMin)
    end if

  end function maxSinZeroPi




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine brktMin(ax,bx,cx,sgn,fct,minThrld,valEquivMin)
!
!   Returns:                         triplet (ax,bx,cx) bracketing a local minimum in sgn*fct, s.t.:
!                                      * ax < bx < cx
!                                      * fct(bx) < fct(ax)
!                                      * fct(bx) < fct(cx)
!
!   Arguments:
!     ax                             initial position; element of resulting triplet
!     bx                             initial position; element of resulting triplet
!     cx                             element of resulting triplet
!     sgn                            sign applied to fct
!     fct                            function
!     minThrld                       upper bound on minima to be considered
!     valEquivMin                    highest value treated as a local minimum
!
!   searches in direction of decreasing sgn*fct from initial points ax, bx
!   adapted from [324] (mnbrak.f90)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp),parameter :: &
     GOLD=1.2_dp, &                ! originally 1.618034_dp
     GLIMIT=2._dp, &               ! originally 100.0_dp
     TINY=1.0e-100_dp
    real(dp) :: &
     ax, &
     bx, &
     cx, &
     sgn, &
     minThrld, &
     fa, &
     fb, &
     fc, &
     fu, &
     q, &
     r, &
     u, &
     ulim, &
     valEquivMin
    intent(in) :: &
     sgn, &
     minThrld, &
     valEquivMin
    intent(inout) :: &
     ax, &
     bx, &
     cx

    interface
      function fct(x)
        import
        implicit none
        real(dp) :: &
         fct, &
         x
        intent(in) :: &
         x
      end function
    end interface

    fa=sgn*fct(ax)
    fb=sgn*fct(bx)
    if (fb > fa) then
      call swap(ax,bx)
      call swap(fa,fb)
    end if
    cx=bx+GOLD*(bx-ax)
    fc=sgn*fct(cx)
    do
      if (fc <= valEquivMin .OR. (fb < fc .AND. fb <= minThrld)) exit ! added conditions fc <= valEquivMin, fb <= minThrld
      r=(bx-ax)*(fb-fc)
      q=(bx-cx)*(fb-fa)
      u=bx-((bx-cx)*q-(bx-ax)*r)/(2.0_dp*sign(max(abs(q-r),TINY),q-r))
      ulim=bx+GLIMIT*(cx-bx)
      if ((bx-u)*(u-cx) > 0.0) then
        fu=sgn*fct(u)
        if (fu < fc .AND. fu <= minThrld) then ! added condition fu <= minThrld
          ax=bx
          fa=fb
          bx=u
          fb=fu
          exit
        else if (fu > fb .AND. fb <= minThrld) then ! added condition fb <= minThrld
          cx=u
          fc=fu
          exit
        end if
        u=cx+GOLD*(cx-bx)
        fu=sgn*fct(u)
      else if ((cx-u)*(u-ulim) > 0.0) then
        fu=sgn*fct(u)
        if (fu < fc) then
          bx=cx
          cx=u
          u=cx+GOLD*(cx-bx)
          call shft4DP(fb,fc,fu,sgn*fct(u))
        end if
      else if ((u-ulim)*(ulim-cx) >= 0.0) then
        u=ulim
        fu=sgn*fct(u)
      else
        u=cx+GOLD*(cx-bx)
        fu=sgn*fct(u)
      end if
      call shft4DP(ax,bx,cx,u)
      call shft4DP(fa,fb,fc,fu)
    end do

  end subroutine brktMin




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function posMin(ax,bx,cx,tol,sgn,fct,minThrld,fx,valEquivMin)
!
!   Returns:                         position of local minimum in sgn*fct between ax and cx
!
!   Arguments:
!     ax                             element of bracketing triplet as returned by brktMin
!     bx                             element of bracketing triplet as returned by brktMin
!     cx                             element of bracketing triplet as returned by brktMin
!     tol                            requested relative precision of result
!     sgn                            sign applied to fct
!     fct                            function
!     minThrld                       upper bound on minima to be considered
!     fx                             ordinate of minimum
!     valEquivMin                    highest value treated as a local minimum
!
!   Requirements:
!     - local minimum bracketed by (ax,bx,cx) (see brktMin)
!
!   adapted from [324] (brent.f90)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i4b),parameter :: &
     ITMAX=10000
    real(dp),parameter :: &
     CGOLD=0.3819660_dp, &
     ZEPS=1.e-3_dp*EPS_DP
    integer(i4b) :: &
     iter
    real(dp) :: &
     ax, &
     bx, &
     cx, &
     tol, &
     sgn, &
     minThrld, &
     posMin, &
     a, &
     b, &
     d, &
     e, &
     etemp, &
     fu, &
     fv, &
     fw, &
     fx, &
     p, &
     q, &
     r, &
     tol1, &
     tol2, &
     u, &
     v, &
     w, &
     x, &
     xm, &
     valEquivMin
    intent(in) :: &
     ax, &
     bx, &
     cx, &
     tol, &
     sgn, &
     minThrld, &
     valEquivMin
    intent(out) :: &
     fx

    interface
      function fct(x)
        import
        implicit none
        real(dp) :: &
         fct, &
         x
        intent(in) :: &
         x
      end function
    end interface

    a=min(ax,cx)
    b=max(ax,cx)
    v=bx
    w=v
    x=v
    e=0.0
    fx=sgn*fct(x)
    fv=fx
    fw=fx
    do iter=1,ITMAX
      xm=0.5_dp*(a+b)
      tol1=tol*abs(x)+ZEPS
      tol2=2.0_dp*tol1
      if (fx <= valEquivMin .OR. (fx <= minThrld .AND. abs(x-xm) <= (tol2-0.5_dp*(b-a)))) then ! added conditions fx <= valEquivMin, fx <= minThrld
        posMin=x
        return
      end if
      if (abs(e) > tol1) then
        r=(x-w)*(fx-fv)
        q=(x-v)*(fx-fw)
        p=(x-v)*q-(x-w)*r
        q=2.0_dp*(q-r)
        if (q > 0.0) p=-p
        q=abs(q)
        etemp=e
        e=d
        if (abs(p) >= abs(0.5_dp*q*etemp) .or. &
          p <= q*(a-x) .or. p >= q*(b-x)) then
          e=merge(a-x,b-x, x >= xm )
          d=CGOLD*e
        else
          d=p/q
          u=x+d
          if (u-a < tol2 .or. b-u < tol2) d=sign(tol1,xm-x)
        end if
      else
        e=merge(a-x,b-x, x >= xm )
        d=CGOLD*e
      end if
      u=merge(x+d,x+sign(tol1,d), abs(d) >= tol1 )
      fu=sgn*fct(u)
      if (fu <= fx) then
        if (u >= x) then
          a=x
        else
          b=x
        end if
        call shft4DP(v,w,x,u)
        call shft4DP(fv,fw,fx,fu)
      else
        if (u < x) then
          a=u
        else
          b=u
        end if
        if (fu <= fw .or. w == x) then
          v=w
          fv=fw
          w=u
          fw=fu
        else if (fu <= fv .or. v == x .or. v == w) then
          v=u
          fv=fu
        end if
      end if
    end do
    call err('exceed maximum iterations while determining periodogram peak width')

  end function posMin



  !================================================[ DENSITIES AND STATISTICS ]=====================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dens1DUni(x)
!
!   Returns:                         number proportional to uniform density U(-1,1)
!
!   Arguments:
!     x                              position to calculate result at
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     dens1DUni, &
     x
    intent(in) :: &
     x

    dens1DUni=max(ceiling(1._dp-abs(x)),0)

  end function dens1DUni




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dens1DTriang(x)
!
!   Returns:                         number proportional to triangular density (support (-1,1))
!
!   Arguments:
!     x                              position to calculate result at
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     dens1DTriang, &
     x
    intent(in) :: &
     x

    dens1DTriang=max(1._dp-abs(x),0._dp)

  end function dens1DTriang




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dens1DEpa(x)
!
!   Returns:                         number proportional to Epanechnikov density (support (-1,1); cf. [402.p139])
!
!   Arguments:
!     x                              position to calculate result at
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     dens1DEpa, &
     x
    intent(in) :: &
     x

    dens1DEpa=max(1._dp-x**2,0._dp)

  end function dens1DEpa




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dens1DNorm(x)
!
!   Returns:                         number proportional to normal density N(0,1)
!
!   Arguments:
!     x                              position to calculate result at
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     dens1DNorm, &
     x
    intent(in) :: &
     x

    dens1DNorm=exp(-x**2/2._dp)

  end function dens1DNorm




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function normalCDF(x,mu,vce)
!
!   Returns:                         CDF of normal distribution
!
!   Arguments:
!     x                              position to calculate result at
!     mu                             mean of normal distribution
!     vce                            variance of normal distribution
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     normalCDF, &
     x, &
     mu, &
     vce
    intent(in) :: &
     x, &
     mu, &
     vce

    normalCDF=.5_dp*(1._dp+erf((x-mu)/sqrt(2._dp*vce)))

  end function normalCDF




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine kerDensEst1D(ker,iKer,arrIn,stDevIn,xRgeArg,M,densEst,flags,xRgeMax,wArrIn,bandw)
!
!   performs kernel density estimation (KDE) based on 1D samples
!
!   Arguments:
!     ker                            kernel
!     iKer                           kernel index
!     arrIn                          samples
!     stDevIn                        original standard deviation which the KDE should be based on
!     xRgeArg                        abscissa range
!     M                              # KDE points
!     densEst                        resulting KDE
!     flags                          option flags
!     xRgeMax                        maximal abscissa range
!     wArrIn                         weights to be applied to samples
!     bandw                          KDE bandwidth
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iKer, &
     allocStat, &
     flags
    integer(i8b) :: &
     M, &
     n, &
     j, &
     j1, &
     j2, &
     iArr(:)
    real(dp) :: &
     arrIn(:), &
     interQuartRge, &
     stDevIn, &
     h, &
     A, &
     d, &
     mul, &
     y(M), &
     invLogMul, &
     negLogA, &
     gMean, &
     wArr(:), &
     invH(:), &
     KDETmp(:), &
     wArrIn(:), &
     bandw
    logical :: &
     useW, &
     logScl
    type(tyCoord2DDbl) :: &
     densEst(M)
    type(tyIntvlDbl) :: &
     xRge, &
     xRgeArg, &
     xRgeMax
    allocatable :: &
     iArr, &
     wArr, &
     invH, &
     KDETmp
    optional :: &
     xRgeMax, &
     wArrIn, &
     bandw
    intent(in) :: &
     M, &
     iKer, &
     stDevIn, &
     xRgeArg, &
     xRgeMax, &
     flags
    intent(inout) :: &
     arrIn, &
     wArrIn, &
     bandw
    intent(out) :: &
     densEst
    interface
      function ker(x)                ! should have support of [-1,1] (if finite) or, e.g., be Normal(0,1)
        import
        real(dp) :: &
         ker, &
         x
        intent(in) :: &
         x
      end function
    end interface

    n=size(arrIn,1)                  ! # data

    useW=present(wArrIn)
    logScl=getFlg(flags,FLG%logScl)

    ! normalise and sort

    allocate(wArr(n),invH(n),stat=allocStat)
    call testAllocErr(allocStat,45)
    if (useW) then
      wArrIn=wArrIn/sum(wArrIn)      ! normalise weights (should not be necessary)
      if (.NOT. getFlg(flags,FLG%sorted)) then
        allocate(iArr(n),stat=allocStat)
        call testAllocErr(allocStat,31)
        call indxArr(arrIn,iArr)
        arrIn=arrIn(iArr)
        wArrIn=wArrIn(iArr)
      end if
      wArr=wArrIn
    else
      if (.NOT. getFlg(flags,FLG%sorted)) call sort(arrIn)
      wArr=1._dp/n
    end if


    ! bandwidth h

    interQuartRge=arrIn(ceiling(.75*n))-arrIn(ceiling(.25*n)) ! cf. [08.11.2010/1b]
    if (present(bandw) .AND. isDef(bandw)) then
      h=bandw
    else
      h=.9_dp*min(stDevIn,interQuartRge/1.34_dp)/n**.2_dp*KER_1D_REL_BANDW(iKer) ! [404.{(3.28),(3.30),(3.31)}]
    end if


    ! x range

    if (getFlg(flags,FLG%xtndRgeX)) then
      if (KER_1D_FIN_SUP(iKer)) then
        xRge=xtndIntvl(xRgeArg,h)
      else
        xRge=xtndIntvl(xRgeArg,h*KER_1D_EXTEND_X(iKer)) ! cf. [08.11.2010/4]
      end if
      if (present(xRgeMax)) xRge=xRge%sect(xRgeMax)
      if (logScl) then
        xRge%lBnd=max(LOG_SCL_PSD_MIN,xRge%lBnd)
        if (xRge%width() <= 0._dp) xRge=xRgeArg
      end if
    else
      xRge=xRgeArg
    end if

    if (logScl) then

      ! x multiplier and offset
      xRge%uBnd=min(HUGE_DP,xRge%uBnd)
      mul=(xRge%uBnd/xRge%lBnd)**(1._dp/max(M-1,1_i8b))
      invLogMul=1._dp/log10(mul)
      A=xRge%lBnd/mul                ! to be able to have lowest index of densEst be 1
      negLogA=-log10(A)

      ! y positions
      !$OMP parallel do
      do j=1,M
        y(j)=xFromILog(j)
      end do
      !$OMP end parallel do

    else

      ! x step and offset
      d=xRge%width()/max(M-1,1_i8b)
      A=xRge%lBnd-d                  ! to be able to have lowest index of densEst be 1

      ! y positions
      do j=1,M
        y(j)=xFromI(j)
      end do

    end if


    ! adjusting the bandwidth (according to [404] or, equivalently, [564])

    invH=1._dp/h

    if (getFlg(flags,FLG%adapt) .AND. M > 1) then

      allocate(KDETmp(n),stat=allocStat)
      call testAllocErr(allocStat,54)

      ! prelim. KDE
      KDETmp=KDE(arrIn,1_i8b,n,dens1DUni,I_KER_1D%uni,.FALSE.)

      ! geom. mean and new window width
      gMean=product(KDETmp(1:n)**(1._dp/n))
      invH=sqrt(KDETmp/gMean)/h

      do j=1,m
        y(j)=arrIn(min(n,max(1,nint(real(j-1)*(n-1)/(M-1)))))
      end do

      deallocate(KDETmp)

    else if (present(bandw) .AND. isUndef(bandw)) then ! return bandwidth

      bandw=h

    end if


    ! KDE

    allocate(KDETmp(M),stat=allocStat)
    call testAllocErr(allocStat,55)

    KDETmp=KDE(y,1_i8b,M,ker,iKer,.TRUE.)

    ! normalisation, final array
    densEst%x=y
    if (logScl) then
      densEst%y=KDETmp/(log10(mul)*(.5_dp*(KDETmp(1)+KDETmp(M))+sum(KDETmp(2:M-1)))) ! different normalisation (cf. [24.02.2011/1]): approx. by trapezoids over a log-scale abscissa
    else
      densEst%y=KER_1D_NORM_FAC(iKer)*KDETmp ! normalisation
    end if


    contains

    function indxL(x)                ! returns index of KDE point whose abscissa does not exceed x
      implicit none
      integer(i8b) :: &
       indxL
      real(dp) :: &
       x

      indxL=floor((x-A)/d)
    end function


    function indxH(x)                ! returns index of KDE point whose abscissa does not fall below x
      implicit none
      integer(i8b) :: &
       indxH
      real(dp) :: &
       x

      indxH=ceiling((x-A)/d)
    end function


    function xFromI(i)               ! returns abscissa for given index
      implicit none
      integer(i8b) :: &
       i
      real(dp) :: &
       xFromI

      xFromI=A+d*i
    end function


    function indxLLog(x)             ! returns index of KDE point whose abscissa does not exceed x (log-scale abscissa)
      implicit none
      integer(i8b) :: &
       indxLLog
      real(dp) :: &
       x

      indxLLog=floor((log10(x)+negLogA)*invLogMul)
    end function


    function indxHLog(x)             ! returns index of KDE point whose abscissa does not fall below x (log-scale abscissa)
      implicit none
      integer(i8b) :: &
       indxHLog
      real(dp) :: &
       x

      indxHLog=ceiling((log10(x)+negLogA)*invLogMul)
    end function


    function xFromILog(i)            ! returns abscissa for given index (log-scale abscissa)
      implicit none
      integer(i8b) :: &
       i
      real(dp) :: &
       xFromILog

      xFromILog=A*mul**i
    end function


    function KDE(y,jMin,jMax,ker,iKer,regSpc) ! performs actual KDE
      implicit none
      integer :: &
       iKer
      integer(i8b) :: &
       jMin, &
       jMax, &
       i, &
       j
      real(dp) :: &
       y(:), &
       x, &
       curW, &
       curInvH, &
       curFac, &
       KDE(jMin:jMax), &
       curH
      logical :: &
       regSpc
      intent(in) :: &
       y, &
       jMin, &
       jMax, &
       iKer, &
       regSpc
      interface
        function ker(x)              ! should have support of [-1,1] (if finite) or, e.g., be Normal(0,1)
          import
          real(dp) :: &
           ker, &
           x
          intent(in) :: &
           x
        end function
      end interface

      KDE=0

      if (KER_1D_FIN_SUP(iKer)) then ! finite support (-1,1)?

        if (regSpc) then

          if (logScl) then

            do i=1,n
              x=arrIn(i)
              curW=wArr(i)
              curInvH=invH(i)
              curH=1._dp/curInvH
              curFac=curW*curInvH
              j1=max(jMin,indxHLog(x-curH))
              j2=min(jMax,indxLLog(x+curH))
              !$OMP parallel do
              do j=j1,j2
                KDE(j)=KDE(j)+curFac*ker((y(j)-x)*curInvH)
              end do
              !$OMP end parallel do
            end do

          else                       ! linear scale

            do i=1,n
              x=arrIn(i)
              curW=wArr(i)
              curInvH=invH(i)
              curH=1._dp/curInvH
              curFac=curW*curInvH
              j1=max(jMin,indxH(x-curH))
              j2=min(jMax,indxL(x+curH))
              !$OMP parallel do
              do j=j1,j2
                KDE(j)=KDE(j)+curFac*ker((y(j)-x)*curInvH)
              end do
              !$OMP end parallel do
            end do

          end if

        else                         ! irreg. spacing of y

          do i=1,n
            x=arrIn(i)
            curW=wArr(i)
            curInvH=invH(i)
            curH=1._dp/curInvH
            curFac=curW*curInvH
            !$OMP parallel sections
            !$OMP section
            j1=posInSortedArr(y,x-curH,.TRUE.)
            !$OMP section
            j2=posInSortedArr(y,x+curH,.FALSE.)
            !$OMP end parallel sections
            if (isUndef(j1)) j1=huge(j1) ! no samples in kernel support → deactivate KDE
            if (isUndef(j2)) then
              j2=jMax
            else
              j2=j2-1
            end if
            !$OMP parallel do
            do j=j1,j2
              KDE(j)=KDE(j)+curFac*ker((y(j)-x)*curInvH)
            end do
            !$OMP end parallel do
          end do

        end if                       ! (regSpc)

      else                           ! infinite kernel support

        do i=1,n
          x=arrIn(i)
          curW=wArr(i)
          curInvH=invH(i)
          curFac=curW*curInvH
          !$OMP parallel do
          do j=jMin,jMax
            KDE(j)=KDE(j)+curFac*ker((y(j)-x)*curInvH)
          end do
          !$OMP end parallel do
        end do

      end if                         ! (KER_1D_FIN_SUP(iKer))
    end function

  end subroutine kerDensEst1D




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function KSStat(theorCDF,mu,vce,arr)
!
!   Returns:                         one-sample Kolmogorov-Smirnov statistic
!
!   Arguments:
!     theorCDF                       theoretical CDF
!     mu                             mean of theoretical density
!     vce                            variance of theoretical density
!     arr                            samples
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     KSStat, &
     arr(:), &
     x, &
     mu, &
     vce
    integer(i8b) :: &
     i, &
     N, &
     iSort(size(arr,1))
    intent(in) :: &
     mu, &
     vce, &
     arr

    interface
      function theorCDF(x,mu,vce)
        import
        implicit none
        real(dp) :: &
         theorCDF, &
         x, &
         mu, &
         vce
        intent(in) :: &
         x, &
         mu, &
         vce
      end function
    end interface

    call indxArr(arr,iSort)

    KSStat=-huge(KSStat)
    N=size(arr,1)
    do i=1,N
      x=arr(iSort(i))
      KSStat=max(KSStat,max(abs(real(i-1,dp)/N-theorCDF(x,mu,vce)),abs(1._dp*i/N-theorCDF(x,mu,vce))))
    end do

  end function KSStat




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function KSSignifLvl(K)
!
!   Returns:                         Kolmogorov-Smirnov p-value of K
!
!   Arguments:
!     K                              the KS statistic times sqrt(n_values), for a one-sample test
!
!   implementation partly inspired by [391.14.3]
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     N_IT=100000
    integer :: &
     i, &
     sgn
    real(dp) :: &
     KSSignifLvl, &
     K, &
     tmp, &
     a, &
     resTmp
    intent(in) :: &
     K

    resTmp=0._dp
    sgn=1
    a=-2._qp*K**2
    do i=1,N_IT
      tmp=exp(a*i**2)
      if (tmp < tiny(tmp)) exit
      resTmp=resTmp+sgn*tmp
      sgn=-sgn
    end do

    KSSignifLvl=2._qp*resTmp

  end function KSSignifLvl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dens2DEpa(x)
!
!   Returns:                         number proportional to Epanechnikov density of |x|² (support: unit sphere in R²; cf. [11.11.2010/1])
!
!   Arguments:
!     x                              abscissa in R²
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     dens2DEpa, &
     x
    intent(in) :: &
     x

    dens2DEpa=max(1._dp-x,0._dp)

  end function dens2DEpa




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dens2DBiWeight(x)
!
!   Returns:                         number proportional to biweight density of |x|² (support: unit sphere in R²; cf. [404.p76])
!
!   Arguments:
!     x                              abscissa in R²
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     dens2DBiWeight, &
     x
    intent(in) :: &
     x

    dens2DBiWeight=max(ceiling(1._dp-x),0)*(1._dp-x)**2 ! saving 1._dp-x will not save time

  end function dens2DBiWeight




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dens2DTriWeight(x)
!
!   Returns:                         number proportional to triweight density of |x|² (support: unit sphere in R²; cf. [404.p76])
!
!   Arguments:
!     x                              abscissa in R²
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     dens2DTriWeight, &
     x
    intent(in) :: &
     x

    dens2DTriWeight=max((1._dp-x)**3,0._dp)

  end function dens2DTriWeight




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dens2DNorm(x)
!
!   Returns:                         number proportional to normal density of |x|² (support: unit sphere in R²; cf. [404.p76])
!
!   Arguments:
!     x                              abscissa in R²
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     dens2DNorm, &
     x
    intent(in) :: &
     x

    dens2DNorm=exp(-.5_dp*x)

  end function dens2DNorm




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dens2DContourHeight(prob,M,dens)
!
!   Returns:                         density level above which a given density over two dimensions covers a certain probability
!
!   Arguments:
!     prob                           requested probability
!     M                              # density samples
!     dens                           density samples
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     prob(:), &
     dens2DContourHeight(size(prob)), &
     hMax, &
     hTmp, &
     dmy, &
     hMinCur, &
     hMaxCur, &
     densSum, &
     curProb, &
     densZ(M**2)
    integer(i8b) :: &
     M, &
     i, &
     iProb, &
     j, &
     s, &
     indx(size(prob))
    type(tyCoord3DDbl) :: &
     dens(M**2)
    intent(in) :: &
     prob, &
     M, &
     dens

    hMax=0._dp
    hMinCur=0._dp
    densSum=0._dp
    s=size(prob)
    call indxArr(prob,indx,OPT_SORT_DESCENDING)

    dmy=dens2DContourHeightPhi(0._dp,0._dp,M,densZ,0._dp,0._dp,clear=.TRUE.)

    densZ=dens%z
    do i=1,M
      do j=1,M
        hTmp=densZ((i-1)*M+j)
        densSum=densSum+hTmp
        hMax=max(hMax,hTmp)
      end do
    end do

    hMaxCur=hMax
    do iProb=1,s
      curProb=prob(indx(iProb))
      hTmp=findZero(dens2DContourHeightPhi,hMinCur,hMaxCur,DIFF_EPS_DP)
      if (isUndef(hTmp)) then
        hTmp=findZero(dens2DContourHeightPhi,0._dp,hMax,DIFF_EPS_DP)
        if (isUndef(hTmp)) then
          call undef(dens2DContourHeight)
          exit
        end if
      end if
      dens2DContourHeight(indx(iProb))=hTmp
      hMinCur=hTmp
    end do


    contains

    function findZero(func,h1,h2,accur) ! finds root known to lie between h1 and h2, using Brent's method (adapted from [324], zbrent.f90)
      implicit none
      real(dp) :: &
       h1, &
       h2, &
       accur, &
       findZero
      intent(in) :: &
       h1, &
       h2, &
       accur
      interface
        function func(h,prob,M,densZ,densSum,hMax,clear)
          import
          implicit none
          integer(i8b) :: &
           M
          real(dp) :: &
           h, &
           func, &
           prob, &
           densZ(M**2), &
           densSum, &
           hMax
          logical :: &
           clear
          optional :: &
           clear
          intent(in) :: &
           h, &
           prob, &
           M, &
           densZ, &
           densSum, &
           hMax, &
           clear
        end function
      end interface
      integer(i4b),parameter :: &
       ITMAX=100
      integer(i4b) :: &
       iter
      real(dp) :: &
       a, &
       b, &
       c, &
       d, &
       e, &
       fa, &
       fb, &
       fc, &
       p, &
       q, &
       r, &
       s, &
       tol1, &
       xm

      a=h1
      b=h2
      fa=func(a,curProb,M,densZ,densSum,hMax)
      fb=func(b,curProb,M,densZ,densSum,hMax)
      if (fa*fb > 0.0) then
        call undef(findZero)         ! err('root must be bracketed for findZero')
        return
      end if
      c=b
      fc=fb
      do iter=1,ITMAX
        if ((fb > 0.0 .and. fc > 0.0) .or. (fb < 0.0 .and. fc < 0.0)) then
          c=a
          fc=fa
          d=b-a
          e=d
        end if
        if (abs(fc) < abs(fb)) then
          a=b
          b=c
          c=a
          fa=fb
          fb=fc
          fc=fa
        end if
        tol1=2.0_dp*DIFF_EPS_DP*abs(b)+0.5_dp*accur
        xm=0.5_dp*(c-b)
        if (abs(xm) <= tol1 .or. fb == 0.0) then
          findZero=b
          return
        end if
        if (abs(e) >= tol1 .and. abs(fa) > abs(fb)) then
          s=fb/fa
          if (a == c) then
            p=2.0_dp*xm*s
            q=1.0_dp-s
          else
            q=fa/fc
            r=fb/fc
            p=s*(2.0_dp*xm*q*(q-r)-(b-a)*(r-1.0_dp))
            q=(q-1.0_dp)*(r-1.0_dp)*(s-1.0_dp)
          end if
          if (p > 0.0) q=-q
          p=abs(p)
          if (2.0_dp*p  <  min(3.0_dp*xm*q-abs(tol1*q),abs(e*q))) then
            e=d
            d=p/q
          else
            d=xm
            e=d
          end if
        else
          d=xm
          e=d
        end if
        a=b
        fa=fb
        b=b+merge(d,sign(tol1,xm), abs(d) > tol1 )
        fb=func(b,curProb,M,densZ,densSum,hMax)
      end do
      findZero=b
    end function

  end function dens2DContourHeight




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dens2DContourHeightPhi(h,prob,M,densZ,densSum,hMax,clear)
!
!   Returns:                         function whose root w.r.t. h is a density level used by dens2DContourHeight
!
!   Arguments:
!     h                              density level
!     prob                           requested probability (see dens2DContourHeight)
!     M                              # density samples
!     densZ                          samples of density ordinates
!     densSum                        sum of elements of densZ
!     hMax                           maximum element of densZ
!     clear                          whether saved data should be reset
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     N_SAVE_MAX=100
    integer :: &
     nSav=0
    integer(i8b) :: &
     M, &
     iMinSav(N_SAVE_MAX), &
     iMaxSav(N_SAVE_MAX), &
     jMinSav(N_SAVE_MAX), &
     jMaxSav(N_SAVE_MAX), &
     i, &
     j, &
     iMin, &
     iMax, &
     jMin, &
     jMax, &
     iMinCur, &
     iMaxCur, &
     jMinCur(M), &
     jMaxCur(M), &
     k, &
     kMinDiff
    real(dp) :: &
     dens2DContourHeightPhi, &
     h, &
     hSav(N_SAVE_MAX), &
     minDiff, &
     sumTmp, &
     hTmp, &
     prob, &
     densSum, &
     hMax, &
     densZ(M**2)
    logical :: &
     clear, &
     incIMin, &
     incJMin
    optional :: &
     clear
    save :: &
     hSav, &
     nSav, &
     iMinSav, &
     iMaxSav, &
     jMinSav, &
     jMaxSav
    intent(in) :: &
     h, &
     prob, &
     M, &
     densZ, &
     densSum, &
     clear, &
     hMax


    sumTmp=0._dp
    incIMin=.TRUE.
    if (present(clear) .AND. clear) then

      nSav=0
      call undef(dens2DContourHeightPhi)

    else if (h == hMax) then

      dens2DContourHeightPhi=-prob

    else if (h == 0) then

      dens2DContourHeightPhi=1._dp-prob

    else

      ! find saved iMin,iMax,… for smaller height

      if (nSav > 0) call findMaxValIndx(merge(h-hSav(1:nSav),HUGE_DP,hSav(1:nSav) < h),minDiff,kMinDiff,OPT_INVERT)
      if (nSav == 0 .OR. minDiff == hMax .OR. isUndef(kMinDiff)) then
        iMin=1
        iMax=M
        jMin=1
        jMax=M
      else
        iMin=iMinSav(kMinDiff)
        iMax=iMaxSav(kMinDiff)
        jMin=jMinSav(kMinDiff)
        jMax=jMaxSav(kMinDiff)
      end if


      ! sum over those density points where density >= h

      jMinCur=jMax
      jMaxCur=jMin

      do i=iMin,iMax                 ! scan in x-dir.

        incJMin=.TRUE.

        do j=jMin,jMax               ! scan in y-dir.

          hTmp=densZ((i-1)*M+j)
          if (hTmp >= h) then        ! we are at density >= h
            sumTmp=sumTmp+hTmp
            if (incJMin) then
              jMinCur(i)=j           ! set only once (for the first time we encounter a point with dens. >= h)
              incJMin=.FALSE.
            end if
            jMaxCur(i)=j             ! set always as long as density >= h
          end if

        end do                       ! scan in y-dir.

        if (.NOT. incJMin) then      ! we encountered a region with density >= h at this x
          if (incIMin) then
            iMinCur=i
            incIMin=.FALSE.
          end if
          iMaxCur=i
        end if

      end do                         ! scan in x-dir.


      if (incIMin) then              ! no points with dens. >= h found at all

        call undef(dens2DContourHeightPhi)

      else

        dens2DContourHeightPhi=sumTmp/densSum-prob


        ! save iMin,iMax,… for later calls

        if (nSav < N_SAVE_MAX) then
          nSav=nSav+1
          hSav(nSav)=h
          iMinSav(nSav)=iMinCur
          iMaxSav(nSav)=iMaxCur
          jMinSav(nSav)=minval(jMinCur)
          jMaxSav(nSav)=maxval(jMaxCur)
        end if

      end if

    end if                           !  (present(clear) .AND. clear)

  end function dens2DContourHeightPhi




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine kerDensEst2D(ker,iKer,arrIn1,arrIn2,covM,xRge1Arg,xRge2Arg,M,densEst,xtndXRge,xRge1Max,xRge2Max,zmCont,xRge1Zm, &
     xRge2Zm)
!
!   performs kernel density estimation (KDE) based on 2D samples
!
!   Arguments:
!     ker                            kernel
!     iKer                           kernel index
!     arrIn1                         first coordinates of samples (x1)
!     arrIn2                         second coordinates of samples (x2)
!     covM                           covariance matrix of samples
!     xRge1Arg                       range of first coordinate
!     xRge2Arg                       range of second coordinate
!     M                              # KDE points
!     densEst                        resulting KDE
!     xtndXRge                       whether coordinate ranges should be extended if necessary
!     xRge1Max                       maximal range of first coordinate
!     xRge2Max                       maximal range of second coordinate
!     zmCont                         requested prob. content of zoomed coordinate ranges
!     xRge1Zm                        zoomed first-coordinate range
!     xRge2Zm                        zoomed second-coordinate range
!
!   Requirements:
!     - arrIn1 and arrIn2 must have the same size
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iKer
    integer(i8b) :: &
     M, &
     i, &
     n, &
     j1Min, &
     j1Max, &
     j2Min, &
     j2Max, &
     j1, &
     j2
    real(dp) :: &
     arrIn1(:), &
     arrIn2(:), &
     tmpArr(M,M), &
     x1, &
     x2, &
     h, &
     A(2), &
     d(2), &
     y1, &
     y2, &
     covM(2,2), &
     covMInv(2,2), &
     eigenVMin, &
     eigenVMax, &
     trCovMInv, &
     delMax1, &
     delMax2, &
     d1, &
     d2, &
     detCovM, &
     cutV, &
     cutA, &
     covMInv11, &
     covMInv22, &
     covMInv12, &
     covMInv11H, &
     covMInv22H, &
     covMInvDiagH, &
     kap, &
     lam, &
     cosAlp, &
     sinAlp, &
     cutCont, &
     zmCont
    logical :: &
     xtndXRge
    type(tyCoord3DDbl) :: &
     densEst(M**2)
    type(tyIntvlDbl) :: &
     xRge1Arg, &
     xRge2Arg, &
     xRge1, &
     xRge2, &
     xRge1Max, &
     xRge2Max, &
     xRge1Zm, &
     xRge2Zm
    optional :: &
     zmCont, &
     xRge1Max, &
     xRge2Max, &
     xRge1Zm, &
     xRge2Zm
    intent(in) :: &
     M, &
     iKer, &
     covM, &
     xRge1Arg, &
     xRge2Arg, &
     arrIn1, &
     arrIn2, &
     xtndXRge, &
     xRge1Max, &
     xRge2Max, &
     zmCont
    intent(out) :: &
     densEst
    intent(inout) :: &
     xRge1Zm, &
     xRge2Zm
    interface
      function ker(x)                ! one of the standard kernels over R²
        import
        real(dp) :: &
         ker, &
         x
        intent(in) :: &
         x
      end function
    end interface


    ! initialisations

    covMInv=invMatrix(covM)
    covMInv11=covMInv(1,1)
    covMInv22=covMInv(2,2)
    covMInv12=covMInv(1,2)


    ! covariance matrix invertible?

    if (isUndef(covMInv11)) then

      do i=1,M**2
        call undef(densEst(i)%x)
        call undef(densEst(i)%y)
        call undef(densEst(i)%z)
      end do

    else

      n=size(arrIn1,1)
      detCovM=det(covM)

      ! bandwidth h

      h=1._dp/n**(1._dp/6)*KER_2D_REL_BANDW(iKer) ! [11.11.2010/1]
      covMInv11H=covMInv11/h**2
      covMInv22H=covMInv22/h**2
      covMInvDiagH=2._dp*covMInv12/h**2

      ! maximum delta(x) where kernel may be nonzero

      call undef(delMax1)
      call undef(delMax2)

      if (KER_2D_FIN_SUP(iKer)) then ! finite support? (assumed to be unit sphere in R²)

        trCovMInv=tr(covM)/detCovM   ! cf. [24.01.2011/4]
        eigenVMin=trCovMInv/2._dp-sqrt((trCovMInv/2._dp)**2-1._dp/detCovM)

        if (eigenVMin /= 0._dp) then
          eigenVMax=trCovMInv/2._dp+sqrt((trCovMInv/2._dp)**2-1._dp/detCovM)
          kap=h/sqrt(eigenVMin)
          lam=h/sqrt(eigenVMax)
          cosAlp=covMInv12/sqrt(covMInv12**2+(eigenVMin-covMInv11)**2)
          sinAlp=sqrt(1._dp-cosAlp**2)
          delMax1=sqrt((kap*cosAlp)**2+(lam*sinAlp)**2)
          delMax2=sqrt((kap*sinAlp)**2+(lam*cosAlp)**2)
        end if

      end if


      ! x,y ranges

      if (xtndXRge) then
        if (isDef(delMax1)) then
          if (KER_2D_FIN_SUP(iKer)) then
            xRge1=xtndIntvl(xRge1Arg,delMax1)
            xRge2=xtndIntvl(xRge2Arg,delMax2)
          else
            xRge1=xtndIntvl(xRge1Arg,delMax1*KER_2D_EXTEND_X(iKer)) ! cf. [08.11.2010/4],[18.11.2010/1]
            xRge2=xtndIntvl(xRge2Arg,delMax2*KER_2D_EXTEND_X(iKer))
          end if
          if (present(xRge1Max)) xRge1=xRge1%sect(xRge1Max)
          if (present(xRge2Max)) xRge2=xRge2%sect(xRge2Max)
        else                         ! delMax… undefined
          if (present(xRge1Max)) then
            xRge1=xRge1Max           ! use max. ranges if given
          else
            xRge1=xRge1Arg
          end if
          if (present(xRge2Max)) then
            xRge2=xRge2Max
          else
            xRge2=xRge2Arg
          end if
        end if
      else
        xRge1=xRge1Arg
        xRge2=xRge2Arg
      end if


      ! x steps and deltas

      d(1)=xRge1%width()/M
      A(1)=xRge1%lBnd-d(1)           ! to be able to have lowest index of densEst be 1
      d(2)=xRge2%width()/M
      A(2)=xRge2%lBnd-d(2)           ! to be able to have lowest index of densEst be 1


      ! do KDE

      if ((KER_2D_FIN_SUP(iKer) .AND. eigenVMin == 0._dp) .OR. .NOT. KER_2D_FIN_SUP(iKer)) then ! KDE for all of the points
        tmpArr=0
        do i=1,n
          x1=arrIn1(i)
          x2=arrIn2(i)
          !$OMP parallel do private(j2,y1,y2,d1,d2)
          do j1=1,M
            do j2=1,M
              y1=xFromI(1,j1)
              y2=xFromI(2,j2)
              d1=x1-y1
              d2=x2-y2
              tmpArr(j1,j2)=tmpArr(j1,j2)+ker(covMInv11H*d1**2+d1*d2*covMInvDiagH+covMInv22H*d2**2)
            end do
          end do
          !$OMP end parallel do
        end do
      else                           ! KDE for part of the points
        tmpArr=0
        do i=1,n
          x1=arrIn1(i)
          x2=arrIn2(i)
          j1Min=max(1_i8b,min(M,indxH(1,x1-delMax1)))
          j1Max=max(1_i8b,min(M,indxL(1,x1+delMax1)))
          j2Min=max(1_i8b,min(M,indxH(2,x2-delMax2)))
          j2Max=max(1_i8b,min(M,indxL(2,x2+delMax2)))
          if (j1Max < j1Min) call swap(j1Min,j1Max) ! can be necessary for very small delMax[12]
          if (j2Max < j2Min) call swap(j2Min,j2Max)

          !$OMP parallel do private(j2,y1,y2,d1,d2)
          do j1=j1Min,j1Max
            do j2=j2Min,j2Max
              y1=xFromI(1,j1)
              y2=xFromI(2,j2)
              d1=x1-y1
              d2=x2-y2
              tmpArr(j1,j2)=tmpArr(j1,j2)+ker(covMInv11H*d1**2+d1*d2*covMInvDiagH+covMInv22H*d2**2)
            end do
          end do
          !$OMP end parallel do
        end do
      end if


      ! alternative, more secure normalisation
      tmpArr=tmpArr/(d(1)*d(2)*(sum(tmpArr)-.5_dp*sum(tmpArr(:,1))+sum(tmpArr(:,M))+sum(tmpArr(1,:))+sum(tmpArr(M,:))- &
       .25_dp*(tmpArr(1,1)+tmpArr(1,M)+tmpArr(M,1)+tmpArr(M,M))))
       ! margin points are only part of half as many elements as inner ones, corner points of 1/4 as many (cf. [20.12.2010/2])


      !$OMP parallel do private(j2,i)
      do j1=1,M
        do j2=1,M
          i=jFin(j1,j2)
          densEst(i)%x=xFromI(1,j1)
          densEst(i)%y=xFromI(2,j2)
          densEst(i)%z=tmpArr(j1,j2)
        end do
      end do
      !$OMP end parallel do


      if (present(zmCont)) then

        ! zooming-bounds: cut away region with prob. content <= cutCont (cf. [20.12.2010/2])

        cutCont=1._dp-zmCont
        cutA=d(1)*d(2)/2._dp

        j1Min=1
        j1Max=M
        j2Min=1
        j2Max=M

        ! x_min
        cutV=0
        do j1=1,M-1
          cutV=cutV+cutA*(sum(tmpArr(j1:j1+1,j2Min:j2Max))- &
           .5_dp*(tmpArr(j1,j2Min)+tmpArr(j1,j2Max)+tmpArr(j1+1,j2Min)+tmpArr(j1+1,j2Max)))
          if (cutV > cutCont/4._dp) then
            j1Min=j1
            exit
          end if
        end do

        ! y_min
        cutV=0
        do j2=1,M-1
          cutV=cutV+cutA*(sum(tmpArr(j1Min:j1Max,j2:j2+1))- &
           .5_dp*(tmpArr(j1Min,j2)+tmpArr(j1Max,j2)+tmpArr(j1Min,j2+1)+tmpArr(j1Max,j2+1)))
          if (cutV > cutCont/4._dp) then
            j2Min=j2
            exit
          end if
        end do

        ! x_max
        cutV=0
        do j1=M,j1Min+2,-1
          cutV=cutV+cutA*(sum(tmpArr(j1-1:j1,j2Min:j2Max))- &
           .5_dp*(tmpArr(j1-1,j2Min)+tmpArr(j1-1,j2Max)+tmpArr(j1,j2Min)+tmpArr(j1,j2Max)))
          if (cutV > cutCont/4._dp) then
            j1Max=j1
            exit
          end if
        end do

        ! y_max
        cutV=0
        do j2=M,j2Min+2,-1
          cutV=cutV+cutA*(sum(tmpArr(j1Min:j1Max,j2-1:j2))- &
           .5_dp*(tmpArr(j1Min,j2-1)+tmpArr(j1Max,j2-1)+tmpArr(j1Min,j2)+tmpArr(j1Max,j2)))
          if (cutV > cutCont/4._dp) then
            j2Max=j2
            exit
          end if
        end do

        if (j1Min == 1 .AND. j1Max == M) then
          call undef(xRge1Zm)
        else
          xRge1Zm=tyIntvlDbl(densEst(jFin(j1Min,1_i8b))%x,densEst(jFin(j1Max,1_i8b))%x)
        end if
        if (j2Min == 1 .AND. j2Max == M) then
          call undef(xRge2Zm)
        else
          xRge2Zm=tyIntvlDbl(densEst(jFin(1_i8b,j2Min))%y,densEst(jFin(1_i8b,j2Max))%y)
        end if

      end if                         ! (present(xRge1Zm) .AND. present(xRge2Zm))

    end if                           ! (isUndef(covMInv11))


    contains

    function indxL(iD,x)             ! returns index of KDE point whose abscissa in a given dimension does not exceed x
      implicit none
      integer :: &
       iD
      integer(i8b) :: &
       indxL
      real(dp) :: &
       x

      indxL=floor((x-A(iD))/d(iD))
    end function


    function indxH(iD,x)             ! returns index of KDE point whose abscissa in a given dimension does not fall below x
      implicit none
      integer :: &
       iD
      integer(i8b) :: &
       indxH
      real(dp) :: &
       x

      indxH=ceiling((x-A(iD))/d(iD))
    end function


    function xFromI(iD,i)            ! returns abscissa for given index and dimension
      implicit none
      integer :: &
       iD
      integer(i8b) :: &
       i
      real(dp) :: &
       xFromI

      xFromI=A(iD)+i*d(iD)
    end function


    function jFin(j1,j2)             ! returns index of final 1D array densEst, given an index for each abscissa
      implicit none
      integer(i8b) :: &
       j1, &
       j2, &
       jFin

      jFin=(j1-1)*M+j2
    end function

  end subroutine kerDensEst2D




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function medianDbl(arr,flags)
!
!   Returns:                         median element of an array
!
!   Arguments:
!     arr                            array in question
!     flags                          option flags
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     flags
    real(dp) :: &
     medianDbl, &
     arr(:), &
     a, &
     b
    integer(i8b) :: &
     n, &
     c
    intent(in) :: &
     arr, &
     flags
    
    n=size(arr,1)
    select case (n)
    case (0)
      call undef(medianDbl)
    case (1)
      medianDbl=arr(1)
    case default
      c=ceiling(.5_dp*n)
      if (getFlg(flags,FLG%sorted)) then
        a=arr(c)
        b=arr(c+1)
      else
        block
          integer(i8b) :: &
           s(size(arr,1))
          call indxArr(arr,s)
          a=arr(s(c))
          b=arr(s(c+1))
        end block
      end if

      if (isEven(n)) then
        medianDbl=.5_dp*(a+b)
      else
        medianDbl=a
      end if
    end select
    
  end function medianDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function medianQP(arr,flags)
!
!   Returns:                         median element of an array
!
!   Arguments:
!     arr                            array in question
!     flags                          option flags
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     flags
    real(qp) :: &
     medianQP, &
     arr(:), &
     a, &
     b
    integer(i8b) :: &
     n, &
     c
    intent(in) :: &
     arr, &
     flags
    
    n=size(arr,1)
    select case (n)
    case (0)
      call undef(medianQP)
    case (1)
      medianQP=arr(1)
    case default
      c=ceiling(.5_dp*n)
      if (getFlg(flags,FLG%sorted)) then
        a=arr(c)
        b=arr(c+1)
      else
        block
          integer(i8b) :: &
           s(size(arr,1))
          call indxArr(arr,s)
          a=arr(s(c))
          b=arr(s(c+1))
        end block
      end if

      if (isEven(n)) then
        medianQP=.5_dp*(a+b)
      else
        medianQP=a
      end if
    end select
    
  end function medianQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrMeanReal(arr,weights)
!
!   Returns:                         mean of a set of samples
!
!   Arguments:
!     arr                            samples
!     weights                        weights to be applied to the samples
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     arr(:), &
     arrMeanReal, &
     weights(:)
    optional :: &
     weights
    intent(in) :: &
     arr, &
     weights

    if (present(weights)) then
      arrMeanReal=real(sum(real(weights,dp)*arr)/sum(weights),sp)
    else
      arrMeanReal=real(sum(real(arr,dp))/size(arr),sp)
    end if

  end function arrMeanReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrMeanDbl(arr,weights)
!
!   Returns:                         mean of a set of samples
!
!   Arguments:
!     arr                            samples
!     weights                        weights to be applied to the samples
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     arr(:), &
     arrMeanDbl, &
     weights(:)
    optional :: &
     weights
    intent(in) :: &
     arr, &
     weights

    if (present(weights)) then
      arrMeanDbl=real(sum(real(weights,qp)*arr)/sum(weights),dp)
    else
      arrMeanDbl=real(sum(real(arr,qp))/size(arr),dp)
    end if

  end function arrMeanDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrMeanQP(arr,weights)
!
!   Returns:                         mean of a set of samples
!
!   Arguments:
!     arr                            samples
!     weights                        weights to be applied to the samples
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(qp) :: &
     arr(:), &
     arrMeanQP, &
     weights(:)
    optional :: &
     weights
    intent(in) :: &
     arr, &
     weights

    if (present(weights)) then
      arrMeanQP=sum(weights*arr)/sum(weights)
    else
      arrMeanQP=sum(arr)/size(arr)
    end if

  end function arrMeanQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrVarReal(arr,meanArg,weights,mnKnown)
!
!   Returns:                         variance of a set of samples
!
!   Arguments:
!     arr                            samples
!     meanArg                        sample (or population) mean
!     weights                        weights to be applied to the samples
!     mnKnown                        whether the mean was known before the samples had been obtained
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     arr(:), &
     arrVarReal, &
     mean, &
     meanArg, &
     weights(:)
    logical :: &
     mnKnown
    optional :: &
     meanArg, &
     weights, &
     mnKnown
    intent(in) :: &
     arr, &
     meanArg, &
     weights

    if (present(meanArg)) then
      mean=meanArg
    else if (present(weights)) then
      mean=arrMean(arr,weights)
    else
      mean=arrMean(arr)
    end if

    if (present(weights)) then
      arrVarReal=real(sum(real(weights,dp)*(arr-mean)**2)/sum(weights),sp)
    else
      arrVarReal=real(sum(real(arr-mean,dp)**2)/(size(arr)-merge(0,1,present(mnKnown) .AND. mnKnown)),sp)
    end if

  end function arrVarReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrVarDbl(arr,meanArg,weights,mnKnown)
!
!   Returns:                         variance of a set of samples
!
!   Arguments:
!     arr                            samples
!     meanArg                        sample (or population) mean
!     weights                        weights to be applied to the samples
!     mnKnown                        whether the mean was known before the samples had been obtained
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     arr(:), &
     arrVarDbl, &
     mean, &
     meanArg, &
     weights(:)
    logical :: &
     mnKnown
    optional :: &
     meanArg, &
     weights, &
     mnKnown
    intent(in) :: &
     arr, &
     meanArg, &
     weights

    if (present(meanArg)) then
      mean=meanArg
    else if (present(weights)) then
      mean=arrMean(arr,weights)
    else
      mean=arrMean(arr)
    end if

    if (present(weights)) then
      arrVarDbl=real(sum(real(weights,qp)*(arr-mean)**2)/sum(weights),dp)
    else
      arrVarDbl=real(sum(real(arr-mean,qp)**2)/(size(arr)-merge(0,1,present(mnKnown) .AND. mnKnown)),dp)
    end if

  end function arrVarDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrVarQP(arr,meanArg,weights,mnKnown)
!
!   Returns:                         variance of a set of samples
!
!   Arguments:
!     arr                            samples
!     meanArg                        sample (or population) mean
!     weights                        weights to be applied to the samples
!     mnKnown                        whether the mean was known before the samples had been obtained
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(qp) :: &
     arr(:), &
     arrVarQP, &
     mean, &
     meanArg, &
     weights(:)
    logical :: &
     mnKnown
    optional :: &
     meanArg, &
     weights, &
     mnKnown
    intent(in) :: &
     arr, &
     meanArg, &
     weights

    if (present(meanArg)) then
      mean=meanArg
    else if (present(weights)) then
      mean=arrMean(arr,weights)
    else
      mean=arrMean(arr)
    end if

    if (present(weights)) then
      arrVarQP=sum(weights*(arr-mean)**2)/sum(weights)
    else
      arrVarQP=sum((arr-mean)**2)/(size(arr)-merge(0,1,present(mnKnown) .AND. mnKnown))
    end if

  end function arrVarQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrCovReal(arr1,arr2,mean1Arg,mean2Arg)
!
!   Returns:                         covariance of two sets of samples
!
!   Arguments:
!     arr1                           first set of samples
!     arr2                           second set of samples
!     mean1Arg                       sample (or population) mean of first set
!     mean2Arg                       sample (or population) mean of second set
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     arr1(:), &
     arr2(:), &
     arrCovReal, &
     mean1, &
     mean2, &
     mean1Arg, &
     mean2Arg
    optional :: &
     mean1Arg, &
     mean2Arg
    intent(in) :: &
     arr1, &
     arr2, &
     mean1Arg, &
     mean2Arg

    if (present(mean1Arg)) then
      mean1=mean1Arg
    else
      mean1=arrMean(arr1)
    end if
    if (present(mean2Arg)) then
      mean2=mean2Arg
    else
      mean2=arrMean(arr2)
    end if

    arrCovReal=sum(real(arr1-mean1,dp)*real(arr2-mean2,dp))/(size(arr1)-1)

  end function arrCovReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrCovDbl(arr1,arr2,mean1Arg,mean2Arg)
!
!   Returns:                         covariance of two sets of samples
!
!   Arguments:
!     arr1                           first set of samples
!     arr2                           second set of samples
!     mean1Arg                       sample (or population) mean of first set
!     mean2Arg                       sample (or population) mean of second set
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     arr1(:), &
     arr2(:), &
     arrCovDbl, &
     mean1, &
     mean2, &
     mean1Arg, &
     mean2Arg
    optional :: &
     mean1Arg, &
     mean2Arg
    intent(in) :: &
     arr1, &
     arr2, &
     mean1Arg, &
     mean2Arg

    if (present(mean1Arg)) then
      mean1=mean1Arg
    else
      mean1=arrMean(arr1)
    end if
    if (present(mean2Arg)) then
      mean2=mean2Arg
    else
      mean2=arrMean(arr2)
    end if

    arrCovDbl=sum(real(arr1-mean1,qp)*real(arr2-mean2,qp))/(size(arr1)-1)

  end function arrCovDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrVolDbl(arr)
!
!   Returns:                         volume of smallest area confined by coordinate planes containing all given points
!
!   Arguments:
!     arr                            set of points
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    real(dp) :: &
     arr(:,:), &
     arrVolDbl
    intent(in) :: &
     arr
    
    arrVolDbl=product([(maxval(arr(:,i))-minval(arr(:,i)), i=1,size(arr,2))])
    
  end function arrVolDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function arrVolQP(arr)
!
!   Returns:                         volume of smallest area confined by coordinate planes containing all given points
!
!   Arguments:
!     arr                            set of points
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    real(qp) :: &
     arr(:,:), &
     arrVolQP
    intent(in) :: &
     arr
    
    arrVolQP=product([(maxval(arr(:,i))-minval(arr(:,i)), i=1,size(arr,2))])
    
  end function arrVolQP




end module
