module sampling                       ! MCMC sampling, PT, Multi-PT, reading and saving samples

! Routines:
! 
! makeChain              Builds an MCMC chain
! accProbSwap            Returns: acceptance probability for parallel tempering swap (may exceed 1)
! multiPT                Builds several PT cold chains to be mixed, and assesses convergence
! calcPSR                Calculates potential scale reductions (PSR) for all pars. in all chains
! printPSR               Prints PSR values
! calcAutocorr           Calculates autocorrelation of parameter samples
! postrSampIO            Reads or saves posterior samples and other information to a saved-samples file
! calcLikh               Calculates likelihood for all chain links
! printPostrSampProps    Prints properties of posterior samples

  use basicTypes
  use crc32
  use random
  use dataModelsNoise
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !=====================================================[ TYPE DEFINITIONS ]========================================================

  type tyMCMCChain
    integer(i8b) :: &
     length, &
     burnInLen
    logical :: &
     sorted(N_L_PARS_MAX)=.FALSE.
    real(dp),allocatable :: &
     links(:,:)                      !(iteration, parameter)
    real(qp),allocatable :: &
     logLikh(:)
  end type


  ! SAVING POSTERIOR SAMPLES

  ! CAUTION: when changing components, make sure that SAV_SAMP_…_LEN and curHdrSiz (in savPostrSamp()) are changed accordingly and
  !          ensure backwards compatibility!

  type tySavSampMetaGlob             ! global header

      ! 1st part (constant length: SAV_SAMP_HDR1_LEN)

    character(22) ::  hIdent                    ! "BASE posterior samples"
    integer(i2b) ::   hRevNr(1)                 ! BASE revision nr.
    character(128) :: hResrvd                   ! reserved, currently unused
    integer(i2b) ::   hLenDiv(3)                ! lengths of command line, host computer name, target system identifier
    integer(i2b) ::   hNConfFilLins(1)          ! # config-file lines
    integer(i4b) :: &
                      hDtStart(8), &            ! date and time BASE was started
                      hDtEnd(8)                 ! date and time BASE was finished
    integer(i2b) ::   hNDtSets(1)               ! # data sets in total
    real(dp) :: &
                      hT1(1), &                 ! earliest time of all data
                      hTN(1), &                 ! latest time of all data
                      hEpoch(1)                 ! epoch
    integer(i1b) ::   hTimesAreJD(1)            ! times given in JD scale?
    integer(i4b) ::   &
                      hNDt(1), &                ! # data in total
                      hDtCRC(1)                 ! CRC of all data
    real(dp) ::       hMS(1)                    ! stellar mass
    integer(i1b) ::   hNNPlMds(1)               ! # n-planet modes (=1 if bin. mode)
    integer(i4b) :: &
                      hNPT(1), &                ! # PT procs. in multi-PT
                      hPTNChains(1)             ! # chains in PT

      ! 2nd part (variable length)

    integer(i1b),allocatable :: &
                      hDtTy(:)                  ! data types for all data sets
    character,allocatable :: &
                      hCmdLin(:), &             ! command line
                      hHostName(:), &           ! host computer name
                      hSysNm(:)                 ! target system identifier
    integer(i2b),allocatable :: &
                      hConfFilLinLen(:)         ! conf.-file line lengths
    character,allocatable :: &
                      hConfFilLins(:)           ! conf.-file lines (1-dim. array containing all chars. of conf. file)
  end type


  type tySavSampMetaNPlMd            ! n-planet-mode header

      ! 1st part (constant length: SAV_SAMP_HDR_MD_1_LEN)

    integer(i2b) :: &
                      hNPl(1), &                ! # planets
                      hNPars(1)                 ! # local pars.
    integer(i8b) :: &
                      hNSamp(1), &              ! # samples per par.
                      hNSampBurn(1), &          ! # samples per par. used as burn-in
                      hNThin(1), &              ! thinning stride
                      hMeanPTBlkLen(1)          ! mean PT block length
    integer(i1b) :: &
                      hTParAdj(1), &            ! tempering pars. adjusted?
                      hFinByUsrReq(1), &        ! finished by user request?
                      hComprMd(1)               ! compression mode (0: none; 1: RLE)

      ! 2nd part (variable length)

    real(dp),allocatable :: &
                      hTPar(:,:)                ! tempering pars. for all PTs (1st index) and all PT chains (2nd index)
    integer(i2b),allocatable :: &
                      hLenParNms(:)            ! lengths of local par. names
    character,allocatable :: &
                      hParNms(:)               ! par. names
    integer(i1b),allocatable :: &
                      hPriTypes(:), &          ! prior types
                      hPriLogScl(:)            ! user-def. prior: log scale?
    integer(i4b),allocatable :: &
                      hPriNPts(:)              ! user-def. prior: # points
    real(dp),allocatable :: &
                      hPriLBnd(:), &           ! lower prior bounds
                      hPriUBnd(:), &           ! upper prior bounds
                      hPriKnees(:), &          ! prior knees (value ignored for all priors except those of modified Jeffreys type)
                      hPriVce(:), &            ! variances of (truncated) normal distributions of normal priors
                      hPriPtsPsdX(:,:), &      ! user-def. prior points: (pseudo-)abscissae values
                      hPriPtsY(:,:), &         ! user-def. prior points: ordinate values
                      hPriInvDelta(:), &       ! user-def. prior: inverse delta
                      hFinPSR(:), &            ! final PSR values
                      hFracIndep(:), &         ! fraction of independent samples
                      hParEstMode(:)           ! posterior mode
  end type

  type tySavSampMetaSamp             ! meta info on samples
    integer(i8b),allocatable :: &
                      sNSeq(:)                 ! # sequences
    integer(i1b) ::   sSeqSiz(1)               ! current sequence size
    real(dp) ::       sSeqVal(1)               ! current value
  end type

  type tySavSamp                     ! complete info (for last #-pl. mode, excluding chain itself)
    character(MAX_LEN_PATH) :: &
                                filNmRead, &
                                filNmSav
    type(tySavSampMetaGlob) ::  hdr
    type(tySavSampMetaNPlMd) :: nPlMd ! contains only last #-pl. mode
    type(tySavSampMetaSamp) ::  samp
  end type


  !========================================================[ CONSTANTS ]============================================================

  integer,parameter :: &
   SAV_SAMP_MAX_SEQ_SIZ=HUGE_INT1, &
   N_SAMP_ZERO_MAX=10000, &
   TPAR_ADJ_MIN_N_UPDOWN=2000, &    ! cf. [05.05.2010/5]; should be <~2000 (13.05.2010) - but >~ 1000, s.t. qual. of 1st candidate is good (05.12.2010)
   TPAR_ADJ_N_UPDOWN_ZERO_MAX=0, &
   TPAR_ADJ_N_REVERSIONS_MAX=2, &   ! # reversions to the last optimal temp. par. after which still to continue trying to adjust temp. par.
   UTH_INFC_LEN_MIN=1000, &
   TH_INFC_LEN_MIN=1000, &
   AUTOCORR_N_LAGS_MAX=100

  integer(i8b),parameter :: &
   TPAR_ADJ_BLK_LEN=15_i8b, &       ! should be >~10 (13.05.2010)
   PSR_CALC_N_LINKS_MIN=1000_i8b, &
   UTH_DISCARD_LEN_MIN=1000_i8b, &
   AUTOCORR_LAG_MIN=1

  real,parameter :: &
   MULTIPT_BLK_DUR_S=5., &          ! duration of one multiPT block in seconds (important for display updates)
   INFC_LEN_FRAC_STD=.5, &
   PSR_THRLD1_2_RATIO_MIN=2., &
   AUTOCORR_LAG_FACT=1.05

  real(dp),parameter :: &
   TPAR_MAX=1._dp, &
   STD_TPAR_MIN=1e-3_dp, &          ! 1e-8 will lead to very poor performance even with temp. par. ajustment! (08.05.2010)
   TPAR_ADJ_PT_ACC_RATE_MIN=.15_dp, & ! should be ~.2 (13.05.2010)
   TPAR_ADJ_WEIGHT_OLD=.9_dp, &
   TPAR_ADJ_MIN_N_UPDOWN_FACT=1.1_dp, & ! should be ~ between 1.5 and 2 (13.05.2010) - may be ~ 1.1  (05.12.2010)
   SPD_MEAS_TPAR_ADJ_DUR_RATIO=1._dp/50, &
   TPAR_ADJ_RUNTM_RATIO=1._dp/10, &  ! ratio (duration of temp. par. adjustment):(remaining runtime)
   TPAR_ADJ_LOCK_RATIO_MAX=1._dp/5, &
   RUNTM_BEF_TPAR_ADJ_MAX=5._dp*60   ! CAUTION: when changing this value, also change corresponding help text

  type(tyIntvlDbl),parameter :: &
   TPAR_ADJ_DUR_S_PER_PT_RGE=tyIntvlDbl(10._dp,10._dp*60), & ! min./max. duration of temp. par. adjustment per PT proc.
   INI_SPD_MEAS_DUR_S_RGE=tyIntvlDbl(4._dp,10._dp), & ! min./max. duration of speed measurement per PT proc. for temp. par. adjustment
   PSR_UPD_INTVL_RATIO_RGE=tyIntvlDbl(1._dp/100,1._dp/5), & ! min./max. ratio (# blocks to next PSR update):(tot. # blocks)
   PSR_UPD_INTVL_DUR_S_RGE=tyIntvlDbl(20._dp,10._dp*60), &
   PSR_UPD_INTVL_RATIO_EL_RGE=tyIntvlDbl(1._dp/20,VAL_UNDEF_DBL)   ! min. ratio (time to next PSR update):(elapsed runtime in current mode)

  type(tyIntvlInt),parameter :: &
   PSR_UPD_DIST_RGE=tyIntvlInt(1000,500000)


  !========================================================[ VARIABLES ]============================================================

   integer :: &
    usrMaxRuntmS, &
    usrMaxLen, &
    curUsrMaxRuntmS, &
    curUsrMaxLen, &
    usrMaxRuntmInMdS(N_N_PL_MODES_MAX), &
    usrMaxLenInMd(N_N_PL_MODES_MAX), &
    usrSpdMeasDurS(N_N_PL_MODES_MAX), &
    usrTParAdjDurS(N_N_PL_MODES_MAX), &
    nPT, &                           ! # PT procedures to be run per mode
    PTNChains, &                     ! # chains run in parallel in the 'prlTempering' procedure
    tParAdjRemS(:), &
    nPSR, &
    usrInfcLenMin(N_N_PL_MODES_MAX), &
    autoCorrNLags(N_N_PL_MODES_MAX)

   integer(i8b) :: &
    meanPTBlkLen(N_N_PL_MODES_MAX), &
    PSRIBlks(:), &
    nThin(N_N_PL_MODES_MAX), &       ! stepping for thinning (1: no thinning)
    autoCorrLags(AUTOCORR_N_LAGS_MAX)

   real :: &
    tParSumOfSq(:), &
    maxRuntmRatioDiscard(N_N_PL_MODES_MAX), & ! max. fraction of runtime to be discarded before start of burn-in
                                              ! ('PSR_THRLD1_CALC_N_BLKS' blocks will be discarded anyway unless
                                              ! 'PSRThrld2' has already been fallen below earlier)
    aftrBurnRatioConv, &
    aftrBurnRatioNotConv, &
    infcLenFrac(N_N_PL_MODES_MAX)

   real(dp) :: &
    tPar(:,:), &
    tParMin, &
    tParMeanSqDevnMax, &
    finPSR(:), &                     ! stores the potential scale reduction (PSR) values
    PSRThrld1Fact(N_N_PL_MODES_MAX), &
    PSRThrld2, &
    fracIndep(N_L_PARS_MAX), &
    fracIndepMin, &
    autoCorr(N_N_PL_MODES_MAX,N_L_PARS_MAX,AUTOCORR_N_LAGS_MAX)

   logical :: &
    usrAdjustTPar(N_N_PL_MODES_MAX), &
    onlySpdMeas, &
    usrReadSamp, &
    usrSavSamp, &
    usrSavCandCSV, &
    dispPSR, &
    maximiseInfcLen, &
    finishedByUsrReq

   type(tyMCMCChain) :: &
    chn                              ! the final cold chain for the current mode

   type(tySavSamp) :: &
    savSamp

   type(tyCoord2DDbl) :: &
    PSR(:,:)

   allocatable :: &
    tParAdjRemS, &
    tPar, &
    tParSumOfSq, &
    PSRIBlks, &
    finPSR, &
    PSR


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


  !===========================================================[ MCMC ]==============================================================


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine makeChain(chain,iniPars,tPar,logPriLikhMax,iLogPriLikhMax,nAcc,unitSavCand,nThin)
!
!   builds an MCMC chain
!
!   Arguments:
!     chain                          MCMC chain
!     iniPars                        initial pars.
!     tPar                           tempering pars.
!     logPriLikhMax                  max. natural logarithm of prior×likelihood
!     iLogPriLikhMax                 index of sample with max. prior×likelihood
!     nAcc                           # accepted proposals
!     unitSavCand                    file unit for saving proposed candidates
!     nThin                          thinning stride
!
!   Requirements:
!     - chain%length >= 0
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     unitSavCand, &
     p
    integer(i8b) :: &
     it, &
     itThin, &
     iLogPriLikhMax, &
     nAcc, &
     nThin
    real(dp) :: &
     iniPars(:), &
     cand(curNPars), &
     cur(curNPars), &
     l, &
     candTmp(curNPars), &
     lMin, &
     lMax, &
     lTmp1, &
     lTmp2, &
     d(curNPars), &
     accProbTmpng, &
     logLikhCur, &
     logLikhCand, &
     logPriCur, &
     logPriCand, &
     logPriLikhMax, &
     logPriLikh, &
     tPar, &
     dCur
    logical :: &
     savCand
    type(tyMCMCChain) :: &
     chain
    intent(in) :: &
     iniPars, &
     tPar, &
     unitSavCand, &
     nThin
    intent(inout) :: &
     chain
    intent(out) :: &
     logPriLikhMax, &
     iLogPriLikhMax, &
     nAcc

    
    savCand=isDef(unitSavCand)


    if (chain%length > 0) then


      cur=iniPars
      logPriLikh=logPrior(cur)+logLikh(cur)
      logPriLikhMax=logPriLikh
      iLogPriLikhMax=1
      nAcc=0



      ! BUILD CHAIN

      do it=1,chain%length

        do itThin=1,nThin              ! loop for thinning

          do                           ! loop to find Hit & Run candidate

            ! generate random unit vector 'd', with distribution independent of direction

            call fillRndNormNonZero(d)
            d=d/sqrt(sum(d**2))


            ! find interval (lMin,lMax) in which jump length must lie

            lMin=-HUGE_DP
            lMax=HUGE_DP
            do p=1,curNPars
              if (parFixed(p)) cycle
              if (parIsCyc(p) .AND. cycParUsMaxRge(p)) then
                d(p)=d(p)*parPeriod(p)
                lMin=max(lMin,-1._dp)
                lMax=min(lMax,1._dp)
              else
                dCur=d(p)*priHalfW(p)
                d(p)=dCur
                lTmp1=(priLBnd(p)-cur(p))/dCur
                lTmp2=(priUBnd(p)-cur(p))/dCur
                lMin=max(lMin,min(lTmp1,lTmp2))
                lMax=min(lMax,max(lTmp1,lTmp2))
              end if
            end do


            ! sample l from this interval

            l=lMin+rndUni()*(lMax-lMin)


            ! calc. candidate link

            where (parFixed(:curNPars))
              candTmp=cur
            elsewhere
              candTmp=cur+l*d
            end where
            cand=candTmp-floor(candTmp*parInvPeriod(:curNPars))*parPeriod(:curNPars)


            ! check that candidate is within prior support

            if (priorIsNonZero(cand)) exit ! condition not met in fraction of <~ 5E-6 of tests: OK


          end do                       ! loop to find Hit & Run candidate


          ! output candidate
          
          if (savCand) then
            !$OMP critical (OMPCrit5)
            write(unitSavCand,*) cand
            !$OMP end critical (OMPCrit5)
          end if


          ! calc. acc. prob.

          logLikhCur=logLikh(cur)
          logLikhCand=logLikh(cand)
          logPriCur=logPrior(cur)
          logPriCand=logPrior(cand)
          accProbTmpng=exp(logPriCand-logPriCur+tPar*(logLikhCand-logLikhCur))


          ! accept/reject; determination of posterior mode

          if (rndUni() <= accProbTmpng) then
            cur=cand
            logLikhCur=logLikhCand
            logPriLikh=logPriCand+logLikhCand
            if (logPriLikh > logPriLikhMax) then
              logPriLikhMax=logPriLikh
              iLogPriLikhMax=it
            end if
            call inc(nAcc)
          end if

        end do                       ! loop for thinning

        chain%links(it,1:curNPars)=cur
        chain%logLikh(it)=real(logLikhCur,qp)

      end do                         ! it=1,chain%length
    
    end if                           ! (chain%length > 0)

  end subroutine makeChain




  !====================================================[ PARALLEL TEMPERING ]=======================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine prlTempering(coldChain,meanBlkLen,nChains,iniParsArg,presrvIniPars,tPar,adjTPar,logPriLikhMax,logPriLikhMaxPars, &
     nAcc,savCandCSV,sumOfSqOpt,iPT)
!
!   builds several MCMC chains using parallel tempering
!
!   Arguments:
!     coldChain                      cold MCMC chain
!     meanBlkLen                     mean PT block length
!     nChains                        # PT chains
!     iniParsArg                     initial pars.
!     presrvIniPars                  whether iniParsArg should remain unaltered
!     tPar                           tempering parameters
!     adjTPar                        whether tempering pars. should be adjusted
!     logPriLikhMax                  max. natural logarithm of prior×likelihood
!     logPriLikhMaxPars              pars. maximising prior×likelihood (i.e. MAP pars.)
!     nAcc                           # accepted proposals for cold chain
!     savCandCSV                     whether candidates should be saved in a CSV file
!     sumOfSqOpt                     best achieved sum of squared deviations of up-drifting fraction from straight line in
!                                    temp.-pars. adjustment
!     iPT                            index of this PT procedure in multiPT
!
!   CAUTION: not thread-safe!
!   iniParsArg MAY NOT BE CONTIGUOUS
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     LBL_UP=1, &
     LBL_DOWN=2
    integer :: &
     iChn, &
     nChains, &                      ! # chains run in parallel
     lbls(nChains), &
     nTParRev, &
     allocStat, &
     iPT
    integer(i8b) :: &
     iBlk, &
     i, &
     j, &
     meanBlkLen, &
     nBlks, &
     maxBlkLen, &
     cChnPos, &
     nLinks, &
     curBlkLen, &
     curNThin, &
     blkFin(:), &
     i1MinDist, &
     iBlk1CurAdjCyc, &
     curNUpDown, &
     nUpDownZero, &
     minNUpDownReq, &
     minNUpDown, &
     nSwapsTriedByChain(nChains-1), &
     nSwapsByChain(nChains-1), &
     nUpDown(nChains,2), &
     nUpDownZeroMax, &
     iChgLockEarliestRetry, &
     iChgLockMinNChg, &
     iLogPriLikhMax(nChains), &
     nAccTmp, &
     nAcc
    real :: &
     fracUpDown(nChains), &
     fAim, &
     sumOfSq, &
     sumOfSqOpt, &
     minDevn, &
     straightLinDev
    real(dp) :: &
     iniParsArg(:,:), &
     tPar(:), &
     logPriLikhMaxPars(:), &
     tmpBlkFin(:), &
     tmStart, &
     tmLstDispUpd, &
     tParOpt(nChains), &
     tParCand(nChains), &
     logARByChain(nChains-1), &
     logARByChainNew(nChains-1), &
     logARMin, &
     logARNewMin, &
     meanDevn, &
     logPriLikhMax, &
     logPriLikhMaxArr(nChains), &
     iniPars(nChains,curNPars), &
     rndTmp, &
     rndTmp2, &
     accRate
    logical :: &
     adjTPar, &
     optEqCurTPar, &
     presrvIniPars, &
     remDurPrinted, &
     savCandCSV
    type(tyMCMCChain) :: &
     coldChain, &
     chains(nChains)
    allocatable :: &
     blkFin, &
     tmpBlkFin
    optional :: &
     sumOfSqOpt, &
     iPT
    intent(in) :: &
     nChains, &
     meanBlkLen, &
     adjTPar, &
     presrvIniPars, &
     savCandCSV
    intent(inout) :: &
     coldChain, &
     iniParsArg, &
     tPar
    intent(out) :: &
     logPriLikhMax, &
     logPriLikhMaxPars, &
     nAcc


    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
    !
    !   PREPARATIONS
    !
    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·

    maxBlkLen=0_i8b
    cChnPos=1_i8b
    nTParRev=0
    iChgLockMinNChg=0_i8b
    curNThin=merge(1_i8b,nThin(iNPlMd),isUndef(nThin(iNPlMd))) ! if auto-thinning is activated, nThin(iNPlMd) will be undefined
    nLinks=coldChain%length
    nAcc=0

    if (nLinks > 0) then

      if (adjTPar) then
        call resetQties(.TRUE.)
        tParOpt=tPar
        call undef(sumOfSqOpt)
        minNUpDownReq=TPAR_ADJ_MIN_N_UPDOWN
        tmLstDispUpd=omp_get_wtime()
        remDurPrinted=.FALSE.
        nUpDownZeroMax=TPAR_ADJ_N_UPDOWN_ZERO_MAX
      end if

      iniPars=iniParsArg
      nBlks=max(1,floor(1.*nLinks/meanBlkLen))

      allocate(tmpBlkFin(nBlks),blkFin(0:nBlks),stat=allocStat)
      call testAllocErr(allocStat,1)

      coldChain%burnInLen=-1
      chains(1:nChains)%burnInLen=-1

      blkFin(0)=0
      blkFin(nBlks)=nLinks
      do i=1,nBlks                   ! find out max. block size and stretch tmpBlkFin s.t. blkFin(nBlks)==nLinks
        blkFin(i)=floor(1.*nLinks*i/nBlks)
        if (blkFin(i)-blkFin(i-1) > maxBlkLen) maxBlkLen=blkFin(i)-blkFin(i-1)
      end do


      do i=1,nChains
        allocate(chains(i)%links(maxBlkLen,curNPars),chains(i)%logLikh(maxBlkLen),stat=allocStat)
        call testAllocErr(allocStat,2)
      end do

      logPriLikhMax=-huge(logPriLikhMax)


    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
    !
    !   BLOCKS LOOP
    !
    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·

      tmStart=omp_get_wtime()
      iBlk1CurAdjCyc=1

      ! execute all blocks
      do iBlk=1,nBlks

        curBlkLen=blkFin(iBlk)-cChnPos+1


    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
    !
    !   CHAINS LOOP
    !
    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·

        ! build all parallel chains

        !$OMP parallel do schedule(dynamic) private(nAccTmp)
        do iChn=1,nChains

          chains(iChn)%length=curBlkLen

          call makeChain(chains(iChn),iniPars(iChn,:),tPar(iChn),logPriLikhMaxArr(iChn),iLogPriLikhMax(iChn),nAccTmp, &
           merge(UN_SAV_CAND_CSV+iChn,VAL_UNDEF_INT,savCandCSV),curNThin)
          
          if (iChn == 1) nAcc=nAcc+nAccTmp ! cold chain
          if (.NOT. adjTPar) iniPars(iChn,:)=chains(iChn)%links(curBlkLen,:) ! for next block; activating this for tPar adjustment will strongly decrease its efficiency! (why?) (05.12.2010)

        end do                       ! CHAINS LOOP
        !$OMP end parallel do


        ! determination of posterior mode

        do iChn=1,nChains
          if (logPriLikhMaxArr(iChn) > logPriLikhMax) then
            logPriLikhMax=logPriLikhMaxArr(iChn)
            logPriLikhMaxPars=chains(iChn)%links(iLogPriLikhMax(iChn),:)
          end if
        end do


    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
    !
    !   SWAP ?
    !
    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·

        if (nChains > 1) then

          ! propose swap of last links of chains (i),(i+1)
          call random_number(rndTmp)
          call random_number(rndTmp2)
          i=1+floor(rndTmp*(nChains-1))

          call inc(nSwapsTriedByChain(i))
          if (rndTmp2 < accProbSwap(chains(i)%links(curBlkLen,:),chains(i+1)%links(curBlkLen,:),tPar(i),tPar(i+1))) then

            call inc(nSwapsByChain(i))

            ! swap links

            call swap(chains(i)%links(curBlkLen,:),chains(i+1)%links(curBlkLen,:))

            ! swap chain initial pars.

            call swap(iniPars(i,:),iniPars(i+1,:))

            if (adjTPar) then

              ! swap up/down labels

              call swap(lbls(i),lbls(i+1))

              ! make sure up/down labels for cold and hottest chains are correct

              lbls(1)=LBL_UP
              lbls(nChains)=LBL_DOWN

              ! update histogram of up/down labels

              do i=1,nChains
                nUpDown(i,lbls(i))=nUpDown(i,lbls(i))+ceiling(lbls(i)/2.)
              end do

            end if

          end if                     !(swap accepted)

        end if                       !(nChains > 1)



    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
    !
    !   ADJUST TEMP. PARS.
    !
    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·


        if (adjTPar) then


          ! REMAINING TIME

          if (omp_get_wtime()-tmLstDispUpd >= 1.) then
            call printQty(UNIT%cons,"Estimated time remaining [h:m:s]", &
             fmtTime(nint(((omp_get_wtime()-tmStart)/iBlk)*((nPT-(iPT-1))*nBlks-iBlk))),OPT_NO_ADVANCE)
            tmLstDispUpd=omp_get_wtime()
            remDurPrinted=.TRUE.
          end if


          ! CHECK THAT MINIMUM NUMBER OF REGISTERED UP/DOWN LABELS (WHERE > 0) IS GREATER THAN A THRESHOLD

          minNUpDown=huge(minNUpDown)
          nUpDownZero=0
          do i=1,nChains
            curNUpDown=sum(nUpDown(i,:))
            if (curNUpDown == 0) call inc(nUpDownZero)
            if (0 < curNUpDown .AND. curNUpDown < minNUpDown) minNUpDown=curNUpDown
          end do
          if (minNUpDown == huge(minNUpDown)) minNUpDown=0

          iChgLockEarliestRetry=max(0_i8b,iChgLockEarliestRetry-1)
          iChgLockMinNChg=max(0_i8b,iChgLockMinNChg-1)

          if ((nUpDownZero <= nUpDownZeroMax .AND. iChgLockEarliestRetry == 0 .AND. minNUpDown > minNUpDownReq) .OR. &
           iChgLockMinNChg == 0) then

            iChgLockMinNChg=nint(TPAR_ADJ_LOCK_RATIO_MAX*nBlks)


            ! UP/DOWN FRACTION

            do i=1,nChains
              if (sum(nUpDown(i,:)) == 0) then
                ! if fraction undefined: define it as 1 or 0 if chain is nearest to cold or hottest chain, respectively
                fracUpDown(i)=merge(1.,0.,i < (1+nChains)/2.)
              else
                fracUpDown(i)=1.*nUpDown(i,LBL_UP)/(nUpDown(i,LBL_UP)+nUpDown(i,LBL_DOWN))
              end if
            end do


            ! QUALITY OF CURRENT TPAR

            sumOfSq=0
            do i=2,nChains-1
              fAim=real(nChains-i)/(nChains-1)
              sumOfSq=sumOfSq+(fracUpDown(i)-fAim)**2
            end do

            if (sumOfSq < sumOfSqOpt .OR. isUndef(sumOfSqOpt)) then
              sumOfSqOpt=sumOfSq
              tParOpt=tPar
              optEqCurTPar=.TRUE.
            end if

            if (optEqCurTPar) then


              ! GENERATE NEW TPAR

                ! generate values and sort them (sorting should not be necessary!)

              tParCand(1)=tPar(1)
              tParCand(nChains)=tParMin
              do i=2,nChains-1
                minDevn=HUGE_SP
                fAim=real(nChains-i)/(nChains-1)
                do j=1,nChains-1
                  if ((fracUpDown(j)-fAim)*(fracUpDown(j+1)-fAim) <= 0) then
                    straightLinDev=sqrt((fracUpDown(j)-real(nChains-j)/(nChains-1))**2+ &
                     (fracUpDown(j+1)-real(nChains-(j+1))/(nChains-1))**2)
                    if (straightLinDev < minDevn) then
                      minDevn=straightLinDev
                      i1MinDist=j
                    end if
                  end if
                end do
                tParCand(i)=tPar(i1MinDist)+(fAim-fracUpDown(i1MinDist))/(fracUpDown(i1MinDist+1)-fracUpDown(i1MinDist))* &
                (tPar(i1MinDist+1)-tPar(i1MinDist))
              end do
              call sort(tParCand,OPT_SORT_DESCENDING)


              ! derive tParCand as mixture with current tPar

              tParCand(2:nChains-1)=TPAR_ADJ_WEIGHT_OLD*tPar(2:nChains-1)+(1._dp-TPAR_ADJ_WEIGHT_OLD)*tParCand(2:nChains-1)

                ! calc. & check mean squared deviation of temp. pars. from original ones

              meanDevn=0
              do i=2,nChains-1
                meanDevn=meanDevn+(tParCand(i)-(TPAR_MAX-real(i-1,dp)/(nChains-1)*(TPAR_MAX-tParMin)))**2
              end do
              meanDevn=meanDevn/(nChains-2)

              if (meanDevn <= tParMeanSqDevnMax) then

                  ! estimate new acceptance rates and check if they are high enough

                ! old acceptance rates
                do i=1,nChains-1
                  if (nSwapsTriedByChain(i) == 0) then
                    logARByChain(i)=1 ! dummy
                  else
                    logARByChain(i)=log(1._dp*nSwapsByChain(i)/nSwapsTriedByChain(i))
                  end if
                end do
                logARMin=minval(logARByChain)

                ! new acceptance rates
                do i=1,nChains-1
                  logARByChainNew(i)=logARByChain(i)*((tParCand(i+1)-tParCand(i))/(tPar(i+1)-tPar(i)))**2
                end do
                logARNewMin=minval(logARByChainNew)

                ! accept candidate?
                if (logARNewMin >= min(logARMin,log(TPAR_ADJ_PT_ACC_RATE_MIN))) then
                  call resetQties
                  tPar=tParCand
                  optEqCurTPar=.FALSE.
                  if (iBlk+2*((iBlk+1)-iBlk1CurAdjCyc) <= nBlks) minNUpDownReq=nint(minNUpDownReq*TPAR_ADJ_MIN_N_UPDOWN_FACT)
                  nUpDownZeroMax=max(0_i8b,nUpDownZeroMax-1)
                  iBlk1CurAdjCyc=iBlk+1
                else
                  iChgLockEarliestRetry=minNUpDownReq ! TPAR_ADJ_N_CHG_LOCK
                end if

              else

                iChgLockEarliestRetry=minNUpDownReq ! TPAR_ADJ_N_CHG_LOCK

              end if                 ! (meanDevn <= tParMeanSqDevnMax)

            else                     ! i.e. (.NOT. optEqCurTPar)

              if (nTParRev == TPAR_ADJ_N_REVERSIONS_MAX) then

                exit

              else

                ! revert to last "optimal" temp. pars.
                call resetQties
                call inc(nTParRev)
                tPar=tParOpt
                optEqCurTPar=.TRUE.
                if (iBlk+2*((iBlk+1)-iBlk1CurAdjCyc) <= nBlks) then
                  minNUpDownReq=nint(minNUpDownReq*TPAR_ADJ_MIN_N_UPDOWN_FACT)
                end if
                iBlk1CurAdjCyc=iBlk+1

              end if

            end if                   ! (optEqCurTPar)

          end if                     ! (nUpDownZero <= nUpDownZeroMax .AND. iChgLockEarliestRetry == 0 .AND. (minNUpDown > minNUpDownReq … ))

        else

          ! append 1st chain to coldChain

          coldChain%links(cChnPos:cChnPos+curBlkLen-1,:)=chains(1)%links(:curBlkLen,:)
          coldChain%logLikh(cChnPos:cChnPos+curBlkLen-1)=chains(1)%logLikh(:curBlkLen)

        end if                       ! (adjTPar)


        cChnPos=cChnPos+curBlkLen

      end do                         ! BLOCKS LOOP

      if (adjTPar) tPar=tParOpt
      if (.NOT. presrvIniPars) iniParsArg=iniPars

      deallocate(tmpBlkFin,blkFin)
      do i=1,nChains
        deallocate(chains(i)%links,chains(i)%logLikh)
      end do
      coldChain%burnInLen=0

    end if                           ! (nLinks > 0)


    contains

    subroutine resetQties(noIniPars) ! resets certain vars. for temp. par. adjustment
      implicit none
      logical :: &
       noIniPars
      optional :: &
       noIniPars
      intent(in) :: &
       noIniPars

      nSwapsByChain=0
      nSwapsTriedByChain=0
      nUpDown=0
      lbls=0
      lbls(1)=LBL_UP
      lbls(nChains)=LBL_DOWN
      iniPars=iniParsArg
      iChgLockEarliestRetry=0
      if (.NOT. (present(noIniPars) .AND. noIniPars)) iniPars=iniParsArg
    end subroutine


  end subroutine prlTempering




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function accProbSwap(parI,parIP1,tParI,tParIP1)
!
!   Returns:                         acceptance probability for parallel tempering swap (may exceed 1)
!
!   Arguments:
!     parI                           link of chain i
!     parIP1                         link of chain i+1
!     tParI                          temp. pars. of chain i
!     tParIP1                        temp. pars. of chain i+1
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     parI(:), &
     parIP1(:), &
     accProbSwap, &
     tParI, &
     tParIP1
    intent(in) :: &
     parI, &
     parIP1, &
     tParI, &
     tParIP1

    accProbSwap=exp(tParI*(logLikh(parIP1)-logLikh(parI))+tParIP1*(logLikh(parI)-logLikh(parIP1)))
      ! cf. [dip.(2.94)]

  end function accProbSwap




  !=========================================================[ MULTI-PT ]============================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine multiPT(finCChn,postrMode,logPostrEv)
!
!   builds several PT cold chains to be mixed, and assesses convergence
!
!   Arguments:
!     finCChn                        resulting cold chain
!     postrMode                      posterior mode
!     logPostrEv                     natural logarithm of posterior×evidence
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iMaxPSR, &
     iChn, &
     i, &
     iPT, &
     p, &
     secRemain, &
     curMaxRuntmRemS, &
     allocStat, &
     tmpInt, &
     iPTMix, &
     tm8StartMultiPTCore
    integer(i8b) :: &
     oriBurnInLen, &
     oriInfcLen, &
     PSRUpdINxtBlk, &
     iBlkBurnInStart, &
     iBlkBurnInStartMax, &
     iBlk1PSR, &
     PSRUpdDistBlks, &
     desirFinCChnLen, &
     multiPTBlkLen, &
     iLink, &
     iBlk, &
     finCChnMaxLen, &
     snglCChnsMaxLen, &
     finCChnPos, &
     nBlks, &
     nAccTmp, &
     nCand, &
     nAcc
    real(dp) :: &
     postrMode(curNPars), &
     iniPars(nPT,PTNChains,curNPars), &
     PSRThrld1, &
     PSRNxtThrld, &
     maxPSR, &
     absLstPSRGrad, &
     curPSRGrad, &
     tParAdjDurS, &
     iniSpdMeasDurS, &
     secElapsMultiPTCore, &
     tParAdjActDurS, &
     tmStartMultiPTCore, &
     tmSpdMeasStart, &
     tmTmp, &
     tmBurnInStartMax, &
     tmLstDispUpd, &
     spdMultiPT, &
     logPriLikhMaxArr(nPT), &
     logPriLikhMaxPars(nPT,curNPars), &
     logPostrEv, &
     accRate
    logical :: &
     burnInStarted, &
     burnInFinished, &
     PSRCalcDeact, &
     warnShortenBurnIn, &
     overrideMaxRuntm
    character(11) :: &
     cMaxPSR
    character(9) :: &
     cCChnLinksPerS
    character(10) :: &
     cAccRate
    character(DISP_TBL_COL_W_MAX) :: &
     statTbl(DISP_TBL_N_COLS)
    type(tyIntvlInt8) :: &
     PSRUpdDistBlksRge
    type(tyMCMCChain) :: &
     finCChn, &
     coldChains1Blk(nPT), &
     coldChains(nPT)
    intent(out) :: &
     finCChn, &
     postrMode, &
     logPostrEv


    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
    !
    !   PREPARATIONS: DETERMINATION OF temp. pars., SPEED MEASUREMENT, TIME MANAGEMENT
    !
    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·

    iBlkBurnInStart=0_i8b
    iBlk1PSR=0_i8b
    iBlk=0_i8b
    tParAdjActDurS=0._dp
    call startSubsect(UNIT%cons+UNIT%log,"Starting Multi-PT procedure")

    ! INITIALIZATIONS

    finishedByUsrReq=.FALSE.
    burnInStarted=.FALSE.
    burnInFinished=.FALSE.
    finCChnPos=1
    dispPSR=.FALSE.
    call undef(PSRThrld1)
    warnShortenBurnIn=.FALSE.
    nPSR=0
    iMaxPSR=1
    call undef(maxPSR)
    call undef(oriInfcLen)
    overrideMaxRuntm=.FALSE.
    curSpdEst=UTH_MULTIPT_EST_SPD_MIN/merge(1_i8b,nThin(iNPlMd),isUndef(nThin(iNPlMd)))


    ! DURATION OF TEMP. PAR. ADJUSTMENT AND INITIAL SPEED MEASUREMENT

    if (onlySpdMeas) then
      iniSpdMeasDurS=usrSpdMeasDurS(iNPlMd)
    else
      if (usrAdjustTPar(iNPlMd)) then
        if (isUndef(usrTParAdjDurS(iNPlMd))) then
          if (isDef(curUsrMaxRuntmS)) then
            tParAdjDurS=putOntoIntvl(stretchIntvl(TPAR_ADJ_DUR_S_PER_PT_RGE,real(nPT,dp)),TPAR_ADJ_RUNTM_RATIO*curUsrMaxRuntmS)
          else
            tParAdjDurS=intvlCentre(TPAR_ADJ_DUR_S_PER_PT_RGE)*nPT
          end if
        else
          tParAdjDurS=1.*usrTParAdjDurS(iNPlMd)
        end if
        iniSpdMeasDurS=putOntoIntvl(INI_SPD_MEAS_DUR_S_RGE,round(SPD_MEAS_TPAR_ADJ_DUR_RATIO*tParAdjDurS,1._dp,ROUND_MD%nearest))
      else
        if (isUndef(usrSpdMeasDurS(iNPlMd))) then
          if (isDef(curUsrMaxRuntmS)) then
            ! obey bounds
            iniSpdMeasDurS=putOntoIntvl(SPD_MEAS_DUR_S_RGE,round(SPD_MEAS_RUNTM_RATIO*curUsrMaxRuntmS,1.,ROUND_MD%nearest))
          else
            iniSpdMeasDurS=nint(intvlCentre(SPD_MEAS_DUR_S_RGE))
          end if
        else
          iniSpdMeasDurS=usrSpdMeasDurS(iNPlMd)
        end if
      end if
    end if


    ! INITIAL PARS.

    call startSubsect(UNIT%cons+UNIT%log,"Generating initial parameters")
    do p=1,curNPars
      if (parFixed(p)) then
        iniPars(:,1,p)=priLBnd(p)
      else
        iniPars(:,1,p)=disprsOvrPriIntvl(p,nPT,sorted=.FALSE.)
      end if
      do iChn=2,PTNChains
        iniPars(:,iChn,p)=iniPars(:,1,p)
      end do
    end do

    call startSubsect(UNIT%cons+UNIT%log,"Results")
    if (nPT == 1) then
      do p=1,curNPars
        call printQty(UNIT%cons+UNIT%log,trim(fmtPar(p,.NOT. OPT_FOR_PLOT)),iniPars(1,1,p))
      end do
    else
      do p=1,curNPars
        call startSubsect(UNIT%cons+UNIT%log,trim(fmtPar(p,.NOT. OPT_FOR_PLOT)))
        if (parFixed(p)) then
          call printQty(UNIT%cons+UNIT%log,"All PT procedures",iniPars(1,1,p))
        else
          do iPT=1,nPT
            call printQty(UNIT%cons+UNIT%log,"PT procedure "//trim(niceFmt(iPT)),iniPars(iPT,1,p))
          end do
        end if
        call endSubsect(UNIT%cons+UNIT%log) ! trim(parNms(p))
      end do
    end if
    call endSubsect(UNIT%cons+UNIT%log) ! Results
    call endSubsect(UNIT%cons+UNIT%log) ! Initial parameters



    ! INITIAL SPEED MEASUREMENT

    tmSpdMeasStart=omp_get_wtime()
    call startSubsect(UNIT%cons+UNIT%log,"Initial speed measurement")
    call printQty(UNIT%cons+UNIT%log,"Estimated duration [s]",iniSpdMeasDurS)
    spdMultiPT=multiPTSpeed(nPT,iniPars,iniSpdMeasDurS)
    curSpdEst=spdMultiPT
    call printQty(UNIT%cons+UNIT%log,"Speed [1/s]",spdMultiPT)
    call endSubsect(UNIT%cons+UNIT%log) ! Initial speed measurement


    if (.NOT. onlySpdMeas) then

      ! AUTO-THINNING

      if (isUndef(nThin(iNPlMd))) then
        nThin(iNPlMd)=1
        do i=1,N_SPD_MEAS
          curMaxRuntmRemS=curUsrMaxRuntmS-nint(omp_get_wtime()-tmSpdMeasStart)
          tmpInt=max(1,ceiling(nThin(iNPlMd)*curMaxRuntmRemS*spdMultiPT/curUsrMaxLen))
          curSpdEst=curSpdEst*nThin(iNPlMd)/tmpInt
          nThin(iNPlMd)=tmpInt
          call printQty(UNIT%cons+UNIT%log,"Resulting stride for thinning",tmpInt)
          call startSubsect(UNIT%cons+UNIT%log,"Speed measurement no. "//trim(niceFmt(i+1)))
          call printQty(UNIT%cons+UNIT%log,"Estimated duration [s]",iniSpdMeasDurS)
          spdMultiPT=multiPTSpeed(nPT,iniPars,iniSpdMeasDurS)
          curSpdEst=spdMultiPT
          call printQty(UNIT%cons+UNIT%log,"Speed [1/s]",spdMultiPT)
          call endSubsect(UNIT%cons+UNIT%log) ! Speed measurement no.…
        end do
        if (nThin(iNPlMd) == 1) then
          call printMsg(UNIT%cons+UNIT%log,"Thinning automatically deactivated")
        end if
      end if


      ! DETERMINE MAX. RUNTIME/CHAIN LENGTH

      multiPTBlkLen=ceiling(MULTIPT_BLK_DUR_S*spdMultiPT)
      if (isDef(curUsrMaxRuntmS)) then
        curMaxRuntmRemS=curUsrMaxRuntmS-nint(omp_get_wtime()-tmSpdMeasStart)
        if (isDef(curUsrMaxLen)) then
          finCChnMaxLen=curUsrMaxLen
          multiPTBlkLen=min(multiPTBlkLen,min(finCChnMaxLen,floor(curMaxRuntmRemS*spdMultiPT,i8b))) ! obey bounds
        else
          finCChnMaxLen=max(0,floor(curMaxRuntmRemS*spdMultiPT))
          multiPTBlkLen=min(multiPTBlkLen,finCChnMaxLen)
        end if
      else
        finCChnMaxLen=curUsrMaxLen
        curMaxRuntmRemS=floor(finCChnMaxLen/spdMultiPT) ! estimated; will not be used to finish program
        multiPTBlkLen=min(multiPTBlkLen,finCChnMaxLen)
      end if
      tmBurnInStartMax=real(maxRuntmRatioDiscard(iNPlMd),dp)*curMaxRuntmRemS
      if (usrAdjustTPar(iNPlMd)) tmBurnInStartMax=min(tmBurnInStartMax,RUNTM_BEF_TPAR_ADJ_MAX)


      ! DETERMINATION OF POSTERIOR MODE

      logPostrEv=-huge(logPostrEv)
      call undef(postrMode)



      ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
      !
      !   START OF MULTI-PT PROCEDURE
      !
      ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·

      if (finCChnMaxLen > 0) then    ! the contrary might happen if max. # links has been determined from speed measurement

        nBlks=floor(1.*finCChnMaxLen/multiPTBlkLen)
        finCChnMaxLen=nBlks*multiPTBlkLen
        iBlkBurnInStartMax=max(1,floor(tmBurnInStartMax*spdMultiPT/multiPTBlkLen)) ! after this block, burn-in will be started at the latest in any event
        nBlks=nBlks+iBlkBurnInStartMax
        snglCChnsMaxLen=nBlks*multiPTBlkLen

        if (allocated(PSR)) deallocate(PSR)
        if (allocated(finPSR)) deallocate(finPSR)
        if (allocated(PSRIBlks)) deallocate(PSRIBlks)
        allocate(PSR(nBlks,curNPars),PSRIBlks(nBlks),finPSR(curNPars),stat=allocStat)
        call testAllocErr(allocStat,3)
        do p=1,curNPars
          if (parFixed(p)) then
            call undef(PSR(:,p)%x)
            call undef(PSR(:,p)%y)
            call undef(finPSR(p))
          end if
        end do

        do iPT=1,nPT
          allocate(coldChains(iPT)%links(snglCChnsMaxLen,curNPars),coldChains1Blk(iPT)%links(multiPTBlkLen,curNPars), &
           coldChains(iPT)%logLikh(snglCChnsMaxLen),coldChains1Blk(iPT)%logLikh(multiPTBlkLen),stat=allocStat)
          call testAllocErr(allocStat,4)
          coldChains(iPT)%length=snglCChnsMaxLen
          coldChains1Blk(iPT)%length=multiPTBlkLen
        end do

        call printQty(UNIT%cons+UNIT%log,"Maximum length of final chain",finCChnMaxLen)
        call startSubsect(UNIT%cons+UNIT%log,"Starting several PT procedures in parallel")
        call printQty(UNIT%cons+UNIT%log,"# PT procedures",nPT)

        if (PTNChains > 1) call printTPar

        tmStartMultiPTCore=omp_get_wtime()
        tm8StartMultiPTCore=time8()
        call printQty(UNIT%cons+UNIT%log,"Estimated latest time of burn-in start", &
         fmtDate(dateandtimeFromTime8(tm8StartMultiPTCore+nint(tmBurnInStartMax,i4b)),short=.TRUE.))

        ! status table header
        call startSubsect(UNIT%log,"Status")
        call startSubsect(UNIT%cons,"Status:")
        call printMsg(UNIT%cons,fmtStatTbl(DISP_TBL_HDR1,hdrArg=.TRUE.),setFlg(FLG%simpPrfx))
        call printMsg(UNIT%cons,fmtStatTbl(DISP_TBL_HDR2,hdrArg=.TRUE.),setFlg(FLG%simpPrfx))
        call printMsg(UNIT%cons,fmtStatTbl(hLinArg=.TRUE.),setFlg(FLG%simpPrfx))


        ! INITIALIZATIONS

        tmLstDispUpd=tmStartMultiPTCore
        PSRCalcDeact=nPT == 1
        if (PSRCalcDeact) then
          call undef(PSR%x)
          call undef(PSR%y)
          call undef(finPSR)
          PSRUpdINxtBlk=nBlks+1
        else
          PSRUpdINxtBlk=1
        end if
        call undef(PSRNxtThrld)
        if (usrSavCandCSV) then
          do i=1,PTNChains
            call tryOpenFilRW("candidate-samples",trim(outDir)//trim(runIdent)// &
             trim(plStr(iNPlMd,psdNPlUndef,"_",prepend=.TRUE.))//"_"//"sav-cand_chain-"//trim(niceFmt(i))//FIL_EXT_CSV, &
             UN_SAV_CAND_CSV+i,"replace","write")
          end do
        end if
        nCand=0
        nAcc=0


      ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
      !
      !   BLOCKS LOOP
      !
      ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·

        do                           ! blocks loop

          call inc(iBlk)

          iBlk1PSR=iBlkBurnInStart+floor((1.-infcLenFrac(iNPlMd))*(iBlk-iBlkBurnInStart))



          ! PT LOOP

          ! using nested parallelism since within prlTempering more teams will be forked
          ! reason for this implementation: team size within prlTempering (i.e. PTNChains) may be < #CPUs

          ! CAUTION: don't use OMP for loop calling prlTempering!
          !          (speed decreased to ~ (60±10)% when calling
          !          'base am.csv rv_tmp_pri.csv rv_tmp_sec.csv --only-speed-meas --speed-meas-dur=5s' (19.05.2010))

          do iPT=1,nPT

            ! PT procedure

            call prlTempering(coldChains1Blk(iPT),meanPTBlkLen(iNPlMd),PTNChains,iniPars(iPT,:,:),OPT_PT_CHG_INIPARS, &
             tPar(iPT,:),OPT_PT_NORMAL,logPriLikhMaxArr(iPT),logPriLikhMaxPars(iPT,:),nAccTmp,usrSavCandCSV) ! iniParsArg NOT CONTIGUOUS

            coldChains(iPT)%links((iBlk-1)*multiPTBlkLen+1:iBlk*multiPTBlkLen,:)=coldChains1Blk(iPT)%links
            nCand=nCand+nThin(iNPlMd)*multiPTBlkLen
            nAcc=nAcc+nAccTmp


            ! posterior mode

            if (logPriLikhMaxArr(iPT) > logPostrEv) then
              logPostrEv=logPriLikhMaxArr(iPT)
              postrMode=logPriLikhMaxPars(iPT,:)
            end if

          end do
          
          accRate=real(nAcc,dp)/nCand


          ! MIX COLD CHAINS TOGETHER VECTOR-WISE

          if (burnInStarted) then
            if (nPT == 1) then
              finCChn%links(finCChnPos:finCChnPos+multiPTBlkLen-1,:curNPars)=coldChains1Blk(1)%links(1:multiPTBlkLen,:curNPars)
              finCChn%logLikh(finCChnPos:finCChnPos+multiPTBlkLen-1)=coldChains1Blk(1)%logLikh(1:multiPTBlkLen)
              finCChnPos=finCChnPos+multiPTBlkLen
            else
              do iLink=1,multiPTBlkLen
                iPTMix=rndInt(1,nPT)
                finCChn%links(finCChnPos,:curNPars)=coldChains1Blk(iPTMix)%links(iLink,:curNPars)
                finCChn%logLikh(finCChnPos)=coldChains1Blk(iPTMix)%logLikh(iLink)
                call inc(finCChnPos)
              end do
            end if
          end if


          ! TIME MANAGEMENT

          secElapsMultiPTCore=omp_get_wtime()-tmStartMultiPTCore
          spdMultiPT=iBlk*multiPTBlkLen/(secElapsMultiPTCore-tParAdjActDurS)
          write(cCChnLinksPerS,"(I6)") nint(spdMultiPT)
          secRemain=nint(((nBlks-iBlk)*multiPTBlkLen)/spdMultiPT)
          if (isDef(curUsrMaxRuntmS) .AND. .NOT. overrideMaxRuntm) &
           secRemain=min(secRemain,max(0,curMaxRuntmRemS-nint(secElapsMultiPTCore)))


          ! UPDATE DISPLAY

          if (omp_get_wtime()-tmLstDispUpd >= DISP_UPD_INTVL_DUR_S) then

            write(cMaxPSR,"(A11)") trim(fmtNr(maxPSR,"(F11.7)",11,NO_FLAGS))
            write(cAccRate,"(A10)") trim(fmtNr(accRate,"(F10.8)",10,NO_FLAGS))

            statTbl(1)=fmtTime(nint(secElapsMultiPTCore),use0=.TRUE.,hrs3=.TRUE.)
            statTbl(2)=fmtTime(secRemain,use0=.TRUE.,hrs3=.TRUE.)
            statTbl(3)=fmtNr(nint(spdMultiPT),7,NO_FLAGS)
            statTbl(4)=cAccRate
            statTbl(5)=fmtNr(real(finCChnPos-1),"(ES8.2E2)",8,NO_FLAGS)
            statTbl(6)=fmtNr(merge(round(fracIndepMin*(finCChnPos-1),1._dp),VAL_UNDEF_DBL,dispPSR .AND. fracIndepMin >= 0), &
             "(ES6.0E2)",6,NO_FLAGS)
            statTbl(7)=merge("compl. ",merge("started","pending",burnInStarted),burnInFinished)
            statTbl(8)=mrg(cMaxPSR,STR_UNDEF_BY_LEN(9)//" ",dispPSR)
            statTbl(9)=mrg(parNms(iMaxPSR),STR_UNDEF_BY_LEN(7),dispPSR)
            statTbl(10)=fmtTime(merge(nint((PSRUpdINxtBlk-iBlk)*multiPTBlkLen/spdMultiPT,i4b),VAL_UNDEF_INT,dispPSR),use0=.TRUE.)

            call printMsg(UNIT%cons,fmtStatTbl(statTbl),setFlg([FLG%simpPrfx,FLG%noAdvc]))

            tmLstDispUpd=omp_get_wtime()

          end if


          ! ASSESS CONVERGENCE

          if (iBlk == nBlks .OR. (iBlk >= PSRUpdINxtBlk .AND. (iBlk-iBlk1PSR)*multiPTBlkLen >= PSR_CALC_N_LINKS_MIN) .OR. &
           (.NOT. burnInStarted .AND. (iBlk >= iBlkBurnInStartMax))) then

            if (.NOT. PSRCalcDeact) then
              call inc(nPSR)
              call calcPSR(coldChains,iBlk1PSR*multiPTBlkLen+1,iBlk*multiPTBlkLen,PSR(nPSR,:)%y,maxPSR,iMaxPSR)
              where (.NOT. parFixed(:curNPars)) PSR(nPSR,:)%x=1._dp*iBlk*multiPTBlkLen
              PSRIBlks(nPSR)=iBlk
              dispPSR=.TRUE.
            end if


            ! START OR FINISH BURN-IN

            if (burnInStarted) then

              ! burn-in started => PSRThrld2 reached?

              burnInFinished=.NOT.PSRCalcDeact .AND. maxPSR <= PSRThrld2 .AND. &
               infcLenFrac(iNPlMd)*(finCChnPos-1)*nThin(iNPlMd) >= UTH_INFC_LEN_MIN .AND. &
               infcLenFrac(iNPlMd)*(finCChnPos-1) >= TH_INFC_LEN_MIN .AND. &
               (1.-infcLenFrac(iNPlMd))*(finCChnPos-1)*nThin(iNPlMd) >= UTH_DISCARD_LEN_MIN .AND. &
               infcLenFrac(iNPlMd)*(finCChnPos-1) >= usrInfcLenMin(iNPlMd)

            else

              ! burn-in not yet started; calc. PSRThrld1 and see if it has already been fallen below

              burnInStarted=iBlk >= iBlkBurnInStartMax
              if (.NOT. PSRCalcDeact) then
                call setIfUndef(PSRThrld1,max(PSR_THRLD1_2_RATIO_MIN*PSRThrld2,PSRThrld2+PSRThrld1Fact(iNPlMd)*(maxPSR-PSRThrld2)))
                call setIfUndef(PSRNxtThrld,PSRThrld1)
                burnInStarted=burnInStarted .OR. maxPSR <= PSRThrld1
              end if

              if (burnInStarted) then
                if (usrAdjustTPar(iNPlMd)) then
                  ! temp. par. adjustment
                  tmTmp=omp_get_wtime()
                  call adjustTPar(iniPars,nint(spdMultiPT*tParAdjDurS))
                  if (.NOT. usrQuiet) write(*,"(A)") indentPrefx(I_UNIT%cons,simp=.TRUE.)
                  call printMsg(UNIT%cons,fmtStatTbl(DISP_TBL_HDR1,hdrArg=.TRUE.),setFlg(FLG%simpPrfx))
                  call printMsg(UNIT%cons,fmtStatTbl(DISP_TBL_HDR2,hdrArg=.TRUE.),setFlg(FLG%simpPrfx))
                  call printMsg(UNIT%cons,fmtStatTbl(hLinArg=.TRUE.),setFlg(FLG%simpPrfx))
                  tParAdjActDurS=omp_get_wtime()-tmTmp
                end if

                iBlkBurnInStart=iBlk
                nBlks=min(nBlks,max(iBlk,nBlks-(iBlkBurnInStartMax-iBlkBurnInStart)))                         ! burn-in earlier => less blocks in total
                if (allocated(finCChn%links)) deallocate(finCChn%links)                                       ! necessary since chains may be reused in main
                if (allocated(finCChn%logLikh)) deallocate(finCChn%logLikh)                                   ! necessary since chains may be reused in main
                allocate(finCChn%links(finCChnMaxLen,curNPars),finCChn%logLikh(finCChnMaxLen),stat=allocStat) ! allocate final cold chain
                call testAllocErr(allocStat,5)

                if (.NOT. PSRCalcDeact) PSRNxtThrld=PSRThrld2

              end if                 ! (burnInStarted)

            end if                   ! (burnInStarted)

            if (.NOT. PSRCalcDeact) call calcPSRUpdINxtBlk

          end if                     ! (PSRCalcDeact)



          ! CHECK IF CANCELED BY USER

          if (usrFinish()) then
            finishedByUsrReq=.TRUE.
            tmpInt=unlink(RUNTM_CTRL_FIL_NM%finish)

            if (.NOT. PSRCalcDeact) where (.NOT. parFixed(:curNPars)) finPSR=PSR(nPSR,:)%y
            oriInfcLen=floor(infcLenFrac(iNPlMd)*(finCChnPos-1))
            oriBurnInLen=finCChnPos-1-oriInfcLen
            finCChn%burnInLen=oriBurnInLen
            finCChn%length=finCChnPos-1
            exit
          end if



          ! CHECK IF BURN-IN FINISHED OR NO TIME LEFT

          if (isUndef(oriInfcLen) .AND. (burnInFinished .OR. iBlk == nBlks .OR. &
           (isDef(curUsrMaxRuntmS) .AND. secElapsMultiPTCore >= curMaxRuntmRemS))) then

            if (.NOT. PSRCalcDeact) where (.NOT. parFixed(:curNPars)) finPSR=PSR(nPSR,:)%y

            oriInfcLen=floor(infcLenFrac(iNPlMd)*(finCChnPos-1))
            oriBurnInLen=finCChnPos-1-oriInfcLen
            if (maximiseInfcLen) oriInfcLen=oriInfcLen+finCChnMaxLen-(finCChnPos-1)

            if (oriInfcLen*nThin(iNPlMd) < UTH_INFC_LEN_MIN) then
              ! original inference length < minimum inference length
              ! maximum runtime may be exceeded
              desirFinCChnLen=int(round(real(oriBurnInLen+ceiling(real(UTH_INFC_LEN_MIN,dp)/nThin(iNPlMd)),dp), &
               real(multiPTBlkLen,dp),ROUND_MD%up))
              if (desirFinCChnLen > finCChnMaxLen) then
                ! final chain cannot be long enough for original burn-in length + minimum inference length
                ! => shorten burn-in
                finCChn%burnInLen=max(0_i8b,finCChnMaxLen-ceiling(1.*UTH_INFC_LEN_MIN/nThin(iNPlMd)))
                finCChn%length=finCChnMaxLen
                warnShortenBurnIn=.TRUE.
              else
                finCChn%burnInLen=oriBurnInLen
                finCChn%length=desirFinCChnLen
                nBlks=min(nBlks,max(iBlk,iBlkBurnInStart+finCChn%length/multiPTBlkLen))
              end if
              overrideMaxRuntm=.TRUE.
            else if (maximiseInfcLen) then
              ! time left, maximise inference length: sample until either the maximum chain length is reached or both maximum runtime is reached and burn-in is finished
              finCChn%burnInLen=oriBurnInLen
              finCChn%length=finCChnMaxLen
            else
              ! exit in a normal way now
              finCChn%burnInLen=oriBurnInLen
              finCChn%length=finCChnPos-1
              exit
            end if

            if (iBlk < nBlks) dispPSR=.FALSE.

          end if                     ! (isUndef(oriInfcLen) .AND. (burnInFinished .OR. iBlk == nBlks .OR. …))

          if (iBlk == nBlks .OR. (maximiseInfcLen .AND. burnInFinished .AND. isDef(curUsrMaxRuntmS) .AND. &
           secElapsMultiPTCore >= curMaxRuntmRemS .AND. .NOT. overrideMaxRuntm)) then
            ! exit now
            finCChn%length=finCChnPos-1
            exit
          end if

        end do                       ! blocks loop



      ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
      !
      !   AFTER COMPLETION OF MULTI-PT PROCEDURE
      !
      ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·

        if (.NOT. burnInStarted) call err("multi-PT procedure finished before start of burn-in; allow longer runtime/length")

        if (finCChnPos == 1) call err("multi-PT procedure produced a chain of length 0; allow longer runtime/length")

        if (.NOT. usrQuiet) write(*,"(A)") NEW_LIN//indentPrefx(I_UNIT%cons,simp=.TRUE.)
        call printMsg(UNIT%cons+UNIT%log,"Multi-PT procedure finished"//trim(mrg(" by user request","",finishedByUsrReq)))

        call endSubsect(UNIT%cons+UNIT%log) ! status

        call printQty(UNIT%cons+UNIT%log,"Duration [h:m:s]",fmtTime(nint(secElapsMultiPTCore)))
        if (isDef(curUsrMaxRuntmS)) call printQty(UNIT%cons+UNIT%log, &
         "Ratio <duration>/<max. total runtime> [%]",fmtNr(100.*secElapsMultiPTCore/curMaxRuntmRemS,"(F6.1)",6,NO_FLAGS))
        call printQty(UNIT%cons+UNIT%log,"Mean speed [1/s]",nint(spdMultiPT))
        call printQty(UNIT%cons+UNIT%log,"Cold-chain acceptance rate",accRate)
        call printQty(UNIT%cons+UNIT%log,"Max. PSR",maxPSR)

        do iPT=1,nPT
          deallocate(coldChains(iPT)%links,coldChains1Blk(iPT)%links,coldChains(iPT)%logLikh,coldChains1Blk(iPT)%logLikh)
        end do

        call endSubsect(UNIT%cons+UNIT%log) ! Starting several PT procedures in parallel



        ! PRINT SUMMARY

        call startSubsect(UNIT%cons+UNIT%log,"Summary of sampling")

        call printQty(UNIT%cons+UNIT%log,"Length of resulting (mixed) chain",finCChn%length)
        call printQty(UNIT%cons+UNIT%log,"Length of burn-in period",finCChn%burnInLen)
        call printQty(UNIT%cons+UNIT%log,"# links used for inference",finCChn%length-finCChn%burnInLen)

        if (.NOT. PSRCalcDeact) then
          if (burnInFinished) then
            call printMsg(UNIT%cons+UNIT%log,"The cold chains of all PT procedures appear to have converged &
            &according to the PSR criterion.")
            if (overrideMaxRuntm) call printMsg(UNIT%cons+UNIT%log,"Maximum runtime requested had to be overridden &
            &so as to use minimum number of links for inference.")
            if (warnShortenBurnIn) then
              call warn("original burn-in length had to be shortened so as to use minimum number of links for inference !")
              call printQty(UNIT%cons+UNIT%log,"Original burn-in length",oriBurnInLen)
              call endSubsect(UNIT%cons+UNIT%log)
            end if
            call printQty(UNIT%cons+UNIT%log,"Ratio <# links used for inference>/<burn-in length>", & !,"(F7.3)", &
             niceFmt(real(finCChn%length-finCChn%burnInLen)/finCChn%burnInLen))
            call printQty(UNIT%cons+UNIT%log,"Requested minimum ratio",aftrBurnRatioConv) !"(F7.3)",aftrBurnRatioConv)
          else
            call warn("according to the PSR, the cold chains of one or more PT procedures have NOT converged !")
          end if
        end if

        call endSubsect(UNIT%cons+UNIT%log) ! Summary of sampling

      else

        call err("max. runtime exceeded during speed measurement")

      end if                         ! (finCChnMaxLen > 0)

    end if                           !.NOT. onlySpdMeas

    call endSubsect(UNIT%cons+UNIT%log) ! Applying multi-PT method


    contains

    function usrFinish()             ! returns whether user has requested finishing of sampling
      implicit none
      logical :: &
       usrFinish
      inquire(file=RUNTM_CTRL_FIL_NM%finish,exist=usrFinish)
    end function


    subroutine calcPSRUpdINxtBlk     ! calcs. block index for next PSR update
      implicit none

      if (PSRCalcDeact) then
        PSRUpdINxtBlk=nBlks+1
      else if (burnInFinished .OR. PSRThrld2 == 1._dp) then
        PSRUpdINxtBlk=nBlks
      else if (iBlk < nBlks) then

        if (nPSR <= 2) then

          PSRUpdDistBlks=1

        else
        
          associate(lstPSR=>PSR(nPSR-1,iMaxPSR)%y)
            if (maxPSR > lstPSR) then ! PSR increasing
              absLstPSRGrad=abs(lstPSR-PSR(nPSR-2,iMaxPSR)%y)/(PSRIBlks(nPSR-1)-PSRIBlks(nPSR-2))
              curPSRGrad=(maxPSR-lstPSR)/(iBlk-PSRIBlks(nPSR-1))
              PSRUpdDistBlks=nint(abs(((-max(absLstPSRGrad,curPSRGrad))-curPSRGrad)/ &
               ((curPSRGrad-absLstPSRGrad)/(iBlk-PSRIBlks(nPSR-1)))))
            else
              PSRUpdDistBlks=nint((PSRNxtThrld-maxPSR)/((maxPSR-lstPSR)/(iBlk-PSRIBlks(nPSR-1))))
            end if
          end associate

          ! obey (strictest) bounds
          PSRUpdDistBlksRge%uBnd=floor(min(nBlks*(PSR_UPD_INTVL_RATIO_RGE%lBnd+ & ! upper bounds
            min(1._dp,sqrt((maxPSR-PSRNxtThrld)/(PSRThrld1-PSRThrld2)))*PSR_UPD_INTVL_RATIO_RGE%width()), &
            1._dp*PSR_UPD_DIST_RGE%uBnd/multiPTBlkLen))
          PSRUpdDistBlksRge%lBnd=ceiling(max(nBlks*PSR_UPD_INTVL_RATIO_RGE%lBnd, & ! lower bounds
            PSR_UPD_INTVL_RATIO_EL_RGE%lBnd*secElapsMultiPTCore*spdMultiPT/multiPTBlkLen, &
            1._dp*PSR_UPD_DIST_RGE%lBnd/multiPTBlkLen))

          if (PSRUpdDistBlksRge%width() < 0) then ! contradicting bounds
            PSRUpdDistBlks=round(intvlCentre(PSRUpdDistBlksRge),1) ! use mean of two bounds
          else
            PSRUpdDistBlks=putOntoIntvl(PSRUpdDistBlksRge,PSRUpdDistBlks)
          end if
          PSRUpdDistBlks=max(1_i8b,nint(putOntoIntvl(PSR_UPD_INTVL_DUR_S_RGE,PSRUpdDistBlks*multiPTBlkLen/spdMultiPT)* &
           spdMultiPT/multiPTBlkLen,i8b))

          if (PSRUpdDistBlks+iBlk >= nBlks) & ! next PSR calc. would be after finish
            ! maybe calc. PSR again on half way to finish; and calc. it at the very end in any case
           PSRUpdDistBlks=min(max(nint((nBlks-iBlk)/2.,i8b),ceiling(1.*PSR_UPD_DIST_RGE%lBnd/multiPTBlkLen,i8b)),nBlks-iBlk)

          if (.NOT. burnInStarted) &
           PSRUpdDistBlks=min(PSRUpdDistBlks,max(1_i8b,iBlkBurnInStartMax-iBlk))

        end if


        PSRUpdINxtBlk=max(iBlk+PSRUpdDistBlks,ceiling((iBlkBurnInStart+1.*PSR_CALC_N_LINKS_MIN/multiPTBlkLen+ &
         (infcLenFrac(iNPlMd)-1.)*iBlkBurnInStart)/infcLenFrac(iNPlMd),i8b))
        ! cf. [10.11.2010/2];   iBlk1PSR=iBlkBurnInStart+floor((1.-infcLenFrac(iNPlMd))*(iBlk-iBlkBurnInStart));
        !    (iBlk-iBlk1PSR)*multiPTBlkLen >= PSR_CALC_N_LINKS_MIN

      end if                         ! (.NOT. PSRCalcDeact)
    end subroutine


    function multiPTSpeed(nPT,iniPars,durS) ! returns multiPT speed [(final cold chain samples)/s]
      implicit none
      integer :: &
       iSpdMeas, &
       nPT, &
       allocStat, &
       iPT
      integer(i8b) :: &
       spdMeasCurLen, &
       spdMeasTotLen, &
       dmyNAcc
      real(dp) :: &
       iniPars(nPT,PTNChains,curNPars), &
       durS, &
       multiPTSpeed, &
       tmStart, &
       tmSpdMeasStart, &
       spdMeasLstDurS, &
       spdMeasCurSpd, &
       spdMeasTotDurS, &
       spdMeasTmLeft, &
       dmyLogPriLikhMax, &
       dmyLogPriLikhMaxPars(curNPars)
      type(tyMCMCChain) :: &
       spdMeasChains(nPT)
      intent(in) :: &
       nPT, &
       durS
      intent(inout) :: &
       iniPars


      iSpdMeas=1
      spdMeasTotLen=0_i8b
      spdMeasTotDurS=0._dp
      tmSpdMeasStart=omp_get_wtime()

      spdMeasCurSpd=curSpdEst

SPD_MEAS_LOOP: &
      do                             ! speed measurement loop

        do                           ! loop to avoid timing error

          spdMeasTmLeft=durS-(omp_get_wtime()-tmSpdMeasStart)
          spdMeasCurLen=max(nint(spdMeasTmLeft*spdMeasCurSpd,i8b),meanPTBlkLen(iNPlMd))

          if (spdMeasCurLen < 0 .OR. (durS > SPD_SUBMEAS_DUR_S_MIN .AND. spdMeasTmLeft < SPD_SUBMEAS_DUR_S_MIN)) exit SPD_MEAS_LOOP

          do iPT=1,nPT
            allocate(spdMeasChains(iPT)%links(spdMeasCurLen,curNPars),spdMeasChains(iPT)%logLikh(spdMeasCurLen),stat=allocStat)
            call testAllocErr(allocStat,12)
            spdMeasChains(iPT)%length=spdMeasCurLen
          end do

          tmStart=omp_get_wtime()

          ! CAUTION: don't use OMP for loop calling prlTempering!
          !          (speed decreased to ~ (60±10)% when calling
          !          'base am.csv rv_tmp_pri.csv rv_tmp_sec.csv --only-speed-meas --speed-meas-dur=5s' (19.05.2010))
          do iPT=1,nPT
            call prlTempering(spdMeasChains(iPT),meanPTBlkLen(iNPlMd),PTNChains,iniPars(iPT,:,:),OPT_PT_PRESERVE_INIPARS, &
             tPar(iPT,:),OPT_PT_NORMAL,dmyLogPriLikhMax,dmyLogPriLikhMaxPars,dmyNAcc,.FALSE.) ! iniParsArg NOT CONTIGUOUS
          end do

          spdMeasLstDurS=omp_get_wtime()-tmStart
          do iPT=1,nPT
            deallocate(spdMeasChains(iPT)%links,spdMeasChains(iPT)%logLikh)
          end do

          spdMeasCurSpd=spdMeasCurLen/spdMeasLstDurS
          if (spdMeasCurSpd > 0. .OR. iSpdMeas > 1) exit ! no timing error (any more)

          spdMeasCurSpd=curSpdEst

        end do                       ! loop to avoid timing error

        if (spdMeasLstDurS > 0.) then
          spdMeasTotLen=spdMeasTotLen+spdMeasCurLen
          spdMeasTotDurS=spdMeasTotDurS+spdMeasLstDurS
        end if

        if (spdMeasLstDurS <= 0. .OR. durS-(omp_get_wtime()-tmSpdMeasStart) < SPD_SUBMEAS_DUR_S_MIN) exit SPD_MEAS_LOOP

        call inc(iSpdMeas)

      end do SPD_MEAS_LOOP           ! speed measurement loop

      multiPTSpeed=spdMeasTotLen/spdMeasTotDurS ! est. speed of multiPT core procedure (# final links per s)
    end function


    subroutine printTPar             ! prints temp. pars.
      implicit none
      integer :: &
       i

      call startSubsect(UNIT%cons+UNIT%log,"Tempering parameters")
      do i=1,PTNChains
        call printQty(UNIT%cons+UNIT%log,"Chain "//trim(niceFmt(i)),tPar(1,i))
      end do
      call endSubsect(UNIT%cons+UNIT%log) ! Tempering parameters
    end subroutine


    subroutine adjustTPar(iniPars,nLinks) ! adjusts tempering pars.
      implicit none
      integer :: &
       iPT, &
       allocStat, &
       i, &
       nLinks
      integer(i8b) :: &
       dmyNAcc
      real(dp) :: &
       iniPars(nPT,PTNChains,curNPars), &
       dmyLogPriLikhMax, &
       dmyLogPriLikhMaxPars(curNPars)
      type(tyMCMCChain) :: &
       coldChains(nPT)
      intent(inout) :: &
       iniPars
      intent(in) :: &
       nLinks

      if (.NOT. usrQuiet) write(*,"(A)") NEW_LIN//indentPrefx(I_UNIT%cons,simp=.TRUE.)
      call startSubsect(UNIT%cons+UNIT%log,"Adjusting tempering parameters")


      do iPT=1,nPT
        allocate(coldChains(iPT)%links(nLinks,curNPars),coldChains(iPT)%logLikh(nLinks),stat=allocStat)
        call testAllocErr(allocStat,6)
        coldChains(iPT)%length=nLinks

        call prlTempering(coldChains(iPT),TPAR_ADJ_BLK_LEN,PTNChains,iniPars(iPT,:,:),OPT_PT_PRESERVE_INIPARS,tPar(iPT,:), &
         OPT_PT_ADJ_TPAR,dmyLogPriLikhMax,dmyLogPriLikhMaxPars,dmyNAcc,.FALSE.,tParSumOfSq(iPT),iPT) ! iniParsArg NOT CONTIGUOUS

        deallocate(coldChains(iPT)%links,coldChains(iPT)%logLikh)
      end do

      if (.NOT. usrQuiet) print *
      call startSubsect(UNIT%cons+UNIT%log,"Results")
      do i=1,PTNChains
        call startSubsect(UNIT%cons+UNIT%log,"Chain "//trim(niceFmt(i)))
        do iPT=1,nPT
          call printQty(UNIT%cons+UNIT%log,"PT procedure "//trim(niceFmt(iPT)),tPar(iPT,i))
        end do
        call endSubsect(UNIT%cons+UNIT%log) ! chain …
      end do

      call startSubsect(UNIT%cons+UNIT%log, &
       "Fraction of up-drifting samples over temperature: sum of squares of residuals from straight line")
      do iPT=1,nPT
        call printQty(UNIT%cons+UNIT%log,"PT procedure "//trim(niceFmt(iPT)),tParSumOfSq(iPT))
      end do
      call endSubsect(UNIT%cons+UNIT%log) ! Sum of squares of straight-line residuals

      call endSubsect(UNIT%cons+UNIT%log) ! Results
      call endSubsect(UNIT%cons+UNIT%log) ! Determining best-suited tempering parameters
    end subroutine


    function fmtStatTbl(str,hLinArg,hdrArg) ! returns a formatted status table
      implicit none
      integer :: &
       i, &
       j, &
       curW, &
       curWReq, &
       curWTot, &
       nSpcL, &
       nSpcR
      logical :: &
       curIsNum, &
       hLinArg, &
       hLin, &
       hdrArg, &
       hdr
      character :: &
       vSep
      character(*) :: &
       str(DISP_TBL_N_COLS)
      character(DISP_TBL_COL_W_MAX) :: &
       curStr
      character((DISP_TBL_COL_W_MAX+1)*DISP_TBL_N_COLS+1) :: &
       fmtStatTbl
      optional :: &
       str, &
       hLinArg, &
       hdrArg
      intent(in) :: &
       str, &
       hLinArg, &
       hdrArg

      curWTot=0
      fmtStatTbl=""
      hLin=present(hLinArg) .AND. hLinArg
      hdr=present(hdrArg) .AND. hdrArg

      i=1
      do
        curW=DISP_TBL_COL_W(i)
        do j=1,DISP_TBL_N_COLS-i
          if (.NOT. hLin .AND. isUndef(str(i+j))) then
            curW=curW+1+DISP_TBL_COL_W(i+j)
          else
            exit
          end if
        end do
        if (hLin) then
          vSep="+"
          curWReq=curW
          curStr=repeat("-",curW)
          curIsNum=.FALSE.
        else
          vSep=merge(" ","|",DISP_TBL_NO_V_BAR .AND. .NOT. hdr)
          curStr=trim(adjustl(str(i)))
          curWReq=len_trim(curStr)
          if (curWReq > curW-2*DISP_TBL_PADD) then
            curStr=curStr(1:curW-2*DISP_TBL_PADD-2)//".."
            curWReq=curW-2*DISP_TBL_PADD
          end if
          curIsNum=isNumeric(curStr)
        end if
        if (curIsNum) then
          ! numbers are generally right-justified
          nSpcR=DISP_TBL_PADD
          nSpcL=curW-curWReq-nSpcR
        else if (hdr) then
          ! text in header is centered (with a touch to the left if needed)
          nSpcL=floor((curW-curWReq)/2.)
          nSpcR=ceiling((curW-curWReq)/2.)
        else
          ! text is left-justified elsewhere
          if (hLin) then
            nSpcL=0
            nSpcR=0
          else
            nSpcL=DISP_TBL_PADD
            nSpcR=curW-curWReq-nSpcL
          end if
        end if
        curStr=repeat(" ",nSpcL)//curStr(1:curWReq)//repeat(" ",nSpcR)
        fmtStatTbl=fmtStatTbl(1:curWTot)//vSep//curStr(1:curW)
        curWTot=curWTot+1+curW
        i=i+j
        if (i > DISP_TBL_N_COLS) exit
      end do
      fmtStatTbl=fmtStatTbl(1:curWTot)//vSep
    end function

  end subroutine multiPT




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcPSR(chains,iMin,iMax,PSRVals,maxPSR,iMaxPSR)
!
!   calculates potential scale reductions (PSR) for all pars. in all chains
!
!   Arguments:
!     chains                         MCMC chains
!     iMin                           min. link index
!     iMax                           max. link index
!     PSRVals                        calculated PSR values
!     maxPSR                         max. PSR
!     iMaxPSR                        par. index with max. PSR
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::


    implicit none
    integer :: &
     iMaxPSR, &
     p, &
     iPT
    integer(i8b) :: &
     iMin, &
     iMax
    real(dp) :: &
     PSRVals(curNPars), &
     curPSR, &
     curPSRStd, &
     maxPSR, &
     W, &
     WStd, &
     B, &
     BStd, &
     VMod, &                         ! (d+3)/(d+1)·V
     VModStd, &
     mean1(nPT), &
     mean1Std(nPT), &
     leng, &
     DOFV, &                         ! est. # degrees of freedom of V (cf. [18.05.2010/1])
     DOFVStd, &
     vceV, &
     vceVStd
    type(tyMCMCChain) :: &
     chains(:)
    intent(in) :: &
     chains, &
     iMin, &
     iMax
    intent(out) :: &
     PSRVals, &
     maxPSR, &
     iMaxPSR


    nPT=size(chains)

    leng=iMax-iMin+1

    !$OMP parallel do private(B,W,BStd,WStd,iPT,mean1,DOFV,vceV,mean1Std,DOFVStd,vceVStd,curPSR,curPSRStd)
    do p=1,curNPars

      if (.NOT. parFixed(p)) then

        B=0
        W=0
        vceV=0
        BStd=0
        WStd=0
        vceVStd=0

        do iPT=1,nPT

          call calcQties1(mean1(iPT),W,vceV,iPT,p)
          if (cycParUsMaxRge(p)) call calcQties1Std(mean1Std(iPT),WStd,vceVStd,iPT,p) ! calculate "standardised" mean and variance, cf. [06.09.2009/2]

        end do

        call calcQties2(DOFV,B,W,VMod,vceV,mean1)
        curPSR=PSRFct(VMod,W)

        if (cycParUsMaxRge(p)) then  ! re-calc. PSR
          call calcQties2(DOFVStd,BStd,WStd,VModStd,vceVStd,mean1Std)
          curPSRStd=PSRFct(VModStd,WStd)
          if (curPSR < curPSRStd) then
            PSRVals(p)=curPSR
            fracIndep(p)=nPT*max(0._dp,min(1._dp,VMod/B))
          else
            PSRVals(p)=curPSRStd
            fracIndep(p)=nPT*max(0._dp,min(1._dp,VModStd/BStd))
          end if
        else
          PSRVals(p)=curPSR
          fracIndep(p)=nPT*max(0._dp,min(1._dp,VMod/B))
        end if

      end if                         ! (.NOT. parFixed(p))
    end do                           ! p=1,curNPars
    !$OMP end parallel do

    call findMaxValIndx(merge(-HUGE_DP,PSRVals,parFixed(:curNPars)),maxPSR,iMaxPSR)
    fracIndepMin=minval(merge(fracIndep(:curNPars),HUGE_DP,isDef(fracIndep(:curNPars))))


    contains

    function PSRFct(VMod,W)
      implicit none
      real(dp) :: &
       VMod, &
       W, &
       PSRFct
      intent(in) :: &
       VMod, &
       W

      PSRFct=sqrt(VMod/W)
    end function


    subroutine calcQties1(mean1,W,vceV,iPT,p)
      implicit none
      integer :: &
       iPT, &
       p
      real(dp) :: &
       mean1, &
       W, &
       vceV, &
       vceTmp
      intent(inout) :: &
       mean1, &
       W, &
       vceV
      intent(in) :: &
       iPT, &
       p

      mean1=1._dp/leng*sum(chains(iPT)%links(iMin:iMax,p))
      vceTmp=sum((chains(iPT)%links(iMin:iMax,p)-mean1)**2)
      W=W+vceTmp
      vceV=vceV+vceTmp**2
    end subroutine


    subroutine calcQties1Std(mean1,W,vceV,iPT,p)

      implicit none
      integer :: &
       iPT, &
       p, &
       j, &
       jMin
      integer(i8b) :: &
       i
      real(dp) :: &
       mean1, &
       W, &
       vceV, &
       vceTmp, &
       u, &
       v1, &
       v1Min, &
       stdMean1, &
       stdMean, &
       m1, &
       m2
      intent(in) :: &
       iPT, &
       p
      intent(inout) :: &
       mean1, &
       W, &
       vceV


      B=0
      W=0
      vceV=0

      u=parPeriod(p)                 ! upper bound for angle-like pars.

      ! MINIMIZE VARIANCE v1

      v1Min=huge(v1Min)
      do j=0,15

        m1=0                         ! mean m1 only used for variance v1
        do i=iMin,iMax
          m1=m1+shAng(chains(iPT)%links(i,p)-j/16._dp*u,u)
        end do
        m1=m1/leng

        v1=0
        do i=iMin,iMax
          v1=v1+(shAng(chains(iPT)%links(i,p)-j/16._dp*u,u)-m1)**2
        end do
        v1=v1/leng

        if (v1 < v1Min) then
          v1Min=v1
          jMin=j
        end if

      end do


      ! CALC. FIRST ESTIMATE FOR STANDARDIZED MEAN 'stdMean1'

      stdMean1=0
      do i=iMin,iMax
        stdMean1=stdMean1+shAng(chains(iPT)%links(i,p)-jMin/16._dp*u,u)
      end do
      stdMean1=stdMean1/leng+jMin/16._dp*u


      ! CALC. FINAL STANDARDIZED MEAN AND VARIANCE

      stdMean=0
      do i=iMin,iMax
        stdMean=stdMean+shAng(chains(iPT)%links(i,p)-stdMean1,u)
      end do
      stdMean=stdMean/leng+stdMean1

      m2=0
      do i=iMin,iMax
        m2=m2+shAng(chains(iPT)%links(i,p)-stdMean1,u)
      end do
      m2=m2/leng

      vceTmp=0
      do i=iMin,iMax
        vceTmp=vceTmp+(shAng(chains(iPT)%links(i,p)-stdMean1,u)-m2)**2
      end do
      W=W+vceTmp
      vceV=vceV+vceTmp**2


      ! USE THE STANDARDIZED QUANTITIES FOR PSR CALC.

      mean1=stdMean

    end subroutine


    subroutine calcQties2(DOFV,B,W,VMod,vceV,mean1)
      implicit none
      real(dp) :: &
       DOFV, &
       B, &
       W, &
       VMod, &
       mean2, &
       vceV, &
       mean1(nPT)
      intent(inout) :: &
       DOFV, &
       B, &
       W, &
       vceV
      intent(in) :: &
       mean1
      intent(out) :: &
       VMod
      W=W/(nPT*(leng-1))
      mean2=sum(mean1)/nPT
      B=(sum((mean1-mean2)**2)*leng)/(nPT-1)
      vceV=2._dp/nPT**2*(vceV/(leng**2*(leng-1))+((nPT+1)*B/leng)**2/(nPT-1)) ! cf. [18.05.2010/3]
      DOFV=2._dp*((W*(leng-1)+(B*(nPT+1))/nPT)/leng)**2/vceV                  ! cf. [18.05.2010/1]
      VMod=(DOFV+3._dp)/(DOFV+1._dp)*(W*(leng-1)+B*(nPT+1)/nPT)/leng          ! cf. [18.05.2010/1]
    end subroutine


    function shAng(a,aM)
      ! shifted angle; shAng(·,aM): [-aM,aM) → [-aM/2,aM/2)

      implicit none
      real(dp) :: &
       shAng, &
       a, &
       aM

      shAng=modulo(a+aM,aM)-aM/2._dp

    end function


  end subroutine calcPSR




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printPSR
!
!   prints PSR values
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i

    call startSubsect(UNIT%cons+UNIT%log,"Final PSR values")
    if (any(isDef(finPSR(:curNPars)))) then
      do i=1,curNPars
        call printQty(UNIT%cons+UNIT%log,parNms(i),finPSR(i))
      end do
    else
      call printMsg(UNIT%cons+UNIT%log,"<not calculated>")
    end if
    call endSubsect(UNIT%cons+UNIT%log) ! Final PSR values

  end subroutine printPSR




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcAutocorr(iNPlMd,chn,p,mean,vce)
!
!   calculates autocorrelation of parameter samples
!
!   Arguments:
!     iNPlMd                         #-pl.-mode index
!     chn                            MCMC chain
!     p                              par. index
!     mean                           sample mean
!     vce                            sample variance
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iNPlMd, &
     p, &
     iLag
    integer(i8b) :: &
     lagPrev, &
     lag, &
     N, &
     nUsed
    real(dp) :: &
     mean, &
     vce
    type(tyMCMCChain) :: &
     chn
    intent(in) :: &
     iNPlMd, &
     chn, &
     p, &
     mean, &
     vce
    
    N=chn%length
    autoCorrNLags(iNPlMd)=0
    lag=AUTOCORR_LAG_MIN
    
    do iLag=1,AUTOCORR_N_LAGS_MAX
      autoCorrLags(iLag)=lag
      nUsed=N-lag-chn%burnInLen
      if (nUsed < 2) then
        call undef(autoCorr(iNPlMd,p,iLag:))
      else
        autoCorr(iNPlMd,p,iLag)=arrMean((chn%links(chn%burnInLen+1:N-lag,p)-mean)*(chn%links(chn%burnInLen+1+lag:N,p)-mean))/vce
        call inc(autoCorrNLags(iNPlMd))
      end if
      lagPrev=lag
      lag=ceiling(lagPrev*AUTOCORR_LAG_FACT) ! (iLag-1)*AUTOCORR_LAG_STRIDE
    end do

  end subroutine calcAutocorr




  !==================================================[ POSTERIOR-SAMPLES I/O ]======================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine postrSampIO(chn,output,finish,postrMode,nPlReq)
!
!   reads or saves posterior samples and other information to a saved-samples file
!
!   Arguments:
!     chn                            MCMC chain
!     output                         whether samples should be saved
!     finish                         whether output file should be finalised
!     postrMode                      posterior mode (saved in such files because it is the basis for successive calculations)
!     nPlReq                         # planets requested when reading
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*),parameter :: &
     SAV_SAMP_FIL_IDENT="BASE posterior samples"
    integer :: &
     nPlReq, &                       ! needs to be supplied after first reading (UNDEFINED if bin. mode)
     allocStat, &
     confFilLenTot, &
     priNPtsMax, &
     iMode, &                        ! for reading
     i, &
     j, &
     iP, &
     curSeqSiz, &
     seqPos, &
     k
    integer,parameter :: &
     REV_NR_USER_DEF_PRI=93, &
     REV_NR_POSTR_MODE=102, &
     REV_NR_ZERO_SAMPLE_BUG=107, &
     REV_NR_PRI_VCE=158, &
     REV_NR_EPOCH=175, &
     REV_NR_NO_END_PRG=179, &
     REV_NR_NO_M_S=180
    integer(i8b) :: &
     iSamp, &
     iSamp2, &
     lstSeqStart, &
     iSeq, &
     iSamp0(N_SAMP_ZERO_MAX+1), &
     i0, &
     nSamp0, &
     nSamp0Burn
    integer(i4b) :: &
     CRCVal, &
     CRCValSav
    real(dp) :: &
     postrMode(:), &
     curVal
    logical :: &
     output, &
     finish, &                       ! for writing, signals that file should be finished; ignored for reading
     fstCall=.TRUE., &
     compr
    character(STR_LEN_STD) :: &
     tmpStr
    type(tyMCMCChain) :: &
     chn
    optional :: &
     nPlReq, &
     postrMode
    save :: &
     iMode, &
     fstCall, &
     CRCVal
    intent(in) :: &
     output, &
     finish, &
     nPlReq
    intent(inout) :: &
     chn, &
     postrMode




    if (output) then                 ! write

      call startSubsect(UNIT%cons+UNIT%log,"Writing posterior samples")

      if (fstCall) then

        ! global header

        fstCall=.FALSE.
        CRCVal=0
        call tryOpenFilRW("posterior-samples",savSamp%filNmSav,UN_SAV_SAMP,access="stream")

          ! global header, 1st part (constant length)

        savSamp%hdr%hIdent=SAV_SAMP_FIL_IDENT
        savSamp%hdr%hRevNr=BASE_REV_NR
        savSamp%hdr%hResrvd=repeat(char(0),len(savSamp%hdr%hResrvd))
        savSamp%hdr%hLenDiv(1)=len_trim(cmdLin)
        savSamp%hdr%hLenDiv(2)=len_trim(hostName)
        savSamp%hdr%hLenDiv(3)=len_trim(dtProps%sysNm)
        savSamp%hdr%hNConfFilLins=nConfFilLins
        savSamp%hdr%hDtStart=dateStartPrg
        savSamp%hdr%hNDtSets=dtProps%nSets
        savSamp%hdr%hT1=dtProps%tRge%lBnd
        savSamp%hdr%hTN=dtProps%tRge%uBnd
        savSamp%hdr%hEpoch=mdlProps%epoch
        savSamp%hdr%hTimesAreJD=boolToInt1(dtProps%timesAreJD)
        savSamp%hdr%hNDt=dtProps%nRec
        savSamp%hdr%hDtCRC=calcDtCRC()
        if (mdlProps%bin) then
          call undef(savSamp%hdr%hNNPlMds)
        else
          savSamp%hdr%hNNPlMds=nNPlMds ! bin. mode: undef.
        end if
        savSamp%hdr%hNPT=nPT
        savSamp%hdr%hPTNChains=PTNChains
        confFilLenTot=sum(len_trim(confFilLins(:nConfFilLins)))

        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hIdent,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hRevNr,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hResrvd,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hLenDiv,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hNConfFilLins,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hDtStart,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hNDtSets,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hT1,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hTN,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hEpoch,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hTimesAreJD,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hNDt,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hDtCRC,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hNNPlMds,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hNPT,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hPTNChains,CRCVal)

          ! global header, 2nd part (variable length)

        if (allocated(savSamp%hdr%hDtTy)) deallocate(savSamp%hdr%hDtTy)
        if (allocated(savSamp%hdr%hCmdLin)) deallocate(savSamp%hdr%hCmdLin)
        if (allocated(savSamp%hdr%hHostName)) deallocate(savSamp%hdr%hHostName)
        if (allocated(savSamp%hdr%hSysNm)) deallocate(savSamp%hdr%hSysNm)
        if (allocated(savSamp%hdr%hConfFilLinLen)) deallocate(savSamp%hdr%hConfFilLinLen)
        if (allocated(savSamp%hdr%hConfFilLins)) deallocate(savSamp%hdr%hConfFilLins)
        allocate(savSamp%hdr%hDtTy(savSamp%hdr%hNDtSets(1)),savSamp%hdr%hCmdLin(savSamp%hdr%hLenDiv(1)), &
         savSamp%hdr%hHostName(savSamp%hdr%hLenDiv(2)),savSamp%hdr%hSysNm(savSamp%hdr%hLenDiv(3)), &
         savSamp%hdr%hConfFilLinLen(savSamp%hdr%hNConfFilLins(1)),savSamp%hdr%hConfFilLins(confFilLenTot),stat=allocStat)
        call testAllocErr(allocStat,35)
        savSamp%hdr%hDtTy(1:savSamp%hdr%hNDtSets(1))=dtProps%dtTyOfSets(1:savSamp%hdr%hNDtSets(1))
        savSamp%hdr%hCmdLin=transfer(trim(cmdLin),savSamp%hdr%hCmdLin)
        savSamp%hdr%hHostName=transfer(trim(hostName),savSamp%hdr%hHostName)
        savSamp%hdr%hSysNm=transfer(trim(dtProps%sysNm),savSamp%hdr%hSysNm)
        j=0
        do i=1,savSamp%hdr%hNConfFilLins(1)
          savSamp%hdr%hConfFilLinLen(i)=len_trim(confFilLins(i))
          savSamp%hdr%hConfFilLins(j+1:j+savSamp%hdr%hConfFilLinLen(i))=transfer(trim(confFilLins(i)),["a"])
          j=j+savSamp%hdr%hConfFilLinLen(i)
        end do

        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hDtTy,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hCmdLin(1:savSamp%hdr%hLenDiv(1)),CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hHostName(1:savSamp%hdr%hLenDiv(2)),CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hSysNm(1:savSamp%hdr%hLenDiv(3)),CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hConfFilLinLen,CRCVal)
        call writeFilCRC(UN_SAV_SAMP,savSamp%hdr%hConfFilLins,CRCVal)

        deallocate(savSamp%hdr%hDtTy,savSamp%hdr%hCmdLin,savSamp%hdr%hHostName,savSamp%hdr%hSysNm,savSamp%hdr%hConfFilLinLen, &
         savSamp%hdr%hConfFilLins)

      end if                         ! (fstCall)


      ! n-planet-mode header

        ! n-planet-mode header, 1st part (constant length)

      if (mdlProps%bin) then
        call undef(savSamp%nPlMd%hNPl)
      else
        savSamp%nPlMd%hNPl=curNPl
      end if
      savSamp%nPlMd%hNPars=curNPars
      savSamp%nPlMd%hNSamp=chn%length
      savSamp%nPlMd%hNSampBurn=chn%burnInLen
      savSamp%nPlMd%hNThin=nThin(iNPlMd)
      savSamp%nPlMd%hMeanPTBlkLen=meanPTBlkLen(iNPlMd)
      savSamp%nPlMd%hTParAdj=boolToInt1(usrAdjustTPar(iNPlMd))
      savSamp%nPlMd%hFinByUsrReq=boolToInt1(finishedByUsrReq)
      if (allocated(savSamp%samp%sNSeq)) deallocate(savSamp%samp%sNSeq)
      allocate(savSamp%samp%sNSeq(curNPars),stat=allocStat)
      call testAllocErr(allocStat,41)
      call breakIntoSeq(.FALSE.)
      if (1.*sum(savSamp%samp%sNSeq) < 8./(8+1)*curNPars*chn%length) then ! cf. [25.7.2011/2]
        savSamp%nPlMd%hComprMd=1     ! RLE
      else
        savSamp%nPlMd%hComprMd=0     ! none
      end if
      compr=savSamp%nPlMd%hComprMd(1) == 1

      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hNPl,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hNPars,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hNSamp,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hNSampBurn,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hNThin,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hMeanPTBlkLen,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hTParAdj,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hFinByUsrReq,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hComprMd,CRCVal)

        ! n-planet-mode header, 2nd part (variable length)

      priNPtsMax=maxval(priNPts(1:curNPars))

      if (allocated(savSamp%nPlMd%hTPar)) deallocate(savSamp%nPlMd%hTPar)
      if (allocated(savSamp%nPlMd%hParNms)) deallocate(savSamp%nPlMd%hParNms)
      if (allocated(savSamp%nPlMd%hLenParNms)) deallocate(savSamp%nPlMd%hLenParNms)
      if (allocated(savSamp%nPlMd%hPriTypes)) deallocate(savSamp%nPlMd%hPriTypes)
      if (allocated(savSamp%nPlMd%hPriLBnd)) deallocate(savSamp%nPlMd%hPriLBnd)
      if (allocated(savSamp%nPlMd%hPriUBnd)) deallocate(savSamp%nPlMd%hPriUBnd)
      if (allocated(savSamp%nPlMd%hPriKnees)) deallocate(savSamp%nPlMd%hPriKnees)
      if (allocated(savSamp%nPlMd%hPriVce)) deallocate(savSamp%nPlMd%hPriVce)
      if (allocated(savSamp%nPlMd%hFinPSR)) deallocate(savSamp%nPlMd%hFinPSR)
      if (allocated(savSamp%nPlMd%hFracIndep)) deallocate(savSamp%nPlMd%hFracIndep)
      if (allocated(savSamp%nPlMd%hPriNPts)) deallocate(savSamp%nPlMd%hPriNPts)
      if (allocated(savSamp%nPlMd%hPriPtsPsdX)) deallocate(savSamp%nPlMd%hPriPtsPsdX)
      if (allocated(savSamp%nPlMd%hPriPtsY)) deallocate(savSamp%nPlMd%hPriPtsY)
      if (allocated(savSamp%nPlMd%hPriLogScl)) deallocate(savSamp%nPlMd%hPriLogScl)
      if (allocated(savSamp%nPlMd%hPriInvDelta)) deallocate(savSamp%nPlMd%hPriInvDelta)
      if (allocated(savSamp%nPlMd%hParEstMode)) deallocate(savSamp%nPlMd%hParEstMode)
      allocate(savSamp%nPlMd%hTPar(savSamp%hdr%hNPT(1),savSamp%hdr%hPTNChains(1)), &
       savSamp%nPlMd%hParNms(sum(len_trim(parNms(:curNPars)))), &
       savSamp%nPlMd%hLenParNms(curNPars), &
       savSamp%nPlMd%hPriTypes(curNPars),savSamp%nPlMd%hPriLBnd(curNPars),savSamp%nPlMd%hPriUBnd(curNPars), &
       savSamp%nPlMd%hPriKnees(curNPars),savSamp%nPlMd%hPriVce(curNPars),savSamp%nPlMd%hFinPSR(curNPars), &
       savSamp%nPlMd%hFracIndep(curNPars), &
       savSamp%nPlMd%hPriNPts(curNPars),savSamp%nPlMd%hPriPtsPsdX(curNPars,priNPtsMax), &
       savSamp%nPlMd%hPriPtsY(curNPars,priNPtsMax),savSamp%nPlMd%hPriLogScl(curNPars),savSamp%nPlMd%hPriInvDelta(curNPars), &
       savSamp%nPlMd%hParEstMode(curNPars),stat=allocStat)
      call testAllocErr(allocStat,36)
      do i=1,savSamp%hdr%hNPT(1)
        savSamp%nPlMd%hTPar(i,:)=tPar(i,:)
      end do
      j=0
      do i=1,curNPars
        savSamp%nPlMd%hLenParNms(i)=len_trim(parNms(i))
        savSamp%nPlMd%hParNms(j+1:j+len_trim(parNms(i)))=transfer(trim(parNms(i)),savSamp%nPlMd%hParNms)
        j=j+len_trim(parNms(i))
        savSamp%nPlMd%hPriTypes(i)=priTypes(i)
        savSamp%nPlMd%hPriLBnd(i)=priLBnd(i)
        if (parFixed(i)) then
          savSamp%nPlMd%hPriUBnd(i)=priLBnd(i) ! priUBnd(i) should be ok but it needs to be safe
        else
          savSamp%nPlMd%hPriUBnd(i)=priUBnd(i)
        end if
        savSamp%nPlMd%hPriKnees(i)=priKnees(i) ! undef. unless mod. Jef. prior
        savSamp%nPlMd%hPriVce(i)=priVce(i)     ! undef. unless normal prior
        savSamp%nPlMd%hPriNPts(i)=priNPts(i)   ! set to zero for non-user-def. priors
        savSamp%nPlMd%hPriPtsPsdX(i,1:priNPts(i))=priPtsPsdX(i,1:priNPts(i))
        savSamp%nPlMd%hPriPtsY(i,1:priNPts(i))=priPtsY(i,1:priNPts(i))
        savSamp%nPlMd%hPriLogScl(i)=boolToInt1(priLogScl(i))
        savSamp%nPlMd%hPriInvDelta(i)=priInvDelta(i)
        savSamp%nPlMd%hFinPSR(i)=finPSR(i)
        savSamp%nPlMd%hFracIndep(i)=fracIndep(i)
        savSamp%nPlMd%hParEstMode(i)=postrMode(i)
      end do

      do i=1,savSamp%hdr%hNPT(1)
        call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hTPar(i,1:savSamp%hdr%hPTNChains(1)),CRCVal)
      end do
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hLenParNms,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hParNms,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hPriTypes,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hPriLBnd,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hPriUBnd,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hPriKnees,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hPriVce,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hPriNPts,CRCVal)
      do i=1,curNPars
        if (priTypes(i) == PRI_TY%fil) call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hPriPtsPsdX(i,1:priNPts(i)),CRCVal)
      end do
      do i=1,curNPars
        if (priTypes(i) == PRI_TY%fil) call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hPriPtsY(i,1:priNPts(i)),CRCVal)
      end do
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hPriLogScl,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hPriInvDelta,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hFinPSR,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hFracIndep,CRCVal)
      call writeFilCRC(UN_SAV_SAMP,savSamp%nPlMd%hParEstMode,CRCVal)

      ! samples
      if (compr) then
        call breakIntoSeq(.TRUE.)
      else
        call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
        do iP=1,curNPars
          if (.NOT. parFixed(iP)) call writeFilCRC(UN_SAV_SAMP,chn%links(1:chn%length,iP),CRCVal)
          call printPrgrs([iP],[1],[curNPars])
        end do
      end if                         ! (compr)

      deallocate(savSamp%samp%sNSeq,savSamp%nPlMd%hTPar,savSamp%nPlMd%hLenParNms,savSamp%nPlMd%hParNms, &
       savSamp%nPlMd%hPriTypes,savSamp%nPlMd%hPriLBnd,savSamp%nPlMd%hPriUBnd,savSamp%nPlMd%hPriKnees,savSamp%nPlMd%hPriVce, &
       savSamp%nPlMd%hPriNPts,savSamp%nPlMd%hPriPtsPsdX,savSamp%nPlMd%hPriPtsY,savSamp%nPlMd%hPriLogScl, &
       savSamp%nPlMd%hPriInvDelta,savSamp%nPlMd%hFinPSR,savSamp%nPlMd%hFracIndep,savSamp%nPlMd%hParEstMode)

      if (finish) then               ! write CRC, close file and reset fstCall
        write(UN_SAV_SAMP) CRCVal
        CRCVal=0
        close(UN_SAV_SAMP)
        fstCall=.TRUE.
        iMode=0
      end if

      call endSubsect(UNIT%cons+UNIT%log) ! Writing posterior samples


    else                             ! read


      if (fstCall) then

        ! global header

        fstCall=.FALSE.
        CRCVal=0
        call tryOpenFilRW("posterior-samples",savSamp%filNmRead,UN_RD_SAMP,"old","read","stream")

          ! global header, 1st part (constant length)

        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hIdent,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hRevNr,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hResrvd,CRCVal) ! currently ignored
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hLenDiv,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hNConfFilLins,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hDtStart,CRCVal)
        if (savSamp%hdr%hRevNr(1) < REV_NR_NO_END_PRG) call readFilCRC(UN_RD_SAMP,savSamp%hdr%hDtEnd,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hNDtSets,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hT1,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hTN,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hEpoch,CRCVal) ! contains ref. time chi until rev. 174
        if (savSamp%hdr%hRevNr(1) < REV_NR_EPOCH) call undef(savSamp%hdr%hEpoch)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hTimesAreJD,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hNDt,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hDtCRC,CRCVal)
        if (savSamp%hdr%hRevNr(1) < REV_NR_NO_M_S) call readFilCRC(UN_RD_SAMP,savSamp%hdr%hMS,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hNNPlMds,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hNPT,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hPTNChains,CRCVal)

        if (savSamp%hdr%hIdent /= SAV_SAMP_FIL_IDENT) call err("invalid file type while reading posterior samples")
        if (dtProps%nSets == 0 .OR. (dtProps%nSets > 0 .AND. savSamp%hdr%hDtCRC(1) /= calcDtCRC())) then ! error cast in main
          dtProps%nSets=0
          dtProps%nRec=0
          call undef(dtProps%tRge)
        end if
        dtProps%tRge=tyIntvlDbl(savSamp%hdr%hT1(1),savSamp%hdr%hTN(1)) ! needed e.g. for periodogram mode
        if (isUndef(savSamp%hdr%hEpoch(1))) then
          mdlProps%epoch=intvlCentre(dtProps%tRge)
          call warn("inference using these samples may be flawed since epoch treatment has been changed in rev. "// &
           trim(niceFmt(REV_NR_EPOCH)))
        else
          mdlProps%epoch=savSamp%hdr%hEpoch(1)
        end if
        dtProps%timesAreJD=int1ToBool(savSamp%hdr%hTimesAreJD(1))
        mdlProps%bin=isUndef(savSamp%hdr%hNNPlMds(1))

          ! global header, 2nd part (variable length)

        if (allocated(savSamp%hdr%hDtTy)) deallocate(savSamp%hdr%hDtTy)
        if (allocated(savSamp%hdr%hCmdLin)) deallocate(savSamp%hdr%hCmdLin)
        if (allocated(savSamp%hdr%hHostName)) deallocate(savSamp%hdr%hHostName)
        if (allocated(savSamp%hdr%hSysNm)) deallocate(savSamp%hdr%hSysNm)
        if (allocated(savSamp%hdr%hConfFilLinLen)) deallocate(savSamp%hdr%hConfFilLinLen)
        allocate(savSamp%hdr%hDtTy(savSamp%hdr%hNDtSets(1)),savSamp%hdr%hCmdLin(savSamp%hdr%hLenDiv(1)), &
         savSamp%hdr%hHostName(savSamp%hdr%hLenDiv(2)),savSamp%hdr%hSysNm(savSamp%hdr%hLenDiv(3)), &
         savSamp%hdr%hConfFilLinLen(savSamp%hdr%hNConfFilLins(1)),stat=allocStat)
        call testAllocErr(allocStat,37)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hDtTy,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hCmdLin,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hHostName,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hSysNm,CRCVal)
        call readFilCRC(UN_RD_SAMP,savSamp%hdr%hConfFilLinLen,CRCVal)
        if (allocated(savSamp%hdr%hConfFilLins)) deallocate(savSamp%hdr%hConfFilLins)
        allocate(savSamp%hdr%hConfFilLins(sum(savSamp%hdr%hConfFilLinLen)),stat=allocStat)
        call testAllocErr(allocStat,43)
        j=0
        do i=1,savSamp%hdr%hNConfFilLins(1)
          call readFilCRC(UN_RD_SAMP,savSamp%hdr%hConfFilLins(j+1:j+savSamp%hdr%hConfFilLinLen(i)),CRCVal)
          j=j+savSamp%hdr%hConfFilLinLen(i)
        end do

        dtProps%dtTyOfSets(1:savSamp%hdr%hNDtSets(1))=savSamp%hdr%hDtTy(1:savSamp%hdr%hNDtSets(1))
        dtProps%nSetsOfTy=0
        call undef(dtProps%onlyDtTy)
        if (dtProps%nRec > 0) then
          do i=1,N_DT_TYPES
            dtProps%nSetsOfTy(i)=count(dtProps%dtTyOfSets(1:dtProps%nSets) == i)
            if (all(dtProps%dtTyOfSets(:dtProps%nSets) == i)) then
              dtProps%onlyDtTy=i
              exit
            end if
          end do
          dtProps%onlyRV=all(dtProps%dtTyOfSets(:dtProps%nSets) == DT_TY%RV)
        end if
        dtProps%sysNm=""
        do i=1,savSamp%hdr%hLenDiv(3)
          dtProps%sysNm(i:i)=savSamp%hdr%hSysNm(i)
        end do


      else                           ! after 1st call

        call startSubsect(UNIT%cons+UNIT%log,"Reading posterior samples")

        do                           ! loop for finding requested n-pl. mode

          call inc(iMode)


          ! n-planet-mode header

            ! n-planet-mode header, 1st part (constant length)

          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hNPl,CRCVal)

          call startSubsect(UNIT%cons+UNIT%log,"Mode "//trim(niceFmt(iMode))//" ("// &
           trim(mrg("binary",plStr(VAL_UNDEF_INT,int(savSamp%nPlMd%hNPl(1),i4b)," "),isUndef(savSamp%nPlMd%hNPl(1))))//")")

          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hNPars,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hNSamp,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hNSampBurn,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hNThin,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hMeanPTBlkLen,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hTParAdj,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hFinByUsrReq,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hComprMd,CRCVal)

          if (isUndef(savSamp%nPlMd%hNPl(1))) then
            mdlProps%bin=.TRUE.
            curNPl=0
            psdNPl=1
            call undef(psdNPlUndef)
          else
            mdlProps%bin=.FALSE.
            curNPl=savSamp%nPlMd%hNPl(1)
            psdNPl=curNPl
            psdNPlUndef=curNPl
          end if
          curNPars=savSamp%nPlMd%hNPars(1)
          chn%length=savSamp%nPlMd%hNSamp(1)
          chn%burnInLen=savSamp%nPlMd%hNSampBurn(1)
          meanPTBlkLen(iNPlMd)=savSamp%nPlMd%hMeanPTBlkLen(1)
          usrAdjustTPar(iNPlMd)=int1ToBool(savSamp%nPlMd%hTParAdj(1))
          finishedByUsrReq=int1ToBool(savSamp%nPlMd%hFinByUsrReq(1))
          compr=savSamp%nPlMd%hComprMd(1) == 1

            ! n-planet-mode header, 2nd part (variable length)

          if (allocated(savSamp%nPlMd%hTPar)) deallocate(savSamp%nPlMd%hTPar)
          if (allocated(savSamp%nPlMd%hLenParNms)) deallocate(savSamp%nPlMd%hLenParNms)
          if (allocated(savSamp%nPlMd%hPriTypes)) deallocate(savSamp%nPlMd%hPriTypes)
          if (allocated(savSamp%nPlMd%hPriLBnd)) deallocate(savSamp%nPlMd%hPriLBnd)
          if (allocated(savSamp%nPlMd%hPriUBnd)) deallocate(savSamp%nPlMd%hPriUBnd)
          if (allocated(savSamp%nPlMd%hPriKnees)) deallocate(savSamp%nPlMd%hPriKnees)
          if (allocated(savSamp%nPlMd%hPriVce)) deallocate(savSamp%nPlMd%hPriVce)
          if (allocated(savSamp%nPlMd%hPriNPts)) deallocate(savSamp%nPlMd%hPriNPts)
          if (allocated(savSamp%nPlMd%hPriLogScl)) deallocate(savSamp%nPlMd%hPriLogScl)
          if (allocated(savSamp%nPlMd%hPriInvDelta)) deallocate(savSamp%nPlMd%hPriInvDelta)
          if (allocated(savSamp%nPlMd%hFinPSR)) deallocate(savSamp%nPlMd%hFinPSR)
          if (allocated(savSamp%nPlMd%hFracIndep)) deallocate(savSamp%nPlMd%hFracIndep)
          if (allocated(savSamp%nPlMd%hParEstMode)) deallocate(savSamp%nPlMd%hParEstMode)
          allocate(savSamp%nPlMd%hTPar(savSamp%hdr%hNPT(1),savSamp%hdr%hPTNChains(1)),savSamp%nPlMd%hLenParNms(curNPars), &
           savSamp%nPlMd%hPriTypes(curNPars),savSamp%nPlMd%hPriLBnd(curNPars),savSamp%nPlMd%hPriUBnd(curNPars), &
           savSamp%nPlMd%hPriKnees(curNPars),savSamp%nPlMd%hPriVce(curNPars), &
           savSamp%nPlMd%hPriNPts(curNPars),savSamp%nPlMd%hPriLogScl(curNPars),savSamp%nPlMd%hPriInvDelta(curNPars), &
           savSamp%nPlMd%hFinPSR(curNPars),savSamp%nPlMd%hFracIndep(curNPars),savSamp%nPlMd%hParEstMode(curNPars),stat=allocStat)
          call testAllocErr(allocStat,38)
          do i=1,savSamp%hdr%hNPT(1)
            call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hTPar(i,1:savSamp%hdr%hPTNChains(1)),CRCVal)
          end do
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hLenParNms,CRCVal)
          if (allocated(savSamp%nPlMd%hParNms)) deallocate(savSamp%nPlMd%hParNms)
          allocate(savSamp%nPlMd%hParNms(sum(savSamp%nPlMd%hLenParNms)),stat=allocStat)
          call testAllocErr(allocStat,44)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hParNms,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hPriTypes,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hPriLBnd,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hPriUBnd,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hPriKnees,CRCVal)
          if (savSamp%hdr%hRevNr(1) >= REV_NR_PRI_VCE) then
            call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hPriVce,CRCVal)
          else
            call undef(savSamp%nPlMd%hPriVce)
          end if

          ! read user-def. points
          if (savSamp%hdr%hRevNr(1) >= REV_NR_USER_DEF_PRI) then
            call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hPriNPts,CRCVal)
            priNPtsMax=maxval(savSamp%nPlMd%hPriNPts)
            if (priNPtsMax > PRI_N_PTS_MAX) call err("saved-samples file contains too many (> "//trim(niceFmt(PRI_N_PTS_MAX))// &
             ") prior points")
            if (allocated(savSamp%nPlMd%hPriPtsPsdX)) deallocate(savSamp%nPlMd%hPriPtsPsdX)
            if (allocated(savSamp%nPlMd%hPriPtsY)) deallocate(savSamp%nPlMd%hPriPtsY)
            allocate(savSamp%nPlMd%hPriPtsPsdX(curNPars,priNPtsMax),savSamp%nPlMd%hPriPtsY(curNPars,priNPtsMax), &
             stat=allocStat)
            call testAllocErr(allocStat,53)
            do i=1,curNPars
              if (savSamp%nPlMd%hPriNPts(i) > 0) &
               call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hPriPtsPsdX(i,1:savSamp%nPlMd%hPriNPts(i)),CRCVal)
            end do
            do i=1,curNPars
              if (savSamp%nPlMd%hPriNPts(i) > 0) &
               call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hPriPtsY(i,1:savSamp%nPlMd%hPriNPts(i)),CRCVal)
            end do
            priAvail=.TRUE.
          else if (any(savSamp%nPlMd%hPriTypes == PRI_TY%fil)) then
            priAvail=.FALSE.
            call warn("at least one parameter has user-defined prior shape with points not available; skipping inference which &
             &explicitly uses priors")
          end if

          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hPriLogScl,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hPriInvDelta,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hFinPSR,CRCVal)
          call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hFracIndep,CRCVal)

          ! read posterior mode
          if (savSamp%hdr%hRevNr(1) >= REV_NR_ZERO_SAMPLE_BUG) then
            call readFilCRC(UN_RD_SAMP,savSamp%nPlMd%hParEstMode,CRCVal)
            postrMode(1:curNPars)=savSamp%nPlMd%hParEstMode
          end if

          if (allocated(tPar)) deallocate(tPar)
          if (allocated(finPSR)) deallocate(finPSR)
          allocate(tPar(savSamp%hdr%hNPT(1),savSamp%hdr%hPTNChains(1)),finPSR(curNPars),stat=allocStat)
          call testAllocErr(allocStat,40)
          tPar=savSamp%nPlMd%hTPar
          k=0
          do i=1,curNPars
            do j=1,savSamp%nPlMd%hLenParNms(i)
              call inc(k)
              tmpStr(j:j)=savSamp%nPlMd%hParNms(k)
            end do
            parNms(i)=tmpStr(1:savSamp%nPlMd%hLenParNms(i))
            priTypes(i)=savSamp%nPlMd%hPriTypes(i)
            priLBnd(i)=savSamp%nPlMd%hPriLBnd(i)
            priUBnd(i)=savSamp%nPlMd%hPriUBnd(i)
            parFixed(i)=priLBnd(i) == priUBnd(i)
            priKnees(i)=savSamp%nPlMd%hPriKnees(i)
            priVce(i)=savSamp%nPlMd%hPriVce(i)
            if (savSamp%hdr%hRevNr(1) >= REV_NR_USER_DEF_PRI) then
              priNPts(i)=savSamp%nPlMd%hPriNPts(i)
              if (priNPts(i) > 0) then
                priPtsPsdX(i,1:priNPts(i))=savSamp%nPlMd%hPriPtsPsdX(i,1:priNPts(i))
                priPtsY(i,1:priNPts(i))=savSamp%nPlMd%hPriPtsY(i,1:priNPts(i))
                priLogScl(i)=int1ToBool(savSamp%nPlMd%hPriLogScl(i))
                priInvDelta(i)=savSamp%nPlMd%hPriInvDelta(i)
                call setPriDiffFromFil(i)
              end if
            end if
            call setPri(i,tyIntvlDbl(priLBnd(i),priUBnd(i)))
            fracIndep(i)=savSamp%nPlMd%hFracIndep(i)
          end do
          if (savSamp%hdr%hRevNr(1) >= REV_NR_USER_DEF_PRI) deallocate(savSamp%nPlMd%hPriPtsPsdX,savSamp%nPlMd%hPriPtsY)

          curNParsFree=curNPars-count(parFixed(:curNPars))
          finPSR=savSamp%nPlMd%hFinPSR
          fracIndep(1:curNPars)=savSamp%nPlMd%hFracIndep


          ! samples

          if (allocated(chn%links)) deallocate(chn%links)
          if (allocated(chn%logLikh)) deallocate(chn%logLikh)
          allocate(chn%links(chn%length,curNPars),chn%logLikh(chn%length),stat=allocStat)
          call testAllocErr(allocStat,39)

          if (compr) then

            if (allocated(savSamp%samp%sNSeq)) deallocate(savSamp%samp%sNSeq)
            allocate(savSamp%samp%sNSeq(curNPars),stat=allocStat)
            call testAllocErr(allocStat,42)
            call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
            do iP=1,curNPars
              if (parFixed(iP)) then
                chn%links(1:chn%length,iP)=priLBnd(iP)
              else
                call readFilCRC(UN_RD_SAMP,savSamp%samp%sNSeq(iP:iP),CRCVal) ! read # sequences
                iSamp=1
                do iSeq=1,savSamp%samp%sNSeq(iP)
                  call readFilCRC(UN_RD_SAMP,savSamp%samp%sSeqSiz,CRCVal) ! sequence length
                  call readFilCRC(UN_RD_SAMP,savSamp%samp%sSeqVal,CRCVal) ! sequence val.
                  do j=0,savSamp%samp%sSeqSiz(1)-1
                    chn%links(min(chn%length,iSamp+j),iP)=savSamp%samp%sSeqVal(1)
                  end do
                  iSamp=iSamp+savSamp%samp%sSeqSiz(1)
                end do
                call printPrgrs([VAL_UNDEF_INT],nTot=curNParsFree)
              end if                 ! (parFixed(iP))
            end do
            deallocate(savSamp%samp%sNSeq)

          else

            call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
            do iP=1,curNPars
              if (.NOT. parFixed(iP)) then
                do iSamp=1,chn%length
                  call readFilCRC(UN_RD_SAMP,savSamp%samp%sSeqVal,CRCVal)
                  chn%links(min(chn%length,iSamp),iP)=savSamp%samp%sSeqVal(1)
                end do
                call printPrgrs([VAL_UNDEF_INT],nTot=curNParsFree)
              end if
            end do

          end if                     ! (compr)

          if (savSamp%hdr%hRevNr(1) <= REV_NR_ZERO_SAMPLE_BUG) then
            call startSubsect(UNIT%cons+UNIT%log,"Checking for zero-valued samples due to bug in earlier BASE versions")
            nSamp0=0
            nSamp0Burn=0
            call printPrgrs([VAL_UNDEF_INT8],reset=.TRUE.)
            do i=1,100
              do iSamp=i1Sub(1_i8b,chn%length,1_i8b,100_i8b,int(i,i8b)),i2Sub(1_i8b,chn%length,1_i8b,100_i8b,int(i,i8b))
                if (all(chn%links(iSamp,:) == 0._dp)) then
                  nSamp0=nSamp0+1
                  if (iSamp <= chn%burnInLen) nSamp0Burn=nSamp0Burn+1
                  if (nSamp0 > N_SAMP_ZERO_MAX) call err("saved-samples file contains more than "//trim(niceFmt(N_SAMP_ZERO_MAX)) &
                   //" zero-valued samples due to bug in earlier BASE versions")
                  iSamp0(nSamp0)=iSamp
                end if
              end do
              call printPrgrs([i],[1],[100])
            end do
            call endSubsect(UNIT%cons+UNIT%log) ! Checking for zero-valued samples due to bug in earlier BASE versions

            if (nSamp0 > 0) then     ! remove zero-valued samples
              call warn(trim(niceFmt(nSamp0))//" saved sample"//trim(mrg("s are"," is",nSamp0 > 1))// &
               " zero-valued due to bug in earlier BASE versions; correct PSR and number of independent samples thus unknown")
              call startSubsect(UNIT%cons+UNIT%log,"Safely discarding zero-valued sample"//trim(merge(" ","s",nSamp0 == 1)))
              iSamp0(nSamp0+1)=chn%length+1
              iSamp=1
              i0=0
              call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
              do iSamp2=1,nSamp0+1
                chn%links(iSamp:iSamp+iSamp0(iSamp2)-i0-2,:)=chn%links(i0+1:iSamp0(iSamp2)-1,:)
                iSamp=iSamp+iSamp0(iSamp2)-i0-1
                i0=iSamp0(iSamp2)
                call printPrgrs([iSamp2],[1_i8b],[nSamp0+1])
              end do
              chn%length=chn%length-nSamp0
              chn%burnInLen=chn%length-nSamp0Burn
              call endSubsect(UNIT%cons+UNIT%log) ! Safely discarding zero-valued samples
              call undef(finPSR)
              call undef(fracIndep)
            end if
          end if                     ! (savSamp%hdr%hRevNr(1) <= REV_NR_ZERO_SAMPLE_BUG)

          if (psdNPlUndef == nPlReq .AND. nThin(iNPlMd) /= 1) then
            call startSubsect(UNIT%cons+UNIT%log,"Thinning samples")
            call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
            iSamp=modulo(chn%length,nThin(iNPlMd))
            if (iSamp == 0) iSamp=nThin(iNPlMd)
            chn%length=ceiling(1._dp*chn%length/nThin(iNPlMd))
            chn%burnInLen=ceiling(1._dp*chn%burnInLen/nThin(iNPlMd))
            chn%links(1:chn%length,:)=chn%links(iSamp:chn%length:nThin(iNPlMd),:)
            call printPrgrs([1],[1],[1])
            call endSubsect(UNIT%cons+UNIT%log) ! Thinning samples
          end if
          
          deallocate(savSamp%nPlMd%hTPar,savSamp%nPlMd%hLenParNms,savSamp%nPlMd%hParNms,savSamp%nPlMd%hPriTypes, &
           savSamp%nPlMd%hPriLBnd,savSamp%nPlMd%hPriUBnd,savSamp%nPlMd%hPriKnees,savSamp%nPlMd%hPriVce,savSamp%nPlMd%hPriNPts, &
           savSamp%nPlMd%hPriLogScl,savSamp%nPlMd%hPriInvDelta,savSamp%nPlMd%hFinPSR,savSamp%nPlMd%hFracIndep)

          call endSubsect(UNIT%cons+UNIT%log) ! … mode

          if (psdNPlUndef == nPlReq .OR. mdlProps%bin .OR. iMode == savSamp%hdr%hNNPlMds(1)) exit

        end do                       ! loop for finding requested n-pl. mode

        call endSubsect(UNIT%cons+UNIT%log) ! Reading posterior samples

        if (psdNPlUndef /= nPlReq) call err("requested number of planets or binary mode not found in saved posterior samples")

        if (iMode == nNPlMds) then   ! check CRC value, close file and reset fstCall
          read(UN_RD_SAMP) CRCValSav
          if (CRCValSav /= CRCVal) call err("CRC-32 failed after reading posterior samples: file corrupted")
          close(UN_RD_SAMP)
          fstCall=.TRUE.
        end if

      end if                         ! (fstCall)

    end if                           ! (output)


    contains

    subroutine breakIntoSeq(wrt)     ! breaks samples into equal-valued sequences of length <= SAV_SAMP_MAX_SEQ_SIZ
      implicit none
      integer :: &
       iP
      integer(i8b) :: &
       iSub
      logical :: &
       wrt
      intent(in) :: &
       wrt

      if (.NOT. wrt) savSamp%samp%sNSeq=0

      do iP=1,curNPars

        if (.NOT. parFixed(iP)) then

          if (wrt) call writeFilCRC(UN_SAV_SAMP,savSamp%samp%sNSeq(iP:iP),CRCVal) ! write # sequences
          lstSeqStart=1
          savSamp%samp%sSeqVal=chn%links(1,iP)

          do iSub=1_i8b,100_i8b
            do iSamp=i1Sub(2_i8b,chn%length,1_i8b,100_i8b,iSub),i2Sub(2_i8b,chn%length,1_i8b,100_i8b,iSub)
              curVal=chn%links(iSamp,iP)
              if (curVal /= savSamp%samp%sSeqVal(1)) then ! new sequence
                call newSeq(iP,wrt)
                lstSeqStart=iSamp
              end if
              savSamp%samp%sSeqVal(1)=curVal
            end do
            if (wrt) call printPrgrs([int(iP,i8b),iSub],[1_i8b,1_i8b],[int(curNPars,i8b),100_i8b])
          end do

          ! last sample
          iSamp=chn%length+1
          call newSeq(iP,wrt)

        end if

      end do
    end subroutine


    subroutine newSeq(iP,wrt)        ! starts new sequence of samples
      implicit none
      integer :: &
       iP
      logical :: &
       wrt
      intent(in) :: &
       iP, &
       wrt

      curSeqSiz=int(iSamp-lstSeqStart,i4b)
      seqPos=1
      do
        savSamp%samp%sSeqSiz=min(SAV_SAMP_MAX_SEQ_SIZ,curSeqSiz-(seqPos-1))
        if (wrt) then
          call writeFilCRC(UN_SAV_SAMP,savSamp%samp%sSeqSiz,CRCVal) ! sequence length
          call writeFilCRC(UN_SAV_SAMP,savSamp%samp%sSeqVal,CRCVal) ! sequence val.
        else
          savSamp%samp%sNSeq(iP)=savSamp%samp%sNSeq(iP)+1
        end if
        seqPos=seqPos+savSamp%samp%sSeqSiz(1)
        if (seqPos > curSeqSiz) exit
      end do
    end subroutine


    function calcDtCRC()             ! returns the data CRC
      implicit none
      integer :: &
       iSet, &
       iRec
      integer(i4b) :: &
       calcDtCRC
      integer(i1b) :: &
       arr1(1)

      calcDtCRC=0
      do iSet=1,dtProps%nSetsOfTy(DT_TY%AMAngPos)
        do iRec=1,dtSetsAMAngPos(iSet)%N
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMAngPos(iSet)%t(iRec),arr1)) ! transfer() needs to be done by component
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMAngPos(iSet)%obs(iRec)%x,arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMAngPos(iSet)%obs(iRec)%y,arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMAngPos(iSet)%unc(iRec)%a,arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMAngPos(iSet)%unc(iRec)%b,arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMAngPos(iSet)%unc(iRec)%phi,arr1))
        end do
      end do

      do iSet=1,dtProps%nSetsOfTy(DT_TY%RV)
        do iRec=1,dtSetsRV(iSet)%N
          call updtCRCInt1(calcDtCRC,transfer(dtSetsRV(iSet)%t(iRec),arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsRV(iSet)%obs(iRec),arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsRV(iSet)%sgm(iRec),arr1))
        end do
      end do

      do iSet=1,dtProps%nSetsOfTy(DT_TY%AMHipN)
        do iRec=1,dtSetsAMHipN(iSet)%N
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%t(iRec),arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%obs(iRec),arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%sgm(iRec),arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%params%psi(iRec),arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%params%cosPsi(iRec),arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%params%sinPsi(iRec),arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%params%gamma(iRec),arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%params%pos0AtT(iRec)%x,arr1))
          call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%params%pos0AtT(iRec)%y,arr1))
        end do
        call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%params%pos0%x,arr1))
        call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%params%pos0%y,arr1))
        call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%params%mu0%x,arr1))
        call updtCRCInt1(calcDtCRC,transfer(dtSetsAMHipN(iSet)%params%mu0%y,arr1))
      end do
    end function

  end subroutine postrSampIO




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcLikh(chn)
!
!   calculates likelihood for all chain links 
!
!   Arguments:
!     chn                            MCMC chain
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     j, &
     iSamp
    type(tyMCMCChain) :: &
     chn
    intent(inout) :: &
     chn

    call startSubsect(UNIT%cons+UNIT%log,"Calculating likelihoods")
    do j=1_i8b,100_i8b
      !$OMP parallel do
      do iSamp=i1Sub(1_i8b,chn%length,1_i8b,100_i8b,j),i2Sub(1_i8b,chn%length,1_i8b,100_i8b,j)
        chn%logLikh(iSamp)=real(logLikh(chn%links(iSamp,:)),qp)
      end do
      !$OMP end parallel do
      call printPrgrs([j],[1_i8b],[100_i8b])
    end do
    call endSubsect(UNIT%cons+UNIT%log) ! Calculating likelihoods

  end subroutine calcLikh




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printPostrSampProps(general)
!
!   prints properties of posterior samples
!
!   Arguments:
!     general                        whether general properties should be printed
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    logical :: &
     general
    intent(in) :: &
     general

    call startSubsect(UNIT%cons+UNIT%log,"Properties of saved samples and original BASE run")

    if (general) then

      call printQty(UNIT%cons+UNIT%log,"Saved with BASE rev.",savSamp%hdr%hRevNr(1))
      call printQty(UNIT%cons+UNIT%log,"Local date & time BASE was started",fmtDate(savSamp%hdr%hDtStart))
      call printQty(UNIT%cons+UNIT%log,"Host computer name",transfer(savSamp%hdr%hHostName,repeat("a",savSamp%hdr%hLenDiv(2))))
      call printQty(UNIT%cons+UNIT%log,"# parallel-tempering procedures",savSamp%hdr%hNPT(1))
      call printQty(UNIT%cons+UNIT%log,"# chains for parallel tempering",savSamp%hdr%hPTNChains(1))
      call startSubsect(UNIT%cons+UNIT%log,"Original command line")
      call printMsg(UNIT%cons+UNIT%log,transfer(savSamp%hdr%hCmdLin,repeat("a",savSamp%hdr%hLenDiv(1))))
      call endSubsect(UNIT%cons+UNIT%log)
      call printQty(UNIT%cons+UNIT%log,"Epoch",savSamp%hdr%hEpoch(1))

    else

      if (isDef(savSamp%nPlMd%hNPl(1))) call printQty(UNIT%cons+UNIT%log,"# planets",savSamp%nPlMd%hNPl(1))
      call printQty(UNIT%cons+UNIT%log,"Finished by user request",finishedByUsrReq)
      call printQty(UNIT%cons+UNIT%log,"Original thinning stride",savSamp%nPlMd%hNThin(1))
      call printQty(UNIT%cons+UNIT%log,"Mean spacing between swap proposals",savSamp%nPlMd%hMeanPTBlkLen(1))
      call printQty(UNIT%cons+UNIT%log,"Length of chain",savSamp%nPlMd%hNSamp(1))
      call printQty(UNIT%cons+UNIT%log,"Length of burn-in period",savSamp%nPlMd%hNSampBurn(1))
      call printQty(UNIT%cons+UNIT%log,"# links used for inference",savSamp%nPlMd%hNSamp(1)-savSamp%nPlMd%hNSampBurn(1))
      call startSubsect(UNIT%cons+UNIT%log,"Tempering parameters")
      call printQty(UNIT%cons+UNIT%log,"Adjusted",usrAdjustTPar(iNPlMd))
      call startSubsect(UNIT%cons+UNIT%log,"Values")
      do i=1,savSamp%hdr%hPTNChains(1)
        call printQty(UNIT%cons+UNIT%log,"Chain "//trim(niceFmt(i)),tPar(1,i))
      end do
      call endSubsect(UNIT%cons+UNIT%log) ! Values
      call endSubsect(UNIT%cons+UNIT%log) ! Tempering parameters
      call printQty(UNIT%cons+UNIT%log,"Compression mode",trim(mrg(STR_NONE,"RLE",savSamp%nPlMd%hComprMd(1) == 0)))

    end if

    call endSubsect(UNIT%cons+UNIT%log) ! Properties of saved samples and original BASE run

  end subroutine printPostrSampProps




end module
